/*
 * COPYRIGHT:  2007 Christopher Smith
 * LICENCE:    GPL v2: See COPYING in the top level directory
 * PROJECT:    regfs
 * FILE:       buffer.c
 * PURPOSE:    Registry filesystem buffer support
 */

#include "regfs.h"

ULONG NTAPI GetIrpBufferTypeAndSize(PIO_STACK_LOCATION IrpSp, LOCK_OPERATION *Operation)
{
	switch(IrpSp->MajorFunction)
	{
	case IRP_MJ_READ:
		*Operation = IoWriteAccess;
		return IrpSp->Parameters.Read.Length;
	case IRP_MJ_WRITE:
		*Operation = IoReadAccess;
		return IrpSp->Parameters.Write.Length;
	case IRP_MJ_QUERY_INFORMATION:
		*Operation = IoWriteAccess;
		return IrpSp->Parameters.QueryFile.Length;
	case IRP_MJ_SET_INFORMATION:
		*Operation = IoReadAccess;
		return IrpSp->Parameters.SetFile.Length;
	case IRP_MJ_QUERY_VOLUME_INFORMATION:
		*Operation = IoWriteAccess;
		return IrpSp->Parameters.QueryVolume.Length;
	case IRP_MJ_SET_VOLUME_INFORMATION:
		*Operation = IoReadAccess;
		return HDRHELP_EXT_IRPSP(IrpSp)->Parameters.SetVolume.Length;
	case IRP_MJ_QUERY_EA:
		*Operation = IoWriteAccess;
		return HDRHELP_EXT_IRPSP(IrpSp)->Parameters.QueryEa.Length;
	case IRP_MJ_SET_EA:
		*Operation = IoReadAccess;
		return HDRHELP_EXT_IRPSP(IrpSp)->Parameters.SetEa.Length;
	case IRP_MJ_DIRECTORY_CONTROL:
		switch(IrpSp->MinorFunction)
		{
		case IRP_MN_QUERY_DIRECTORY:
			*Operation = IoWriteAccess;
			return HDRHELP_EXT_IRPSP(IrpSp)->Parameters.QueryDirectory.Length;
		case IRP_MN_NOTIFY_CHANGE_DIRECTORY:
			*Operation = IoReadAccess;
			return HDRHELP_EXT_IRPSP(IrpSp)->Parameters.NotifyDirectory.Length;
		default:
			return 0;
		}
	case IRP_MJ_DEVICE_CONTROL:
		*Operation = IoModifyAccess;
		return IrpSp->Parameters.DeviceIoControl.OutputBufferLength;
	case IRP_MJ_FILE_SYSTEM_CONTROL:
		*Operation = IoModifyAccess;
		return HDRHELP_EXT_IRPSP(IrpSp)->Parameters.FileSystemControl.OutputBufferLength;
	case IRP_MJ_QUERY_SECURITY:
		*Operation = IoWriteAccess;
		return IrpSp->Parameters.QuerySecurity.Length;
	case IRP_MJ_SET_QUOTA:
		*Operation = IoReadAccess;
		return HDRHELP_EXT_IRPSP(IrpSp)->Parameters.SetQuota.Length;
	case IRP_MJ_QUERY_QUOTA:
		*Operation = IoWriteAccess;
		return HDRHELP_EXT_IRPSP(IrpSp)->Parameters.QueryQuota.Length;
	default:
		return 0;
	}
}

NTSTATUS NTAPI LockAndMapMdl(IN PMDL Mdl, IN KPROCESSOR_MODE RequestorMode, IN LOCK_OPERATION Operation, IN UCHAR LockCount, OUT PVOID *Buffer)
{
	NTSTATUS Status;

	try
	{
		MmProbeAndLockPages(Mdl, RequestorMode, Operation);

		while(LockCount)
			*Buffer = MmMapLockedPagesSafest(Mdl, KernelMode); //safest method availaible; see osver.h
				
		if(!*Buffer)
			TRY_LEAVE(Status = STATUS_NO_MEMORY)

		*Buffer = (PVOID) (((ULONG) *Buffer) | MmGetMdlByteOffset(Mdl)); //to resolve MDL offset change in NT4Sp4; see KB199311
		Status = STATUS_SUCCESS;
	}
	except(EXCEPTION_EXECUTE_HANDLER)
	{
		*Buffer = NULL;
		Status = GetExceptionCode();
	}

	return Status;
}

NTSTATUS NTAPI RegfsStartLocalBuffer(IN PIRP Irp, OUT PVOID *Buffer, OUT PULONG Length, OUT LOCK_OPERATION *Operation)
{
	NTSTATUS Status;

	*Length = GetIrpBufferTypeAndSize(IoGetCurrentIrpStackLocation(Irp), Operation);

	DPRINT_V("Starting local buffer Len:%x ReqMode:%x IrpFlags:%x Operation:%x ", *Length, Irp->RequestorMode, Irp->Flags, *Operation);

	if(*Length == 0)
	{
		*Buffer = NULL;
		DPRINT_V("no buffer.\n");
		return STATUS_SUCCESS;
	}

	if(Irp->AssociatedIrp.SystemBuffer)
	{
		*Buffer = Irp->AssociatedIrp.SystemBuffer;
		DPRINT_V("using SystemBuffer %p.\n", *Buffer);
		return STATUS_SUCCESS;
	}

	if(Irp->MdlAddress)
	{
		//Status = LockAndMapMdl(Irp->MdlAddress, Irp->RequestorMode, *Operation, 1, Buffer);
		DPRINT_V("mapping MDL with Flags:%x SystemVa:%p", Irp->MdlAddress->MdlFlags, Irp->MdlAddress->MappedSystemVa);
		*Buffer = MmMapLockedPagesSafest(Irp->MdlAddress, Irp->RequestorMode);
		//*Buffer = Irp->MdlAddress->MappedSystemVa;
		//*Buffer = MmGetSystemAddressForMdlSafe(Irp->MdlAddress, NormalPagePriority);
		Status = STATUS_SUCCESS;
	}
	else //use the UserBuffer directly
		try
		{
			if(Irp->RequestorMode == UserMode)
			{
				DPRINT_V("probing for neither ");
				switch(*Operation)
				{
				case IoReadAccess:
					ProbeForRead(Irp->UserBuffer, *Length, 1);
					break;
				case IoWriteAccess:
					ProbeForWrite(Irp->UserBuffer, *Length, 1);
					break;
				case IoModifyAccess:
					ProbeForRead(Irp->UserBuffer, *Length, 1);
					ProbeForWrite(Irp->UserBuffer, *Length, 1);
					break;
				}
			}
			Status = STATUS_SUCCESS;
			*Buffer = Irp->UserBuffer;
		}
		except(EXCEPTION_EXECUTE_HANDLER)
		{
			*Length = 0;
			*Buffer = NULL;
			Status = GetExceptionCode();
		}

	DPRINT_V("Buffer %p Status:%x\n", *Buffer, Status);

	return Status;
}

VOID RegfsEndBuffer(PIRP Irp, PVOID Addr)
{
	if(!Addr)
		return;

	if(Irp->AssociatedIrp.SystemBuffer)
		return;

	if(Irp->MdlAddress)
	{
		MmUnmapLockedPages(Addr, Irp->MdlAddress);
		MmUnlockPages(Irp->MdlAddress);
		return;
	}

	return; //cleanup req'd only for MDL
}

NTSTATUS NTAPI RegfsStartRemoteBuffer(IN PIRP Irp, OUT PVOID *Buffer, OUT PULONG Length, OUT LOCK_OPERATION *Operation)
{
	NTSTATUS Status;

	Status = RegfsStartLocalBuffer(Irp, Buffer, Length, Operation);

	if(!NT_SUCCESS(Status) || *Buffer == NULL)
		return Status;

	if(Irp->AssociatedIrp.SystemBuffer || Irp->MdlAddress) //buffer already available in other processes
		return STATUS_SUCCESS;

	//make an MDL
	if(*Length > REGFS_MDL_MAXIMUM)
		return STATUS_INSUFFICIENT_RESOURCES;

	if(!IoAllocateMdl(Irp->UserBuffer, *Length, FALSE, TRUE, Irp))
		return STATUS_INSUFFICIENT_RESOURCES;

	return LockAndMapMdl(Irp->MdlAddress, Irp->RequestorMode, *Operation, 2, Buffer); //one lock for us to free, one for IO manager to free
}

//copies as much of the buffer as possible, returns copied length
ULONG NTAPI RegfsCopyData(PVOID Dest, void const *Src, ULONG SrcLength, ULONG UnusedBuffer, ULONG_PTR MaxBufferAddr)
{
	ULONG Length = min(SrcLength, UnusedBuffer);
	DPRINT_V("CopyData Dest:%p Src:%p SrcLength:%X Length:%X UnusedBuffer:%X\n", Dest, Src, SrcLength, Length, UnusedBuffer);
	if((ULONG_PTR)Dest + Length > MaxBufferAddr)
		DPRINT("*************ATTEMPTED BUFFER OVERRUN: %p>%p\n", (PUCHAR)Dest + Length, MaxBufferAddr);
	else
		RtlCopyMemory(Dest, (PVOID)Src, Length);
	
	{
		UNICODE_STRING UCS;
		UCS.Buffer = (PWCHAR)Dest;
		UCS.MaximumLength = UCS.Length = (USHORT)Length;
		DPRINT_V("Copied:%wZ\n", &UCS);
	}
	return Length;
}

NTSTATUS NTAPI RegfsExpandBufferIfNecessary(PVOID *Buffer, PULONG CurrentLength, ULONG RequiredLength, ULONG Tag)
{
	if(RequiredLength > *CurrentLength)
	{
		if(*Buffer)
			ExFreePoolWithTag(*Buffer, Tag);
		*CurrentLength = RequiredLength;
		*Buffer = ExAllocatePoolWithTag(PagedPool, RequiredLength, Tag);
		if(*Buffer == NULL)
		{
			DPRINT("Tried to alloc %X, failed.\n", RequiredLength);
			return STATUS_NO_MEMORY;
		}
	}
	return STATUS_SUCCESS;
}