/*
 * COPYRIGHT:  2007 Christopher Smith
 * LICENCE:    GPL v2: See COPYING in the top level directory
 * PROJECT:    regfs
 * FILE:       create.c
 * PURPOSE:    Registry filesystem create/close
 */

#include "regfs.h"

#undef DPRINT_V
#define DPRINT_V DPRINT

//tries to create or open (from CreateDisposition) specified key.
//returns KeyAccessMask, KeyHandle and updates Irp->IoStatus.Information (for disposition) if successful
NTSTATUS NTAPI RegfsFileCreateKey(IN PIO_STACK_LOCATION IrpSp, IN OUT PIRP Irp, IN ULONG CreateDisposition, IN ULONG KeyAttributes, IN UNICODE_STRING FileName, OUT PHANDLE KeyHandle, IN HANDLE RelatedKeyHandle, IN KPROCESSOR_MODE RequestorMode, OUT PACCESS_MASK KeyAccessMask, IN ACCESS_MASK PreviousKeyAccess, IN PSECURITY_SUBJECT_CONTEXT SecurityContext)
{
	NTSTATUS Status;
	UNICODE_STRING KeyName;
	ULONG KeyOptions = 0;
	ULONG KeyDisposition = 0;
	REGFS_STRUCTURED_EA EaStruct;

	//request any remaining access, always need KEY_NOTIFY to support notifications
	*KeyAccessMask = (RegfsFileMaskToKeyMask(IrpSp->Parameters.Create.SecurityContext->DesiredAccess) | KEY_NOTIFY) & ~PreviousKeyAccess;

	//either open target dir or open normal dir
	if(IrpSp->Flags & SL_OPEN_TARGET_DIRECTORY)
		RegfsSeperateLastName(FileName, NULL, &KeyName); //KeyName = RegfsTruncateLastName(FileName);
	else
		KeyName = FileName;

	//set up extended attributes
	EaStruct.Class.Buffer = NULL;
	EaStruct.Class.Length = EaStruct.Class.MaximumLength = 0;
	EaStruct.Volatile = FALSE;
	EaStruct.TitleIndex = 0;

	Status = RegfsStructureEaList(TRUE, (PFILE_FULL_EA_INFORMATION)Irp->AssociatedIrp.SystemBuffer, IrpSp->Parameters.Create.EaLength, &EaStruct);

	if(!NT_SUCCESS(Status))
		return Status;

	if(EaStruct.Volatile)
		KeyOptions |= REG_OPTION_VOLATILE;
		
	switch(CreateDisposition)
	{
	case FILE_SUPERSEDE:
	case FILE_OPEN_IF:
	case FILE_OVERWRITE_IF:
		KeyAttributes |= OBJ_OPENIF;
	case FILE_CREATE:
		Status = RegfsCreateKey(KeyHandle, *KeyAccessMask, KeyAccessMask, RequestorMode, SecurityContext, &KeyName, KeyAttributes, RelatedKeyHandle, NULL, EaStruct.TitleIndex, &EaStruct.Class, KeyOptions, NULL);
		break;
	case FILE_OPEN:
	case FILE_OVERWRITE:
		Status = RegfsOpenKey(KeyHandle, *KeyAccessMask, KeyAccessMask, RequestorMode, SecurityContext, &KeyName, KeyAttributes, RelatedKeyHandle);
		break;
	default:
		DPRINT("Invalid create disposition.\n");
		return STATUS_INVALID_PARAMETER;
	}

	DPRINT_V("Trying to open as directory. KeyName:%wZ OpenIf?%X Disposition:%X\n", &KeyName, !!(KeyAttributes & OBJ_OPENIF), CreateDisposition);

	if(Status == STATUS_OBJECT_NAME_NOT_FOUND)
		DPRINT_V("Key not found.\n");

	if(!NT_SUCCESS(Status))
		return Status;

	//return disposition
	switch(KeyDisposition)
	{
	case REG_CREATED_NEW_KEY:
		DPRINT_V("Created key.\n");
		Irp->IoStatus.Information = FILE_CREATED;
		break;
	case REG_OPENED_EXISTING_KEY:
		DPRINT_V("Opened key.\n");
		Irp->IoStatus.Information = FILE_OPENED;
		break;
	}

	return Status;
}

NTSTATUS NTAPI RegfsFileCreateKeyValue(IN PIO_STACK_LOCATION IrpSp, IN PIRP Irp, IN ULONG CreateDisposition, IN ULONG KeyAttributes, IN UNICODE_STRING FileName, OUT PHANDLE KeyHandle, IN HANDLE RelatedKeyHandle, IN KPROCESSOR_MODE RequestorMode, OUT PACCESS_MASK KeyAccessMask, IN ACCESS_MASK PreviousKeyAccess, IN PSECURITY_SUBJECT_CONTEXT SecurityContext)
{
	NTSTATUS Status;
	UNICODE_STRING KeyName;
	UNICODE_STRING ValueName;
	BOOLEAN ShouldWriteEmptyValue;
	REGFS_STRUCTURED_EA EaStruct;

	RegfsSeperateLastName(FileName, &ValueName, &KeyName);

	if(RtlEqualUnicodeString(&ValueName, &REGFS_DEFAULT_VALUE_ALIAS, FALSE)) //special name for an unnamed value
		ValueName.Length = 0;

	*KeyAccessMask = (RegfsFileMaskToKeyValueMask(IrpSp->Parameters.Create.SecurityContext->DesiredAccess) | KEY_NOTIFY) & ~PreviousKeyAccess;

	DPRINT_V("Trying to open as file. KeyName:%wZ ValueName:%wZ\n", &KeyName, &ValueName);

	//set up extended attributes
	EaStruct.Type = DefaultKeyValueType;
	EaStruct.TitleIndex = 0;

	Status = RegfsStructureEaList(TRUE, (PFILE_FULL_EA_INFORMATION)Irp->AssociatedIrp.SystemBuffer, IrpSp->Parameters.Create.EaLength, &EaStruct);

	if(!NT_SUCCESS(Status))
		return Status;

	//open key
	Status = RegfsOpenKey(KeyHandle, *KeyAccessMask, KeyAccessMask, RequestorMode, SecurityContext, &KeyName, KeyAttributes, RelatedKeyHandle);

	if(!NT_SUCCESS(Status))
		return Status;

	if(CreateDisposition == FILE_SUPERSEDE) //FILE_SUPERSEDE doesn't care if the value already exists, so don't check
	{
		DPRINT_V("Superseding.\n");
		ShouldWriteEmptyValue = TRUE;
		Irp->IoStatus.Information = FILE_SUPERSEDED;
	}
	else
	{
		//find out if the value (file) already exists
		KEY_VALUE_BASIC_INFORMATION kvbi;
		ULONG ResultLength;
		BOOLEAN Exists;

		Status = ZwQueryValueKey(*KeyHandle, &ValueName, KeyValueBasicInformation, &kvbi, sizeof kvbi, &ResultLength);
		if(Status == STATUS_OBJECT_NAME_NOT_FOUND)
			Exists = FALSE;
		else if(NT_SUCCESS(Status) || Status == STATUS_BUFFER_OVERFLOW)
			Exists = TRUE;
		else
		{
			ZwClose(*KeyHandle);
			DPRINT("Error: detirmining value existance failed. Status %X", Status);
			return Status;
		}

		Status = STATUS_SUCCESS;

		//either we know if the value should be (over)written or we complete with an error because it already exists (FILE_SUPERCEDE was already handled)
		if(Exists)
			switch(CreateDisposition)
			{
			case FILE_CREATE:
				Irp->IoStatus.Information = FILE_EXISTS;
				ZwClose(*KeyHandle);
				DPRINT_V("Value already exists.\n");
				return STATUS_OBJECT_NAME_EXISTS;
			case FILE_OVERWRITE:
			case FILE_OVERWRITE_IF:
				DPRINT_V("Overwriting.\n");
				ShouldWriteEmptyValue = TRUE;
				Irp->IoStatus.Information = FILE_OVERWRITTEN;
				break;
			case FILE_OPEN:
			case FILE_OPEN_IF:
				DPRINT_V("Opening.\n");
				Irp->IoStatus.Information = FILE_OPENED;
				ShouldWriteEmptyValue = FALSE;
				break;
			default:
				ZwClose(*KeyHandle);
				DPRINT("Invalid create disposition.\n");
				return STATUS_INVALID_PARAMETER;
			}
		else //!Exists
			switch(CreateDisposition)
			{
			case FILE_CREATE:
			case FILE_OVERWRITE_IF:
			case FILE_OPEN_IF:
				DPRINT_V("Creating.\n");
				Irp->IoStatus.Information = FILE_CREATED;
				ShouldWriteEmptyValue = TRUE;
				break;
			case FILE_OPEN:
			case FILE_OVERWRITE:
				Irp->IoStatus.Information = FILE_DOES_NOT_EXIST;
				ZwClose(*KeyHandle);
				DPRINT_V("Value doesn't exist.\n");
				return STATUS_OBJECT_NAME_NOT_FOUND;
			default:
				ZwClose(*KeyHandle);
				DPRINT("Invalid create disposition.\n");
				return STATUS_INVALID_PARAMETER;
			}
	}//did we care if the value already existed?
	
	DPRINT_V("ShouldWrite? %x\n", ShouldWriteEmptyValue);
	if(ShouldWriteEmptyValue)
	{
		//create an empty buffer and write it to the file
		PVOID Buffer;
		ULONG BufferSize = Irp->Overlay.AllocationSize.LowPart;
		UCHAR EmptyBuffer;
		DPRINT_V("Setting initial value: %wZ\n", &ValueName);
		if(Irp->Overlay.AllocationSize.HighPart)
		{
			ZwClose(*KeyHandle);
			DPRINT("Error: Invalid initial size");
			return STATUS_INVALID_PARAMETER;
		}

		if(BufferSize == 0) //avoid allocating zero bytes
			Buffer = &EmptyBuffer;
		else
			Buffer = ExAllocatePoolWithTag(PagedPool, BufferSize, TAG_BLANK_BUFFER);

		if(!Buffer)
		{
			ZwClose(*KeyHandle);
			DPRINT("Error: Out of memory!");
			return STATUS_NO_MEMORY;
		}
		RtlZeroMemory(Buffer, BufferSize);
		Status = ZwSetValueKey(*KeyHandle, &ValueName, EaStruct.TitleIndex, EaStruct.Type, Buffer, BufferSize);
		
		if(BufferSize != 0)
			ExFreePoolWithTag(Buffer, TAG_BLANK_BUFFER);
		
		if(!NT_SUCCESS(Status))
		{
			ZwClose(*KeyHandle);
			DPRINT("Error: Initial SetValueKey failed.");
			return Status;
		}
	}
	return Status;
}

NTSTATUS NTAPI RegfsCreate(IN PDEVICE_OBJECT DeviceObject, IN PIRP Irp, IN OUT PVOID Buffer, IN ULONG BufferLength, IN PSECURITY_SUBJECT_CONTEXT SecurityContext)
{
	PIO_STACK_LOCATION IrpSp;
	PFILE_OBJECT FileObject;
	PREGFS_DEVICE_EXTENSION DeviceExt;
	HANDLE KeyHandle, RelatedKeyHandle;
	ULONG StartingKeyAttributes;
	UNICODE_STRING FileName;
	KPROCESSOR_MODE RequestorMode;
	ACCESS_MASK PreviousKeyAccess, KeyAccessMask;
	ULONG CreateDisposition;
	BOOLEAN IsDirectory;
	NTSTATUS Status;

	DeviceExt = (PREGFS_DEVICE_EXTENSION)DeviceObject->DeviceExtension;
	IrpSp = (PIO_STACK_LOCATION)IoGetCurrentIrpStackLocation(Irp);
	FileObject = IrpSp->FileObject;
	FileName = FileObject->FileName;

	FileObject->FsContext2 = (PVOID)-1;

	DPRINT("RegfsCreate(DeviceObject %p Irp %p FileObject %p FileName %wZ)\n", DeviceObject, Irp, FileObject, &FileName);

	//remove trailing backslash, if any
	if ((FileName.Length != 0) && (FileName.Buffer[FileName.Length/sizeof(WCHAR)-1] == L'\\'))
		FileName.Length -= sizeof(WCHAR);

	if(FileName.Length == 0)
	{
		FileObject->FsContext2 = NULL;
		return RegfsCompleteStatus(Irp, FALSE, STATUS_SUCCESS, "Direct Volume Open.");
	}

	RequestorMode = Irp->RequestorMode == UserMode || IrpSp->Flags & SL_FORCE_ACCESS_CHECK ? UserMode : KernelMode;

	//not set by dispatch because it's in the create options:
	SecurityContext = &IrpSp->Parameters.Create.SecurityContext->AccessState->SubjectSecurityContext;

	PreviousKeyAccess = RegfsFileMaskToKeyMask(IrpSp->Parameters.Create.SecurityContext->AccessState->PreviouslyGrantedAccess);

	if((IrpSp->Parameters.Create.Options & FILE_OPEN_FOR_BACKUP_INTENT) &&
		(IrpSp->Parameters.Create.SecurityContext->AccessState->Flags & SE_BACKUP_PRIVILEGE))
			PreviousKeyAccess |= KEY_READ;
	
	if((IrpSp->Parameters.Create.Options & FILE_OPEN_FOR_RECOVERY) &&
		(IrpSp->Parameters.Create.SecurityContext->AccessState->Flags & SE_RESTORE_PRIVILEGE))
			PreviousKeyAccess |= KEY_WRITE;

	if(SePrivilegeCheck(&KeyNotifyPrivilegeSet, SecurityContext, RequestorMode))
		PreviousKeyAccess |= KEY_NOTIFY;

	Irp->IoStatus.Information = 0;
	KeyHandle = NULL;

	//translate create parameters
	CreateDisposition = (IrpSp->Parameters.Create.Options >> 24) & 0x000000ff;

	StartingKeyAttributes = 0;

	if(!(IrpSp->Flags & SL_CASE_SENSITIVE))
		StartingKeyAttributes |= OBJ_CASE_INSENSITIVE;

	//use related key handle if any
	if(FileObject->RelatedFileObject && FileObject->RelatedFileObject->FsContext2)
	{
		PREGFS_DFCB RelatedDfcb = ((PREGFS_CCB)FileObject->RelatedFileObject->FsContext2)->Dfcb;
		RelatedKeyHandle = RegfsGetDfcbKeyHandle(RelatedDfcb);
		DPRINT_V("Parent key handle:%p\n", RelatedKeyHandle);
	}
	else
		RelatedKeyHandle = NULL;

	if((IrpSp->Parameters.Create.Options & FILE_DIRECTORY_FILE) || (IrpSp->Flags & SL_OPEN_TARGET_DIRECTORY)) //needs to be a key
	{
		DPRINT_V("Caller needs directory.\n");
		Status = RegfsFileCreateKey(IrpSp, Irp, CreateDisposition, StartingKeyAttributes, FileName, &KeyHandle, RelatedKeyHandle, RequestorMode, &KeyAccessMask, PreviousKeyAccess, SecurityContext);
		if(Status == STATUS_OBJECT_NAME_NOT_FOUND)
		{
			//maybe it exists as a file?
			Status = RegfsFileCreateKeyValue(IrpSp, Irp, FILE_OPEN, StartingKeyAttributes, FileName, &KeyHandle, RelatedKeyHandle, RequestorMode, &KeyAccessMask, PreviousKeyAccess, SecurityContext);
			if(NT_SUCCESS(Status))
			{
				ZwClose(KeyHandle); //we don't need this
				return RegfsCompleteStatus(Irp, TRUE, STATUS_NOT_A_DIRECTORY, "Exists as value only, but needed key.");
			}
			if(Status == STATUS_OBJECT_NAME_NOT_FOUND)
				return RegfsCompleteKeyStatus(Irp, TRUE, Status, "Directory not found.");
		}
		if(!NT_SUCCESS(Status))
			return RegfsCompleteKeyStatus(Irp, TRUE, Status, "Error opening key.");
		
		IsDirectory = TRUE;
	}
	else if(IrpSp->Parameters.Create.Options & FILE_NON_DIRECTORY_FILE) //needs to be a key value
	{
		DPRINT_V("Caller needs file.\n");
		Status = RegfsFileCreateKeyValue(IrpSp, Irp, CreateDisposition, StartingKeyAttributes, FileName, &KeyHandle, RelatedKeyHandle, RequestorMode, &KeyAccessMask, PreviousKeyAccess, SecurityContext);
		if(Status == STATUS_OBJECT_NAME_NOT_FOUND)
		{
			//maybe it exists as a directory?
			Status = RegfsFileCreateKey(IrpSp, Irp, FILE_OPEN, StartingKeyAttributes, FileName, &KeyHandle, RelatedKeyHandle, RequestorMode, &KeyAccessMask, PreviousKeyAccess, SecurityContext);
			if(NT_SUCCESS(Status))
			{
				ZwClose(KeyHandle); //we don't need this
				return RegfsCompleteStatus(Irp, TRUE, STATUS_FILE_IS_A_DIRECTORY, "Exists as key only, but needed value.");
			}
			if(Status == STATUS_OBJECT_NAME_NOT_FOUND)
				return RegfsCompleteKeyStatus(Irp, TRUE, Status, "Directory not found.");
		}
		if(!NT_SUCCESS(Status))
			return RegfsCompleteKeyStatus(Irp, TRUE, Status, "Error opening value.");
		
		IsDirectory = FALSE;
	}
	else //caller didn't specify file type
	{
		DPRINT_V("Caller didn't specify file/directory.\n");
		//try to find as directory
		Status = RegfsFileCreateKey(IrpSp, Irp, FILE_OPEN, StartingKeyAttributes, FileName, &KeyHandle, RelatedKeyHandle, RequestorMode, &KeyAccessMask, PreviousKeyAccess, SecurityContext);
		if(Status == STATUS_OBJECT_NAME_NOT_FOUND)
		{
			//default to open/create as file
			Status = RegfsFileCreateKeyValue(IrpSp, Irp, CreateDisposition, StartingKeyAttributes, FileName, &KeyHandle, RelatedKeyHandle, RequestorMode, &KeyAccessMask, PreviousKeyAccess, SecurityContext);
			if(!NT_SUCCESS(Status))
				return RegfsCompleteKeyStatus(Irp, TRUE, Status, "Couldn't open as key or value.");

			IsDirectory = FALSE;
		}
		else
			IsDirectory = TRUE;

		if(!NT_SUCCESS(Status))
			return RegfsCompleteKeyStatus(Irp, TRUE, Status, "Error opening as key.");
	}

	//allocate, set context for future IO

	{
		PVOID KeyObject;
		POBJECT_NAME_INFORMATION ActualKeyName;
		ULONG ActualKeyNameLength;
		PREGFS_DFCB Dfcb;
		PREGFS_CCB Ccb;

		//get actual name of key (case corrected and post symlinks)
		Status = ObReferenceObjectByHandle(KeyHandle, 0, NULL, KernelMode, &KeyObject, NULL);

		if(!NT_SUCCESS(Status))
		{
			ZwClose(KeyHandle);
			return RegfsCompleteStatus(Irp, TRUE, Status, "Error referencing key.");
		}

		Status = ObQueryNameString(KeyObject, NULL, 0, &ActualKeyNameLength);

		if(Status != STATUS_INFO_LENGTH_MISMATCH && !NT_SUCCESS(Status))
		{
			ZwClose(KeyHandle);
			ObDereferenceObject(KeyObject);
			return RegfsCompleteStatus(Irp, TRUE, Status, "Error finding actual key name length.");
		}

		ActualKeyName = ExAllocatePoolWithTag(PagedPool, ActualKeyNameLength, TAG_ACTUAL_NAME);

		if(!ActualKeyName)
		{
			ZwClose(KeyHandle);
			ObDereferenceObject(KeyObject);
			return RegfsCompleteStatus(Irp, TRUE, STATUS_NO_MEMORY, "Out of memory!");
		}

		Status = ObQueryNameString(KeyObject, ActualKeyName, ActualKeyNameLength, PAGED_BIT_BUCKET);

		ObDereferenceObject(KeyObject);

		if(!NT_SUCCESS(Status))
		{
			ZwClose(KeyHandle);
			ExFreePoolWithTag(ActualKeyName, TAG_ACTUAL_NAME);
			return RegfsCompleteStatus(Irp, TRUE, Status, "Error querying actual key name.");
		}

		if(IsDirectory)
		{
			Status = RegfsReplaceFileName(FileObject, &ActualKeyName->Name, NULL);

			if(!NT_SUCCESS(Status))
			{
				ExFreePoolWithTag(ActualKeyName, TAG_ACTUAL_NAME);
				return RegfsCompleteStatus(Irp, TRUE, Status, "Error setting filename.");
			}

			Status = RegfsDemandDfcb(&Dfcb, &ActualKeyName->Name, NULL, KeyHandle);
			//KeyHandle now owned, possibly closed by RegfsDemandDfcb

			if(!NT_SUCCESS(Status))
			{
				ExFreePoolWithTag(ActualKeyName, TAG_ACTUAL_NAME);
				return RegfsCompleteStatus(Irp, TRUE, Status, "Error getting Dfcb.");
			}
		}
		else
		{
			PKEY_VALUE_FULL_INFORMATION Kvfi = NULL;
			ULONG DataLength;
			
			UNICODE_STRING ValueName;
			RegfsSeperateLastName(FileName, &ValueName, NULL);

			if(RtlEqualUnicodeString(&ValueName, &REGFS_DEFAULT_VALUE_ALIAS, FALSE)) //special name for unnamed key value
				ValueName.Length = ValueName.MaximumLength = 0;

			//don't care about data right now, only name, name length and data length
			Kvfi = ExAllocatePoolWithTag(PagedPool, SIZEOF_KEY_VALUE_FULL_INFORMATION_HEADER + ValueName.Length, TAG_ACTUAL_NAME);

			if(!Kvfi)
			{
				ZwClose(KeyHandle);
				ExFreePoolWithTag(ActualKeyName, TAG_ACTUAL_NAME);
				return RegfsCompleteStatus(Irp, TRUE, STATUS_NO_MEMORY, "Out of memory!");
			}

			Status = ZwQueryValueKey(KeyHandle, &ValueName, KeyValueFullInformation, Kvfi, SIZEOF_KEY_VALUE_FULL_INFORMATION_HEADER + ValueName.Length, PAGED_BIT_BUCKET);

			if(!NT_SUCCESS(Status) && Status != STATUS_BUFFER_OVERFLOW)
			{
				ZwClose(KeyHandle);
				ExFreePoolWithTag(ActualKeyName, TAG_ACTUAL_NAME);
				return RegfsCompleteStatus(Irp, TRUE, Status, "Error querying actual value name.");
			}

			DPRINT_V("Actual ValueName:%wZ\n", &ValueName);

			if(ValueName.Length) //special case for unnamed key value
				Status = RegfsReplaceFileName(FileObject, &ActualKeyName->Name, &ValueName);
			else
				Status = RegfsReplaceFileName(FileObject, &ActualKeyName->Name, (PUNICODE_STRING)&REGFS_DEFAULT_VALUE_ALIAS);
			
			DataLength = Kvfi->DataLength;

			ExFreePoolWithTag(Kvfi, TAG_ACTUAL_NAME);

			if(!NT_SUCCESS(Status))
			{
				ZwClose(KeyHandle);
				ExFreePoolWithTag(ActualKeyName, TAG_ACTUAL_NAME);
				return RegfsCompleteStatus(Irp, TRUE, Status, "Error setting filename.");
			}

			Status = RegfsDemandDfcb(&Dfcb, &ActualKeyName->Name, &FileObject->FileName, KeyHandle);
			//KeyHandle now owned, possibly closed by RegfsDemandDfcb

			if(!NT_SUCCESS(Status))
			{
				ExFreePoolWithTag(ActualKeyName, TAG_ACTUAL_NAME);
				return RegfsCompleteStatus(Irp, TRUE, Status, "Error getting Dfcb.");
			}

			FileObject->FsContext = &Dfcb->Fcb.FcbFsRtl->FsRtl;
			FileObject->SectionObjectPointer = &Dfcb->Fcb.FcbFsRtl->SecPtrs;
		}

		ExFreePoolWithTag(ActualKeyName, TAG_ACTUAL_NAME);

		Ccb = ExAllocatePoolWithTag(PagedPool, sizeof(REGFS_CCB), TAG_CCB);

		if(!Ccb)
		{
			RegfsDereferenceDfcb(Dfcb);
			if(FileObject->FsContext)
				ExFreePoolWithTag(FileObject->FsContext, TAG_FCB_FSRTL);
			return RegfsCompleteStatus(Irp, TRUE, STATUS_NO_MEMORY, "Out of memory!");
		}

		RtlZeroMemory(Ccb, sizeof(REGFS_CCB));

		Ccb->Lock = AllocSimpleMutex(TAG_CCB_LOCK);

		if(!Ccb->Lock)
		{
			ExFreePoolWithTag(Ccb, TAG_CCB);
			if(FileObject->FsContext)
				ExFreePoolWithTag(FileObject->FsContext, TAG_FCB_FSRTL);
			RegfsDereferenceDfcb(Dfcb);
			return RegfsCompleteStatus(Irp, TRUE, STATUS_NO_MEMORY, "Out of memory!");
		}

		if(IrpSp->Parameters.Create.Options & FILE_DELETE_ON_CLOSE)
			InterlockedOr(&Dfcb->Flags, REGFS_DFCB_DELETE_ON_CLOSE);

		Ccb->Dfcb = Dfcb;
		Ccb->EnumMode = EnumDotsMode;
		Ccb->EaEnumIndex = 1; //EAs are 1-index based

		FileObject->FsContext2 = Ccb;
	}

	return RegfsCompleteStatus(Irp, TRUE, STATUS_SUCCESS, "Success.");
}

NTSTATUS NTAPI RegfsCleanup(IN PDEVICE_OBJECT DeviceObject, IN PIRP Irp, IN OUT PVOID Buffer, IN ULONG BufferLength, IN PSECURITY_SUBJECT_CONTEXT SecurityContext)
{
	PIO_STACK_LOCATION IrpSp;
	PFILE_OBJECT FileObject;
	PREGFS_CCB Ccb;
	PREGFS_DFCB Dfcb;
	PREGFS_FCB_FSRTL FcbFsRtl;

	IrpSp = IoGetCurrentIrpStackLocation(Irp);
	FileObject = IrpSp->FileObject;
	Ccb = (PREGFS_CCB)FileObject->FsContext2;
	FcbFsRtl = (PREGFS_FCB_FSRTL)FileObject->FsContext;

	if(!Ccb) //not a real file
		return RegfsCompleteStatus(Irp, FALSE, STATUS_SUCCESS, "RegfsCleanup volume file. Success.");

	DPRINT("RegfsCleanup Irp:%p %wZ\n", Irp, &Ccb->Dfcb->FileName);

	if(!FcbFsRtl) //not a fsrtl file
		return RegfsCompleteStatus(Irp, FALSE, STATUS_SUCCESS, "Success.");

	{//no lingering sections, please
		BOOLEAN Result = MmForceSectionClosed(&FcbFsRtl->SecPtrs, TRUE);
		DPRINT_V("MmForceSectionClosed Result:%x.\n", Result);
	}

	return RegfsCompleteStatus(Irp, FALSE, STATUS_SUCCESS, "Success.");
}

NTSTATUS NTAPI RegfsClose(IN PDEVICE_OBJECT DeviceObject, IN PIRP Irp, IN OUT PVOID Buffer, IN ULONG BufferLength, IN PSECURITY_SUBJECT_CONTEXT SecurityContext)
{
	PIO_STACK_LOCATION IrpSp;
	PFILE_OBJECT FileObject;
	PREGFS_CCB Ccb;
	PREGFS_DFCB Dfcb;
	
	IrpSp = IoGetCurrentIrpStackLocation(Irp);
	FileObject = IrpSp->FileObject;
	Ccb = (PREGFS_CCB)FileObject->FsContext2;

	if(!Ccb) //not a real file
		return RegfsCompleteStatus(Irp, FALSE, STATUS_SUCCESS, "RegfsClose volume file. Success.");	

	Dfcb = Ccb->Dfcb;

	DPRINT("RegfsClose Irp:%p %wZ\n", Irp, &Dfcb->FileName);

	if(Ccb->QueryString.Buffer)
	{
		DPRINT_V("Freeing Ccb->QueryString.Buffer \"%wZ\" %p\n", &Ccb->QueryString, Ccb->QueryString.Buffer);
		ExFreePoolWithTag(Ccb->QueryString.Buffer, TAG_CCB_QUERY_STRING);
	}

	if(Dfcb->Flags & REGFS_DFCB_DELETE_ON_CLOSE && Dfcb->RefCount == 1)
	{
		HANDLE KeyHandle = RegfsGetDfcbKeyHandle(Dfcb);

		if(Dfcb->Flags & REGFS_DFCB_DIRECTORY)
			ZwDeleteKey(KeyHandle);
		else
			ZwDeleteValueKey(KeyHandle, &Dfcb->ShortName);
	}

	RegfsDereferenceDfcb(Dfcb);

	FreeSimpleMutex(Ccb->Lock, TAG_CCB_LOCK);

	ExFreePoolWithTag(Ccb, TAG_CCB);

	FileObject->FsContext2 = (PVOID)-2;

	return RegfsCompleteStatus(Irp, FALSE, STATUS_SUCCESS, "Success.");
}