/*
 * COPYRIGHT:  2007 Christopher Smith
 * LICENCE:    GPL v2: See COPYING in the top level directory
 * PROJECT:    regfs
 * FILE:       dirinf.c
 * PURPOSE:    Registry filesystem directory query
 */

#include "regfs.h"

VOID FORCEINLINE NTAPI KeyInfoToCommonDirectoryHeader(ULONG HeaderSize, PFILE_DIRECTORY_INFORMATION FileInfo, PKEY_BASIC_INFORMATION KeyInfo)
{
	FileInfo->CreationTime = KeyInfo->LastWriteTime;
	FileInfo->LastAccessTime = KeyInfo->LastWriteTime;
	FileInfo->LastWriteTime = KeyInfo->LastWriteTime;
	FileInfo->ChangeTime = KeyInfo->LastWriteTime;
	FileInfo->FileAttributes = FILE_ATTRIBUTE_DIRECTORY;
	//FileInfo->FileNameLength = KeyInfo->NameLength; //set during copy
}

VOID FORCEINLINE NTAPI KeyValueInfoToCommonDirectoryHeader(ULONG HeaderSize, PFILE_DIRECTORY_INFORMATION FileInfo, PKEY_FULL_INFORMATION KeyInfo, PKEY_VALUE_FULL_INFORMATION ValueInfo)
{
	FileInfo->CreationTime = KeyInfo->LastWriteTime;
	FileInfo->LastAccessTime = KeyInfo->LastWriteTime;
	FileInfo->LastWriteTime = KeyInfo->LastWriteTime;
	FileInfo->ChangeTime = KeyInfo->LastWriteTime;
	FileInfo->EndOfFile.QuadPart = ValueInfo->DataLength;
	FileInfo->AllocationSize.QuadPart = ValueInfo->DataLength;
	FileInfo->FileAttributes = FILE_ATTRIBUTE_NORMAL;
	//FileInfo->FileNameLength = ValueInfo->NameLength; //set during copy
}

BOOLEAN FORCEINLINE NTAPI DoesNameMatchQueryConstraint(UNICODE_STRING Constraint, PWCHAR NameBuffer, ULONG NameLength, BOOLEAN CaseInSensitive)
{
	UNICODE_STRING Name;

	Name.Buffer = NameBuffer;
	Name.MaximumLength = Name.Length = (USHORT)NameLength;

	DPRINT_V("Comparing %wZ and %wZ\n", &Constraint, &Name);

	if(!Constraint.Buffer)
		return TRUE;
	
	return FsRtlIsNameInExpression(&Constraint, &Name, CaseInSensitive, NULL);
}

typedef union _FILE_INFO_BUFFER
{
	PCHAR Addr;
	PULONG NextBlockOffs;
	PFILE_BOTH_DIR_INFORMATION BothInfo;
	PFILE_DIRECTORY_INFORMATION DirInfo;
	PFILE_FULL_DIR_INFORMATION FullInfo;
	PFILE_NAMES_INFORMATION NamesInfo;
}FILE_INFO_BUFFER, *PFILE_INFO_BUFFER;

NTSTATUS NTAPI RegfsQueryDirectory(IN PDEVICE_OBJECT DeviceObject, IN PIRP Irp, IN OUT PVOID IoBuffer, IN ULONG IoBufferLength, IN PSECURITY_SUBJECT_CONTEXT SecurityContext OPTIONAL)
{
	PIO_STACK_LOCATION IrpSp;
	PFILE_OBJECT FileObject;
	PREGFS_CCB Ccb;
	PREGFS_DFCB Dfcb;
	HANDLE KeyHandle;
	KEY_QUERY_BUFFER QB;
	KEY_FULL_INFORMATION KeyInfo;
	ULONG ReqQBSize;
	NTSTATUS Status;
	PSTR CompleteMsg;
	FILE_INFO_BUFFER FB;
	ULONG UnusedBuffer;
	ULONG NextSize;
	ULONG HeaderSize;
	ULONG NextVariableSize;
	ULONG SpecifiedIndex;
	PULONG LastOffset;
	BOOLEAN FullOnly = FALSE;
	ULONG_PTR MaxBufferAddr;

	FB.Addr = IoBuffer;
	MaxBufferAddr = (ULONG_PTR)IoBuffer + IoBufferLength;

	IrpSp = (PIO_STACK_LOCATION)IoGetCurrentIrpStackLocation(Irp);
	FileObject = IrpSp->FileObject;
	Ccb = (PREGFS_CCB)FileObject->FsContext2;

	if(!Ccb)
		return RegfsCompleteStatus(Irp, FALSE, STATUS_INVALID_DEVICE_REQUEST, "No directory listing on volume file.");

	Dfcb = Ccb->Dfcb;

	if(!(Dfcb->Flags & REGFS_DFCB_DIRECTORY))
		return RegfsCompleteStatus(Irp, FALSE, STATUS_INVALID_DEVICE_REQUEST, "No directory listing on a non-directory.");

	KeyHandle = RegfsGetDfcbKeyHandle(Dfcb);

	DPRINT("RegfsQueryDirectory DeviceObject:%p, Irp:%p InfoClass:%x FileName:%wZ\n", DeviceObject, Irp, HDRHELP_EXT_IRPSP(IrpSp)->Parameters.QueryDirectory.FileInformationClass, &Dfcb->FileName);

	Status = ZwQueryKey(KeyHandle, KeyFullInformation, &KeyInfo, sizeof KeyInfo, PAGED_BIT_BUCKET);

	if(!NT_SUCCESS(Status) && Status != STATUS_BUFFER_OVERFLOW)
		return RegfsCompleteStatus(Irp, FALSE, Status, "Error querying key.");

	AcquireSimpleMutex(Ccb->Lock);

	if(IrpSp->Flags & SL_INDEX_SPECIFIED)
	{
		Ccb->EnumIndex = 0;
		Ccb->TotalEnumIndex = 0;
		Ccb->EnumMode = EnumDotsMode;

		SpecifiedIndex = HDRHELP_EXT_IRPSP(IrpSp)->Parameters.QueryDirectory.FileIndex;
	}
	else
		SpecifiedIndex = 0;

	if(IrpSp->Flags & SL_RESTART_SCAN)
	{
		DPRINT_V("Restarting scan.\n");

		Ccb->EnumIndex = 0;
		Ccb->TotalEnumIndex = 0;
		Ccb->EnumMode = EnumDotsMode;
		Ccb->Flags &= ~REGFS_CCB_FILES_ENUMERATED;
	}

	//prepare QueryString
	{
		PUNICODE_STRING SrcQueryString = (PUNICODE_STRING)HDRHELP_EXT_IRPSP(IrpSp)->Parameters.QueryDirectory.FileName;

		if(SrcQueryString)
		{
			DPRINT_V("Replacing query string \"%wZ\" at %p.\n", &Ccb->QueryString, Ccb->QueryString.Buffer);

			if(Ccb->QueryString.Buffer)
				ExFreePoolWithTag(Ccb->QueryString.Buffer, TAG_CCB_QUERY_STRING);

			if(SrcQueryString->Length > 0 && !(SrcQueryString->Length == sizeof(WCHAR) && SrcQueryString->Buffer[0] == L'*'))
			{
				if(IrpSp->Flags & SL_CASE_SENSITIVE)
				{
					Ccb->QueryString.Buffer = ExAllocatePoolWithTag(PagedPool, SrcQueryString->Length, TAG_CCB_QUERY_STRING);
					Ccb->QueryString.MaximumLength = Ccb->QueryString.Length = SrcQueryString->Length;

					if(!Ccb->QueryString.Buffer)
					{
						ReleaseSimpleMutex(Ccb->Lock);
						return RegfsCompleteStatus(Irp, FALSE, STATUS_NO_MEMORY, "Error: Out of memory!");
					}
					RtlCopyUnicodeString(&Ccb->QueryString, SrcQueryString);
				}
				else //if case insensitive, upcase QueryString
				{
					Status = RtlUpcaseUnicodeString(&Ccb->QueryString, SrcQueryString, TRUE);

					if(!NT_SUCCESS(Status))
					{
						ExFreePoolWithTag(Ccb->QueryString.Buffer, TAG_CCB_QUERY_STRING);
						Ccb->QueryString.Buffer = NULL;
						RegfsCompleteStatus(Irp, FALSE, Status, "Error creating uppercase query string.");
					}
				}
				DPRINT("New query string: \"%wZ\" at %p\n", &Ccb->QueryString, Ccb->QueryString.Buffer);
			}
			else
			{
				DPRINT_V("New query string is null.\n");
				Ccb->QueryString.Buffer = NULL;
			}

			Ccb->Flags &= ~REGFS_CCB_FILES_ENUMERATED;
		}
	}

	switch(HDRHELP_EXT_IRPSP(IrpSp)->Parameters.QueryDirectory.FileInformationClass)
	{
	case FileBothDirectoryInformation:
		HeaderSize = SIZEOF_FILE_BOTH_DIR_INFORMATION_HEADER;
		break;
	case FileDirectoryInformation:
		HeaderSize = SIZEOF_FILE_DIRECTORY_INFORMATION_HEADER;
		break;
	case FileFullDirectoryInformation:
		HeaderSize = SIZEOF_FILE_FULL_DIR_INFORMATION_HEADER;
		break;
	case FileNamesInformation:
		HeaderSize = SIZEOF_FILE_NAMES_INFORMATION_HEADER;
		break;
	default:
		ReleaseSimpleMutex(Ccb->Lock);
		return RegfsCompleteStatus(Irp, FALSE, STATUS_INVALID_INFO_CLASS, "Error: Unsupported information class.");
	}

	UnusedBuffer = IoBufferLength;
	
	QB.Length = 0;
	QB.Addr = NULL;
	ReqQBSize = REGFS_DEFAULT_QUERY_BUFFER_SIZE;

	LastOffset = NULL;
	CompleteMsg = NULL; //to be printed after leave
	Status = STATUS_SUCCESS;

	try
	{
		while(TRUE)
		{
			Status = RegfsExpandBufferIfNecessary(&QB.Addr, &QB.Length, ReqQBSize, TAG_QUERY_BUFFER);
			if(!NT_SUCCESS(Status))
				TRY_LEAVE(NOTHING)

			//what are we currently enumerating?
			switch(Ccb->EnumMode)
			{
			case EnumDotsMode:
				Status = ZwQueryKey(KeyHandle, KeyBasicInformation, QB.KeyInfo, QB.Length, &ReqQBSize);

				switch(Ccb->EnumIndex)
				{
				case 0: //one dot
					QB.KeyInfo->Name[0] = L'.';
					QB.KeyInfo->NameLength = 1 * sizeof(WCHAR);
					break;
				case 1: //two dots
					QB.KeyInfo->Name[0] = L'.';
					QB.KeyInfo->Name[1] = L'.';
					QB.KeyInfo->NameLength = 2 * sizeof(WCHAR);
					break;
				default:
					Ccb->EnumMode = EnumKeysMode;
					Ccb->EnumIndex = 0;
					DPRINT_V("No more dot entries: switch to keys.\n");
					continue;
				}

				if(SpecifiedIndex > Ccb->TotalEnumIndex)
				{
					Ccb->TotalEnumIndex++;
					Ccb->EnumIndex++;
					continue;
				}

				goto CommonKeyProc;

			case EnumKeysMode:
				Status = ZwEnumerateKey(KeyHandle, Ccb->EnumIndex, KeyBasicInformation, QB.KeyInfo, QB.Length, &ReqQBSize);
				
				if(ReqQBSize > QB.Length)
					continue; //go to start to realloc

				if(Status == STATUS_NO_MORE_ENTRIES)
				{
					//switch to enumerating values
					Ccb->EnumMode = EnumValuesMode;
					Ccb->EnumIndex = 0;
					DPRINT_V("No more key entries: switch to values.\n");
					continue;
				} else if(!NT_SUCCESS(Status))
					TRY_LEAVE(Status = RegfsKeyStatusToFileStatus(Status))

				if(SpecifiedIndex > Ccb->TotalEnumIndex)
				{
					Ccb->TotalEnumIndex++;
					Ccb->EnumIndex++;
					continue;
				}

CommonKeyProc:;
				if(!DoesNameMatchQueryConstraint(Ccb->QueryString, QB.KeyInfo->Name, QB.KeyInfo->NameLength, !(IrpSp->Flags & SL_CASE_SENSITIVE)))
				{
					DPRINT_V("Didn't match contstraint (Case opt %x) %wZ.\n", !(IrpSp->Flags & SL_CASE_SENSITIVE), &Ccb->QueryString);
					Ccb->TotalEnumIndex++;
					Ccb->EnumIndex++;
					continue;
				}

				NextSize = QB.KeyInfo->NameLength + HeaderSize;

				DPRINT_V("NextSize %X = %X + %X\n", NextSize, QB.KeyInfo->NameLength, HeaderSize);

				if(NextSize > UnusedBuffer)
				{
					if(FullOnly) //no room for more than 1, so all done
					{
						DPRINT_V("Full only, no more room: needed %X have %X", NextSize, UnusedBuffer);
						TRY_LEAVE(Status = STATUS_SUCCESS)
					}
					else //the IO Manager guarntees there's enough room for the header, and copyname is safe: we can finish this iteration
					{
						DPRINT_V("Not enough room, partial return: needed %X have %X\n", NextSize, UnusedBuffer);
						Status = STATUS_BUFFER_OVERFLOW;
					}
				}

				UnusedBuffer -= HeaderSize;
				RtlZeroMemory(FB.Addr, HeaderSize);

				switch(HDRHELP_EXT_IRPSP(IrpSp)->Parameters.QueryDirectory.FileInformationClass)
				{
				case FileBothDirectoryInformation:
					KeyInfoToCommonDirectoryHeader(HeaderSize, FB.DirInfo, QB.KeyInfo);
					NextVariableSize = 
					FB.BothInfo->FileNameLength =
						RegfsCopyData(FB.BothInfo->FileName, QB.KeyInfo->Name, QB.KeyInfo->NameLength, UnusedBuffer, MaxBufferAddr);
					RegfsQueryKeyEaSetLength(KeyInfo.MaxClassLen, &FB.BothInfo->EaSize);
					break;
				case FileDirectoryInformation:
					KeyInfoToCommonDirectoryHeader(HeaderSize, FB.DirInfo, QB.KeyInfo);
					NextVariableSize = 
					FB.DirInfo->FileNameLength =
						RegfsCopyData(FB.DirInfo->FileName, QB.KeyInfo->Name, QB.KeyInfo->NameLength, UnusedBuffer, MaxBufferAddr);
					break;
				case FileFullDirectoryInformation:
					KeyInfoToCommonDirectoryHeader(HeaderSize, FB.DirInfo, QB.KeyInfo);
					NextVariableSize = 
					FB.FullInfo->FileNameLength =
						RegfsCopyData(FB.FullInfo->FileName, QB.KeyInfo->Name, QB.KeyInfo->NameLength, UnusedBuffer, MaxBufferAddr);
					RegfsQueryKeyEaSetLength(KeyInfo.MaxClassLen, &FB.FullInfo->EaSize);
					break;
				case FileNamesInformation:
					FB.NamesInfo->FileNameLength = QB.KeyInfo->NameLength;
					FB.NamesInfo->FileIndex = Ccb->TotalEnumIndex;
					NextVariableSize = 
					FB.NamesInfo->FileNameLength =
						RegfsCopyData(FB.NamesInfo->FileName, QB.KeyInfo->Name, QB.KeyInfo->NameLength, UnusedBuffer, MaxBufferAddr);
					break;
				}
				break;
			
			case EnumValuesMode:
				Status = ZwEnumerateValueKey(KeyHandle, Ccb->EnumIndex, KeyValueFullInformation, QB.KeyValueInfo, QB.Length, &ReqQBSize);

				if(ReqQBSize > QB.Length)
					continue; //go to start to realloc

				if(Status == STATUS_NO_MORE_ENTRIES)
				{
					Ccb->EnumMode = EnumEndMode;
					DPRINT_V("Continuing to end mode.\n");
					continue;
				} else if(!NT_SUCCESS(Status))
					TRY_LEAVE(Status = RegfsKeyStatusToFileStatus(Status))

				if(SpecifiedIndex > Ccb->TotalEnumIndex)
				{
					Ccb->TotalEnumIndex++;
					Ccb->EnumIndex++;
					continue;
				}

				if(!DoesNameMatchQueryConstraint(Ccb->QueryString, QB.KeyValueInfo->Name, QB.KeyValueInfo->NameLength, !(IrpSp->Flags & SL_CASE_SENSITIVE)))
				{
					DPRINT_V("Didn't match contstraint (Case opt %x) %wZ.\n", !(IrpSp->Flags & SL_CASE_SENSITIVE), &Ccb->QueryString);
					Ccb->EnumIndex++;
					Ccb->TotalEnumIndex++;
					continue;
				}

				//special case for the unnamed value
				if(QB.KeyValueInfo->NameLength == 0)
				{
					QB.KeyValueInfo->NameLength = REGFS_DEFAULT_VALUE_ALIAS.Length;
					RtlCopyMemory(QB.KeyValueInfo->Name, REGFS_DEFAULT_VALUE_ALIAS.Buffer, REGFS_DEFAULT_VALUE_ALIAS.Length);
				}

				NextSize = QB.KeyValueInfo->NameLength + HeaderSize;

				DPRINT_V("NextSize %X = %X + %X\n", NextSize, QB.KeyValueInfo->NameLength, HeaderSize);

				if(NextSize > UnusedBuffer)
				{
					if(FullOnly) //no room for more than 1, so all done
					{
						DPRINT_V("Full only, no more room: needed %X have %X", NextSize, UnusedBuffer);
						TRY_LEAVE(Status = STATUS_SUCCESS)
					}
					else //the IO Manager guarntees there's enough room for the header, and copyname is safe: we can do just this first entry for this call
					{
						DPRINT_V("Not enough room, partial return: needed %X have %X\n", NextSize, UnusedBuffer);
						Status = STATUS_BUFFER_OVERFLOW;
					}
				}

				UnusedBuffer -= HeaderSize;
				RtlZeroMemory(FB.Addr, HeaderSize);

				switch(HDRHELP_EXT_IRPSP(IrpSp)->Parameters.QueryDirectory.FileInformationClass)
				{
				case FileBothDirectoryInformation:
					KeyValueInfoToCommonDirectoryHeader(HeaderSize, FB.DirInfo, &KeyInfo, QB.KeyValueInfo);
					NextVariableSize = 
					FB.BothInfo->FileNameLength =
						RegfsCopyData(FB.BothInfo->FileName, QB.KeyValueInfo->Name, QB.KeyValueInfo->NameLength, UnusedBuffer, MaxBufferAddr);
					RegfsQueryEaSetLength(KeyHandle, FALSE, &FB.BothInfo->EaSize);
					break;
				case FileDirectoryInformation:
					KeyValueInfoToCommonDirectoryHeader(HeaderSize, FB.DirInfo, &KeyInfo, QB.KeyValueInfo);
					NextVariableSize = 
					FB.DirInfo->FileNameLength =
						RegfsCopyData(FB.DirInfo->FileName, QB.KeyValueInfo->Name, QB.KeyValueInfo->NameLength, UnusedBuffer, MaxBufferAddr);
					break;
				case FileFullDirectoryInformation:
					KeyValueInfoToCommonDirectoryHeader(HeaderSize, FB.DirInfo, &KeyInfo, QB.KeyValueInfo);
					NextVariableSize = 
					FB.FullInfo->FileNameLength =
						RegfsCopyData(FB.FullInfo->FileName, QB.KeyValueInfo->Name, QB.KeyValueInfo->NameLength, UnusedBuffer, MaxBufferAddr);
					RegfsQueryEaSetLength(KeyHandle, FALSE, &FB.FullInfo->EaSize);
					break;
				case FileNamesInformation:
					FB.NamesInfo->FileNameLength = QB.KeyValueInfo->NameLength;
					FB.NamesInfo->FileIndex = Ccb->EnumIndex;
					NextVariableSize = 
					FB.NamesInfo->FileNameLength =
						RegfsCopyData(FB.NamesInfo->FileName, QB.KeyValueInfo->Name, QB.KeyValueInfo->NameLength, UnusedBuffer, MaxBufferAddr);
					break;
				}
				break;

			case EnumEndMode:
				if(Ccb->Flags & REGFS_CCB_FILES_ENUMERATED) //was anything returned this enumeration?
				{
					if(FB.Addr == IoBuffer) //was nothing returned this call?
						TRY_LEAVE(Status = STATUS_NO_MORE_FILES; CompleteMsg = "No more files.")
					else
						TRY_LEAVE(Status = STATUS_SUCCESS; CompleteMsg = "End of enumeration.")
				}
				else
					TRY_LEAVE(Status = STATUS_NO_SUCH_FILE; CompleteMsg = "No such file.")
				break;
			}

			if(!NT_SUCCESS(Status))
				TRY_LEAVE(Status = RegfsKeyStatusToFileStatus(Status))

			Ccb->Flags |= REGFS_CCB_FILES_ENUMERATED;

			Ccb->EnumIndex++;
			Ccb->TotalEnumIndex++;

			if(IrpSp->Flags & SL_RETURN_SINGLE_ENTRY)
			{
				UnusedBuffer -= NextVariableSize; //header was already accounted
				TRY_LEAVE(Status = STATUS_SUCCESS; CompleteMsg = "Returned a single entry.")
			}

			LastOffset = FB.NextBlockOffs; //the first member is always a pointer to the next block

			NextSize = AlignUpTo(HeaderSize + NextVariableSize, 3); //8 byte

			FB.Addr += NextSize;
			UnusedBuffer -= NextSize - HeaderSize; //header was already accounted, need pad + variable

			//save expected next buffer offset
			if(LastOffset)
				*LastOffset = NextSize;

			DPRINT_V("Last offset is %p\n", *LastOffset);

			FullOnly = TRUE; //only return full entries for remaining items during this call
		}
	}
	except(EXCEPTION_EXECUTE_HANDLER)
	{
		Status = GetExceptionCode();
	}

	if(QB.Addr)
		ExFreePoolWithTag(QB.Addr, TAG_QUERY_BUFFER);
	ReleaseSimpleMutex(Ccb->Lock);

	//transfer is the buffer length, except what we didn't use
	Irp->IoStatus.Information = IoBufferLength - UnusedBuffer;
	
	if(LastOffset)
		try
		{
			*LastOffset = 0; //it was the last
		}
		except(EXCEPTION_EXECUTE_HANDLER)
		{
			Status = GetExceptionCode();
		}

	return RegfsCompleteStatus(Irp, TRUE, Status, CompleteMsg);
}