/*
 * COPYRIGHT:  2007 Christopher Smith
 * LICENCE:    GPL v2: See COPYING in the top level directory
 * PROJECT:    regfs
 * FILE:       dispatch.c
 * PURPOSE:    Registry filesystem dispatch preparation
 */

#include "regfs.h"

typedef NTSTATUS (*PREGFS_IRP_DISPATCH_PREP)(PDEVICE_OBJECT DeviceObject, PIRP Irp, PREGFS_DCB Dcb, BOOLEAN RequiresSecurityContext, PREGFS_IRP_DISPATCH Next);

PREGFS_IRP_DISPATCH_PREP DispatchPrepRoutine;

PREGFS_IRP_DISPATCH DispatchRoutines[IRP_MJ_MAXIMUM_FUNCTION];

NTSTATUS NTAPI RegfsDispatch(PDEVICE_OBJECT DeviceObject, PIRP Irp)
{
	PREGFS_CCB Context;
	PREGFS_DCB Dcb;
	PIO_STACK_LOCATION IrpSp;
	BOOLEAN RequiresSecurityContext;

	IrpSp = IoGetCurrentIrpStackLocation(Irp);
	Context = IrpSp->FileObject->FsContext2;

	if(Context)
	{
		if(Context->Dfcb->Flags & REGFS_DFCB_DIRECTORY)
			Dcb = &Context->Dfcb->Dcb;
		else
			Dcb = &Context->Dfcb->Fcb.Directory->Dcb;
	}
	else
		Dcb = NULL;

	switch(IrpSp->MajorFunction)
	{
	case IRP_MJ_SET_INFORMATION: //for move
		if(IrpSp->Parameters.SetFile.FileInformationClass == FileRenameInformation)
			RequiresSecurityContext = TRUE;
		else
			RequiresSecurityContext = FALSE;
		break;

	case IRP_MJ_FILE_SYSTEM_CONTROL:
		RequiresSecurityContext = TRUE;
		break;

	default://note that the IO manager already provides the context for create
		RequiresSecurityContext = FALSE;
	}

	if(DispatchRoutines[IrpSp->MajorFunction])
		return DispatchPrepRoutine(DeviceObject, Irp, Dcb, RequiresSecurityContext, DispatchRoutines[IrpSp->MajorFunction]);
	else
	{
		DPRINT("Invalid IRP function %x. ", IrpSp->MajorFunction);
		Irp->IoStatus.Status = STATUS_INVALID_DEVICE_REQUEST;
		IoCompleteRequest(Irp, IO_NO_INCREMENT);
		return STATUS_INVALID_DEVICE_REQUEST;
	}
}

NTSTATUS NTAPI RegfsDispatchMethodLocal(PDEVICE_OBJECT DeviceObject, PIRP Irp, PREGFS_DCB Dcb, BOOLEAN RequiresSecurityContext, PREGFS_IRP_DISPATCH Next)
{
	NTSTATUS Status = STATUS_SUCCESS;
	PVOID Buffer;
	ULONG Length;
	SECURITY_SUBJECT_CONTEXT SecurityContext;
	LOCK_OPERATION Operation;

	FsRtlEnterFileSystem();

	Status = RegfsStartLocalBuffer(Irp, &Buffer, &Length, &Operation);

	if(!NT_SUCCESS(Status))
	{
		DPRINT("Error setting up buffer.");
		Irp->IoStatus.Status = Status;
		IoCompleteRequest(Irp, IO_NO_INCREMENT);
		return Status;
	}

	if(RequiresSecurityContext)
		SeCaptureSubjectContext(&SecurityContext);

	Status = Next(DeviceObject, Irp, Buffer, Length, RequiresSecurityContext ? &SecurityContext : NULL);

	if(RequiresSecurityContext)
		SeReleaseSubjectContext(&SecurityContext);

	if(Irp->Tail.Overlay.DriverContext[0]) //DriverContext[0] is message
		DPRINT((PCH)Irp->Tail.Overlay.DriverContext[0]);
	DPRINT(" Completed. Status:%x\n", Status);

	//RegfsEndBuffer(Irp, Buffer);

	Irp->IoStatus.Status = Status;

	IoCompleteRequest(Irp, Irp->Tail.Overlay.DriverContext[1] ? IO_DISK_INCREMENT : IO_NO_INCREMENT); //DriverContext[1] is DidIo

	FsRtlExitFileSystem();

	return Status;
}

NTSTATUS NTAPI PrepareWorkItem(OUT PREGFS_WORK_ITEM Wi, PDEVICE_OBJECT DeviceObject, PIRP Irp, PREGFS_DCB Dcb, BOOLEAN RequiresSecurityContext, PREGFS_IRP_DISPATCH Next)
{
	PVOID Buffer;
	ULONG Length;
	LOCK_OPERATION Operation;
	NTSTATUS Status;

	Status = RegfsStartRemoteBuffer(Irp, &Buffer, &Length, &Operation);

	if(!NT_SUCCESS(Status))
	{
		DPRINT("Error setting up buffer.\n");
		return Status;
	}

	if(RequiresSecurityContext)
		SeCaptureSubjectContext(&Wi->DIrp.SecurityContext);

	Wi->Type = WtDispatchIrp;
	Wi->DIrp.Proc = Next;
	Wi->DIrp.DeviceObject = DeviceObject;
	Wi->DIrp.Irp = Irp;
	Wi->DIrp.IoBuffer = Buffer;
	Wi->DIrp.IoBufferLength = Length;
	Wi->DIrp.HasSecurityContext = RequiresSecurityContext;

	return STATUS_SUCCESS;
}

NTSTATUS NTAPI RegfsDispatchMethodWorkThread(PDEVICE_OBJECT DeviceObject, PIRP Irp, PREGFS_DCB Dcb, BOOLEAN RequiresSecurityContext, PREGFS_IRP_DISPATCH Next)
{
	NTSTATUS Status;
	REGFS_WORK_ITEM Wi;

	FsRtlEnterFileSystem();

	Status = PrepareWorkItem(&Wi, DeviceObject, Irp, Dcb, RequiresSecurityContext, Next);

	if(!NT_SUCCESS(Status))
	{
		DPRINT("Error preparing work item.\n");
		Irp->IoStatus.Status = Status;
		IoCompleteRequest(Irp, IO_NO_INCREMENT);
		return Status;
	}

	IoMarkIrpPending(Irp);

	Status = RegfsQueueWorkItem(&Wi);

	if(!NT_SUCCESS(Status))
	{
		if(RequiresSecurityContext)
			SeReleaseSubjectContext(&Wi.DIrp.SecurityContext);

		DPRINT("Error queueing work item.\n");
		Irp->IoStatus.Status = Status;
		IoCompleteRequest(Irp, IO_NO_INCREMENT);
		return Status;
	}

	FsRtlExitFileSystem();

	return STATUS_PENDING;
}

VOID NTAPI RegfsDispatchMethodWorkThreadSystemContext(PREGFS_WORK_ITEM Wi)
{
	NTSTATUS Status;
	PIRP Irp = Wi->DIrp.Irp;

	if(Irp->Cancel) //probably unreachable
	{
		Irp->IoStatus.Status = STATUS_CANCELLED;
		IoCompleteRequest(Irp, IO_NO_INCREMENT);
	}

	Status = Wi->DIrp.Proc(Wi->DIrp.DeviceObject, Irp, Wi->DIrp.IoBuffer, Wi->DIrp.IoBufferLength, &Wi->DIrp.SecurityContext);

	if(Wi->DIrp.HasSecurityContext)
		SeReleaseSubjectContext(&Wi->DIrp.SecurityContext);

	if(Irp->Tail.Overlay.DriverContext[0]) //DriverContext[0] is message
		DPRINT((PCH)Irp->Tail.Overlay.DriverContext[0]);
	DPRINT(" Completed. Status:%x\n", Status);

	RegfsEndBuffer(Irp, Wi->DIrp.IoBuffer);

	Irp->IoStatus.Status = Status;

	IoCompleteRequest(Irp, Irp->Tail.Overlay.DriverContext[1] ? IO_DISK_INCREMENT : IO_NO_INCREMENT); //DriverContext[1] is DidIo
}

NTSTATUS NTAPI RegfsDispatchMethodExWorker(PDEVICE_OBJECT DeviceObject, PIRP Irp, PREGFS_DCB Dcb, BOOLEAN RequiresSecurityContext, PREGFS_IRP_DISPATCH Next)
{
	NTSTATUS Status;
	PVOID Buffer;
	ULONG Length;
	LOCK_OPERATION Operation;
	REGFS_WORK_ITEM Wi;

	FsRtlEnterFileSystem();

	Status = PrepareWorkItem(&Wi, DeviceObject, Irp, Dcb, RequiresSecurityContext, Next);

	if(!NT_SUCCESS(Status))
	{
		DPRINT("Error preparing work item.\n");
		Irp->IoStatus.Status = Status;
		IoCompleteRequest(Irp, IO_NO_INCREMENT);
		return Status;
	}

	IoMarkIrpPending(Irp);

	Status = RegfsQueueExWorkItem(&Wi);

	if(!NT_SUCCESS(Status))
	{
		if(RequiresSecurityContext)
			SeReleaseSubjectContext(&Wi.DIrp.SecurityContext);

		DPRINT("Error queueing work item.\n");
		Irp->IoStatus.Status = Status;
		IoCompleteRequest(Irp, IO_NO_INCREMENT);
		return Status;
	}

	FsRtlExitFileSystem();

	return STATUS_PENDING;
}

#pragma code_seg(push, INIT)

PDRIVER_DISPATCH NTAPI RegfsInitDispatch(REGFS_DISPATCH_METHOD Method, PNTSTATUS Status)
{
	switch(Method)
	{
	case DmObjKernelHandle:
		DispatchPrepRoutine = &RegfsDispatchMethodLocal;
		DefaultObjectAttributes |= OBJ_KERNEL_HANDLE;
		*Status = STATUS_SUCCESS;
		break;

	case DmSystemThread:
		*Status = RegfsInitWorkThreads(1, 1);
		DispatchPrepRoutine = &RegfsDispatchMethodWorkThread;
		break;

	case DmExWorkItem:
		*Status = RegfsInitExWorkItems(2);
		DispatchPrepRoutine = &RegfsDispatchMethodExWorker;
		break;

	default:
		*Status = STATUS_INVALID_PARAMETER;
		DispatchPrepRoutine = NULL;
		break;
	}

	if(!NT_SUCCESS(*Status))
		return NULL;

	RtlZeroMemory(DispatchRoutines, IRP_MJ_MAXIMUM_FUNCTION * sizeof(PREGFS_IRP_DISPATCH));

	DispatchRoutines[IRP_MJ_CREATE] = &RegfsCreate;
	DispatchRoutines[IRP_MJ_CLOSE] = &RegfsClose;
	DispatchRoutines[IRP_MJ_READ] = &RegfsReadWrite;
	DispatchRoutines[IRP_MJ_WRITE] = &RegfsReadWrite;
	DispatchRoutines[IRP_MJ_QUERY_INFORMATION] = &RegfsQueryInformation;
	DispatchRoutines[IRP_MJ_SET_INFORMATION] = &RegfsSetInformation;
	DispatchRoutines[IRP_MJ_QUERY_VOLUME_INFORMATION] = &RegfsQueryVolumeInformation;
	DispatchRoutines[IRP_MJ_QUERY_EA] = &RegfsQueryEa;
	DispatchRoutines[IRP_MJ_SET_EA] = &RegfsSetEa;
	DispatchRoutines[IRP_MJ_CLEANUP] = &RegfsCleanup;
	DispatchRoutines[IRP_MJ_FLUSH_BUFFERS] = &RegfsFlush;
	DispatchRoutines[IRP_MJ_DIRECTORY_CONTROL] = &RegfsDirectoryControl;
	DispatchRoutines[IRP_MJ_FILE_SYSTEM_CONTROL] = &RegfsFsControl;
	DispatchRoutines[IRP_MJ_QUERY_SECURITY] = &RegfsQuerySecurity;
	DispatchRoutines[IRP_MJ_SET_SECURITY] = &RegfsSetSecurity;

	return &RegfsDispatch;
}

#pragma code_seg(pop, INIT)