/*
 * COPYRIGHT:  2007 Christopher Smith
 * LICENCE:    GPL v2: See COPYING in the top level directory
 * PROJECT:    regfs
 * FILE:       ea.c
 * PURPOSE:    Registry filesystem basic and extended attributes
 */

#include "regfs.h"

//There are 3 EAs for keys: titleindex, classname and volatile
//			2 EAs for values: titleindex, type
//The EA volaitle can also be set on creation of keys

DECLARE_CONST_STRING(EA_TITLE_INDEX, "TitleIndex");
DECLARE_CONST_STRING(EA_CLASS_NAME, "ClassName");
DECLARE_CONST_STRING(EA_VOLATILE, "Volatile");
DECLARE_CONST_STRING(EA_TYPE, "Type");

NTSTATUS NTAPI RegfsQueryKeyEaSetLength(IN ULONG ClassLength, OUT PULONG Length)
{
	*Length = SIZEOF_FILE_FULL_EA_INFORMATION_HEADER + EA_TITLE_INDEX.MaximumLength + sizeof(ULONG); //there is always a title index
	*Length = AlignUpTo(*Length, 2);//align to 4 bytes
	*Length += SIZEOF_FILE_FULL_EA_INFORMATION_HEADER + EA_CLASS_NAME.MaximumLength + ClassLength;

	DPRINT_V("Ea length of %x from ClassLength %x\n", *Length, ClassLength);

	return STATUS_SUCCESS;
}

NTSTATUS NTAPI RegfsQueryEaSetLength(IN HANDLE KeyHandle, IN BOOLEAN IsDirectory, OUT PULONG Length)
{
	NTSTATUS Status = STATUS_SUCCESS;
	*Length = SIZEOF_FILE_FULL_EA_INFORMATION_HEADER + EA_TITLE_INDEX.MaximumLength + sizeof(ULONG); //there is always a title index
	*Length = AlignUpTo(*Length, 2);//align to 4 bytes

	if(IsDirectory)
	{
		KEY_NODE_INFORMATION Kni; 

		Status = ZwQueryKey(KeyHandle, KeyNodeInformation, &Kni, sizeof Kni, PAGED_BIT_BUCKET);

		if(!NT_SUCCESS(Status) && Status != STATUS_BUFFER_OVERFLOW)
			return Status;

		*Length += SIZEOF_FILE_FULL_EA_INFORMATION_HEADER + EA_CLASS_NAME.MaximumLength + Kni.ClassLength;
	}
	else
		*Length += SIZEOF_FILE_FULL_EA_INFORMATION_HEADER + EA_TYPE.MaximumLength + sizeof(ULONG);

	DPRINT_V("Ea length of %x\n", *Length);

	return Status;
}

NTSTATUS NTAPI FillInEaInfo(STRING const *Name, PVOID Data, USHORT DataLen, PFILE_FULL_EA_INFORMATION EaInfo, PULONG UnusedBuffer, ULONG_PTR MaxBufferAddr)
{
	NTSTATUS Status;

	DPRINT_V("FillInEaInfo Data:%p DataLen:%X EaInfo:%p UnusedBuffer:%X MaxBuffer:%p\n", Data, DataLen, EaInfo, *UnusedBuffer, MaxBufferAddr);

	if(*UnusedBuffer < SIZEOF_FILE_FULL_EA_INFORMATION_HEADER)
		return STATUS_BUFFER_TOO_SMALL;

	EaInfo->Flags = 0;
	EaInfo->EaNameLength = (UCHAR)Name->Length;
	EaInfo->EaValueLength = DataLen;
	EaInfo->NextEntryOffset = 0;

	*UnusedBuffer -= SIZEOF_FILE_FULL_EA_INFORMATION_HEADER;

	if(*UnusedBuffer < (ULONG)(Name->MaximumLength + DataLen))
		Status = STATUS_BUFFER_OVERFLOW;
	else
		Status = STATUS_SUCCESS;

	if(*UnusedBuffer)
		*UnusedBuffer -= RegfsCopyData(EaInfo->EaName, Name->Buffer, Name->MaximumLength, *UnusedBuffer, MaxBufferAddr);

	if(*UnusedBuffer)
		*UnusedBuffer -= EaInfo->EaValueLength = (USHORT)RegfsCopyData(EaInfo->EaName + Name->MaximumLength, Data, DataLen, *UnusedBuffer, MaxBufferAddr); //cast to USHORT safe: classname is the longest value and also a USHORT

	return Status;
}

NTSTATUS NTAPI RegfsQueryEa(IN PDEVICE_OBJECT DeviceObject, IN PIRP Irp, IN OUT PVOID IoBuffer, IN ULONG IoBufferLength, IN PSECURITY_SUBJECT_CONTEXT SecurityContext OPTIONAL)
{
	NTSTATUS Status;
	PIO_STACK_LOCATION IrpSp;
	HANDLE KeyHandle;
	PREGFS_CCB Ccb;
	PREGFS_DFCB Dfcb;
	PFILE_OBJECT FileObject;
	PFILE_GET_EA_INFORMATION GetEaItem;
	PFILE_FULL_EA_INFORMATION EaInfo;
	PULONG PreviousNextEntryOffset = NULL;
	ULONG UnusedBuffer;
	ULONG_PTR MaxBufferAddr;
	PSTR CompleteMsg;

	MaxBufferAddr = (ULONG_PTR)IoBuffer + IoBufferLength;

	IrpSp = (PIO_STACK_LOCATION)IoGetCurrentIrpStackLocation(Irp);
	FileObject = IrpSp->FileObject;
	
	Ccb = (PREGFS_CCB)FileObject->FsContext2;

	DPRINT("RegfsQueryEa DeviceObject:%p, Irp:%p FileName:%wZ\n", DeviceObject, Irp, &FileObject->FileName);

	if(!Ccb)
		return RegfsCompleteStatus(Irp, FALSE, STATUS_INVALID_DEVICE_REQUEST, "No EAs on volume file.");

	Dfcb = Ccb->Dfcb;

	KeyHandle = RegfsGetDfcbKeyHandle(Dfcb);

	AcquireSimpleMutex(Ccb->Lock);

	if(IrpSp->Flags & SL_RESTART_SCAN)
		Ccb->EaEnumIndex = 1; //EAs are 1-index based

	if(IrpSp->Flags & SL_INDEX_SPECIFIED)
	{
		if(HDRHELP_EXT_IRPSP(IrpSp)->Parameters.QueryEa.EaIndex > 2)
		{
			ReleaseSimpleMutex(Ccb->Lock);
			return RegfsCompleteStatus(Irp, FALSE, STATUS_NONEXISTENT_EA_ENTRY, "Invalid EA index.");
		}
		else
			Ccb->EaEnumIndex = HDRHELP_EXT_IRPSP(IrpSp)->Parameters.QueryEa.EaIndex;
	}

	GetEaItem = HDRHELP_EXT_IRPSP(IrpSp)->Parameters.QueryEa.EaList;

	UnusedBuffer = IoBufferLength;
	
	EaInfo = (PFILE_FULL_EA_INFORMATION)IoBuffer;
	CompleteMsg = NULL; //to be printed after leave
	Status = STATUS_SUCCESS;

	try
	{
		while(TRUE)
		{
			PFILE_FULL_EA_INFORMATION PreviousEaInfo = EaInfo;
			
			PreviousNextEntryOffset = &EaInfo->NextEntryOffset;

			(ULONG_PTR)EaInfo = (ULONG_PTR)IoBuffer + (IoBufferLength - UnusedBuffer); //position to next

			if(EaInfo != IoBuffer) //only need to re-align after having returned an entry
				UnusedBuffer -= AlignUpToAndZero(&EaInfo, (PVOID)MaxBufferAddr, 2); //align to 4 bytes

			if(PreviousNextEntryOffset)
					*PreviousNextEntryOffset = (ULONG_PTR)EaInfo - (ULONG_PTR)PreviousEaInfo; //set offset of this new entry (may be zeroed if it was the last)

			if(GetEaItem)
			{
				STRING SearchName = {GetEaItem->EaNameLength, GetEaItem->EaNameLength, GetEaItem->EaName};
				if(RtlEqualString(&EA_TITLE_INDEX, &SearchName, REGFS_ARE_EA_NAMES_CINS))
					Ccb->EaEnumIndex = 1; //1 is title index for both
				else if(RtlEqualString(&EA_CLASS_NAME, &SearchName, REGFS_ARE_EA_NAMES_CINS))
					Ccb->EaEnumIndex = 2; //2 is class name for keys
				//else if(RtlEqualString(&EA_VOLATILE, &SearchName, REGFS_ARE_EA_NAMES_CINS)) //volaitle can only be set on create
				else if(RtlEqualString(&EA_TYPE, &SearchName, REGFS_ARE_EA_NAMES_CINS))
					Ccb->EaEnumIndex = 2; //2 is type for values
				else
					TRY_LEAVE(Status = STATUS_NONEXISTENT_EA_ENTRY; CompleteMsg = "Non-existant EA queried.")
			}

			if(Ccb->EaEnumIndex > 2) //both keys and values have two EAs, numbered 1 and 2
			{
				if(IoBuffer == EaInfo) //if no items returned yet
					TRY_LEAVE(Status = STATUS_NO_MORE_EAS; CompleteMsg = "No more EAs.")
				else
					TRY_LEAVE(Status = STATUS_SUCCESS; CompleteMsg = "Success; all EAs returned.")
			}

			if(Dfcb->Flags & REGFS_DFCB_DIRECTORY)
			{				
				switch(Ccb->EaEnumIndex)
				{
				case 1: //TitleIndex
					{
						KEY_BASIC_INFORMATION KeyInfo;

						Status = ZwQueryKey(KeyHandle, KeyBasicInformation, &KeyInfo, sizeof KeyInfo, PAGED_BIT_BUCKET);

						if(!NT_SUCCESS(Status) && Status != STATUS_BUFFER_OVERFLOW)
							TRY_LEAVE(Status = RegfsKeyStatusToFileStatus(Status); CompleteMsg = "Error querying key info.")
						
						Status = FillInEaInfo(&EA_TITLE_INDEX, &KeyInfo.TitleIndex, sizeof KeyInfo.TitleIndex, EaInfo, &UnusedBuffer, MaxBufferAddr);
					}
					break;

				case 2: //ClassName
					{
						ULONG KeyInfoSize;
						PKEY_FULL_INFORMATION KeyInfo;

						Status = ZwQueryKey(KeyHandle, KeyFullInformation, NULL, 0, &KeyInfoSize);

						if(!NT_SUCCESS(Status) && !IsBufferSizeStatus(Status))
							TRY_LEAVE(Status = RegfsKeyStatusToFileStatus(Status); CompleteMsg = "Error getting key info size.")

						KeyInfo = ExAllocatePoolWithTag(PagedPool, KeyInfoSize, TAG_EA_KEY_INFO_CLASS);

						if(!KeyInfo)
							TRY_LEAVE(Status = STATUS_NO_MEMORY; CompleteMsg = "Out of memory!")

						Status = ZwQueryKey(KeyHandle, KeyFullInformation, KeyInfo, KeyInfoSize, PAGED_BIT_BUCKET);
						
						if(!NT_SUCCESS(Status) && Status != STATUS_BUFFER_OVERFLOW)
						{
							ExFreePoolWithTag(KeyInfo, TAG_EA_KEY_INFO_CLASS);
							TRY_LEAVE(Status = RegfsKeyStatusToFileStatus(Status); CompleteMsg = "Error querying key info.")
						}

						Status = FillInEaInfo(&EA_CLASS_NAME, KeyInfo->Class, (USHORT)KeyInfo->ClassLength, EaInfo, &UnusedBuffer, MaxBufferAddr);

						ExFreePoolWithTag(KeyInfo, TAG_EA_KEY_INFO_CLASS);
					}
					break;
				}
			}
			else
			{
				KEY_VALUE_BASIC_INFORMATION KeyValueInfo;

				Status = ZwQueryValueKey(KeyHandle, &Dfcb->ShortName, KeyValueBasicInformation, &KeyValueInfo, sizeof KeyValueInfo, PAGED_BIT_BUCKET);

				if(!NT_SUCCESS(Status) && Status != STATUS_BUFFER_OVERFLOW)
					TRY_LEAVE(Status = RegfsKeyStatusToFileStatus(Status); CompleteMsg = "Error querying key value info.")
				
				switch(Ccb->EaEnumIndex)
				{
				case 1: //TitleIndex
					Status = FillInEaInfo(&EA_TITLE_INDEX, &KeyValueInfo.TitleIndex, sizeof KeyValueInfo.TitleIndex, EaInfo, &UnusedBuffer, MaxBufferAddr);
					break;

				case 2: //Type
					Status = FillInEaInfo(&EA_TYPE, &KeyValueInfo.Type, sizeof KeyValueInfo.Type, EaInfo, &UnusedBuffer, MaxBufferAddr);
					break;
				}
			}

			if(Status == STATUS_BUFFER_TOO_SMALL || Status == STATUS_BUFFER_OVERFLOW && //if there wasn't room for another entry and
				EaInfo != IoBuffer) //we're not returning the first entry
					TRY_LEAVE(Status = STATUS_SUCCESS; CompleteMsg = "Success."); //it's ok we didn't need to return more than 1 per request to be a succes

			if(!NT_SUCCESS(Status))
				TRY_LEAVE(NOTHING)

			Ccb->EaEnumIndex++;
		}
	}
	except(EXCEPTION_EXECUTE_HANDLER)
	{
		Status = GetExceptionCode();
	}

	ReleaseSimpleMutex(Ccb->Lock);

	Irp->IoStatus.Information = IoBufferLength - UnusedBuffer;

	if(PreviousNextEntryOffset)
		try
		{
			*PreviousNextEntryOffset = 0; //it was the last
		}
		except(EXCEPTION_EXECUTE_HANDLER)
		{
			Status = GetExceptionCode();
		}
#if DBG
	{
		ULONG EaLen;
		ULONG ErrorOffset;
		NTSTATUS LocalStatus;

		RegfsQueryEaSetLength(KeyHandle, (Dfcb->Flags & REGFS_DFCB_DIRECTORY) != 0, &EaLen);

		LocalStatus = IoCheckEaBufferValidity((PFILE_FULL_EA_INFORMATION)IoBuffer, EaLen, &ErrorOffset);
		DPRINT_V("EA validity status:%X Offset:%X EaLen:%X BufLen:%X Returned:%X\n", LocalStatus, ErrorOffset, EaLen, IoBufferLength, Irp->IoStatus.Information);
		if(!NT_SUCCESS(LocalStatus))
			DPRINT("Bad EA output. Status:%X Offset:%X\n", LocalStatus, ErrorOffset);
	}
#endif

	return RegfsCompleteStatus(Irp, TRUE, Status, CompleteMsg);
}

NTSTATUS NTAPI RegfsStructureEaList(IN BOOLEAN IsDirectory, IN PFILE_FULL_EA_INFORMATION EaBuf, ULONG EaBufLen, IN OUT PREGFS_STRUCTURED_EA EaStruct)
{
	NTSTATUS Status;
	ULONG ErrorOffset;
	ULONG_PTR MinAddr, MaxAddr;

	if(!EaBuf || !EaBufLen)
		return STATUS_SUCCESS; //no EAs

	DPRINT("RegfsStructureEaList\n");

	//Status = IoCheckEaBufferValidity(EaBuf, EaBufLen, &ErrorOffset);

	//if(!NT_SUCCESS(Status))
	//{
	//	DPRINT("Invalid EA status %X at %X\n", Status, ErrorOffset);
	//	return Status;
	//}

	MinAddr = (ULONG_PTR)EaBuf;
	MaxAddr = MinAddr + EaBufLen;

	while(TRUE)
	{
		USHORT ValueLen; //cache to avoid user buffer race
		PVOID ValueData;
		ANSI_STRING EaName;

		if(!RangeInRange((ULONG_PTR)EaBuf, (ULONG_PTR)EaBuf + SIZEOF_FILE_FULL_EA_INFORMATION_HEADER, MinAddr, MaxAddr)) //recheck in case of change since IoCheckEa
			return STATUS_EA_LIST_INCONSISTENT;
		
		EaName.Buffer = EaBuf->EaName;
		EaName.Length = EaBuf->EaNameLength;
		EaName.MaximumLength = EaName.Length + 1; //with null
		ValueLen = EaBuf->EaValueLength;
		ValueData = EaBuf->EaName + EaName.MaximumLength;

		if(!RangeInRange((ULONG_PTR)ValueData, (ULONG_PTR)ValueData + ValueLen, MinAddr, MaxAddr)) //recheck in case of change
			return STATUS_EA_LIST_INCONSISTENT;

		DPRINT("Searching for %Z (%xb)", &EaName, EaName.Length);

		if(IsDirectory)
		{
			if(RtlEqualString(&EA_TITLE_INDEX, &EaName, REGFS_ARE_EA_NAMES_CINS))
			{
				if(ValueLen != sizeof EaStruct->TitleIndex)
					return STATUS_EA_CORRUPT_ERROR;

				RtlCopyMemory(&EaStruct->TitleIndex, ValueData, ValueLen); //use CopyMemory to avoid alignment issues

				DPRINT("Found title index %x\n", EaStruct->TitleIndex);
			}
			else if(RtlEqualString(&EA_CLASS_NAME, &EaName, REGFS_ARE_EA_NAMES_CINS))
			{
				EaStruct->Class.Buffer = (PWCHAR)ValueData;
				EaStruct->Class.Length = EaStruct->Class.MaximumLength = ValueLen;
				DPRINT("Found class name %x\n", EaStruct->TitleIndex);
			}
			else if(RtlEqualString(&EA_VOLATILE, &EaName, REGFS_ARE_EA_NAMES_CINS))
			{
				if(ValueLen != sizeof EaStruct->Volatile)
					return STATUS_EA_LIST_INCONSISTENT;
				
				RtlCopyMemory(&EaStruct->Volatile, ValueData, ValueLen); //use CopyMemory to avoid alignment issues
			}
			else
				return STATUS_NONEXISTENT_EA_ENTRY;
		}
		else //a non-directory (a key value)
		{
			if(RtlEqualString(&EA_TITLE_INDEX, &EaName, REGFS_ARE_EA_NAMES_CINS))
			{
				if(ValueLen != sizeof EaStruct->TitleIndex)
					return STATUS_EA_LIST_INCONSISTENT;
				
				RtlCopyMemory(&EaStruct->TitleIndex, ValueData, ValueLen); //use CopyMemory to avoid alignment issues

				DPRINT("Found value title index %x\n", EaStruct->TitleIndex);
			}
			else if(RtlEqualString(&EA_TYPE, &EaName, REGFS_ARE_EA_NAMES_CINS))
			{
				if(ValueLen != sizeof EaStruct->Type)
					return STATUS_EA_LIST_INCONSISTENT;
				
				RtlCopyMemory(&EaStruct->Type, ValueData, ValueLen); //use CopyMemory to avoid alignment issues

				DPRINT("Found value type %x\n", EaStruct->Type);
			}
			else
				return STATUS_NONEXISTENT_EA_ENTRY;
		}

		if(EaBuf->NextEntryOffset)
			EaBuf = (PFILE_FULL_EA_INFORMATION)((ULONG_PTR)EaBuf + EaBuf->NextEntryOffset);
		else
			break;
	}

	return STATUS_SUCCESS;
}

NTSTATUS NTAPI RegfsSetEa(IN PDEVICE_OBJECT DeviceObject, IN PIRP Irp, IN OUT PVOID IoBuffer, IN ULONG IoBufferLength, IN PSECURITY_SUBJECT_CONTEXT SecurityContext OPTIONAL)
{
	NTSTATUS Status;
	PIO_STACK_LOCATION IrpSp;
	PFILE_OBJECT FileObject;
	HANDLE KeyHandle;
	PREGFS_CCB Ccb;
	PREGFS_DFCB Dfcb;
	REGFS_STRUCTURED_EA EaStruct;
	PSTR CompleteMsg;

	IrpSp = (PIO_STACK_LOCATION)IoGetCurrentIrpStackLocation(Irp);
	FileObject = IrpSp->FileObject;

	DPRINT("RegfsSetEa DeviceObject:%p, Irp:%p FileName:%wZ\n", DeviceObject, Irp, &FileObject->FileName);

	Ccb = (PREGFS_CCB)FileObject->FsContext2;

	if(!Ccb)
		return RegfsCompleteStatus(Irp, FALSE, STATUS_INVALID_DEVICE_REQUEST, "No EAs on volume file.");

	Dfcb = Ccb->Dfcb;

	if(Dfcb->Flags & REGFS_DFCB_DIRECTORY)
		return RegfsCompleteStatus(Irp, FALSE, STATUS_SUCCESS, "No EA fields can be set on a directory (no ZwSetInformationKey).");

	KeyHandle = RegfsGetDfcbKeyHandle(Dfcb);
	
	{
		ULONG KeyValueInfoLength;
		PKEY_VALUE_PARTIAL_INFORMATION KeyValueInfo;

		ExAcquireResourceExclusive(Dfcb->Fcb.Lock, TRUE);

		Status = ZwQueryValueKey(KeyHandle, &Dfcb->ShortName, KeyValuePartialInformation, NULL, 0, &KeyValueInfoLength);

		if(Status != STATUS_BUFFER_TOO_SMALL && NT_ERROR(Status))
		{
			ExReleaseResourceLite(Dfcb->Fcb.Lock);
			return RegfsCompleteKeyStatus(Irp, TRUE, Status, "Error querying key value size.");
		}

		KeyValueInfo = ExAllocatePoolWithTag(PagedPool, KeyValueInfoLength, TAG_EA_KEYVAL_INFO);

		if(!KeyValueInfo)
		{
			ExReleaseResourceLite(Dfcb->Fcb.Lock);
			return RegfsCompleteStatus(Irp, TRUE, STATUS_NO_MEMORY, "Out of memory!");
		}

		Status = ZwQueryValueKey(KeyHandle, &Dfcb->ShortName, KeyValuePartialInformation, KeyValueInfo, KeyValueInfoLength, PAGED_BIT_BUCKET);

		if(!NT_SUCCESS(Status) && Status != STATUS_BUFFER_OVERFLOW)
		{
			ExFreePoolWithTag(KeyValueInfo, TAG_EA_KEYVAL_INFO);
			ExReleaseResourceLite(Dfcb->Fcb.Lock);
			return RegfsCompleteKeyStatus(Irp, TRUE, Status, "Error querying key value.");
		}

		RegfsUpdateFsRtlSizes(FileObject, KeyValueInfo->DataLength);

		EaStruct.TitleIndex = KeyValueInfo->TitleIndex;
		EaStruct.Type = KeyValueInfo->Type;

		Status = RegfsStructureEaList(FALSE, (PFILE_FULL_EA_INFORMATION)IoBuffer, IoBufferLength, &EaStruct);

		if(!NT_SUCCESS(Status))
		{
			ExFreePoolWithTag(KeyValueInfo, TAG_EA_KEYVAL_INFO);
			ExReleaseResourceLite(Dfcb->Fcb.Lock);
			return RegfsCompleteStatus(Irp, TRUE, Status, "Error in EA list.");
		}

		Status = ZwSetValueKey(KeyHandle, &Dfcb->ShortName, EaStruct.TitleIndex, EaStruct.Type, KeyValueInfo->Data, KeyValueInfo->DataLength);

		ExFreePoolWithTag(KeyValueInfo, TAG_EA_KEYVAL_INFO);
		ExReleaseResourceLite(Dfcb->Fcb.Lock);

		if(!NT_SUCCESS(Status))
			return RegfsCompleteKeyStatus(Irp, TRUE, Status, "Error setting key value.");
	}

	return RegfsCompleteKeyStatus(Irp, TRUE, STATUS_SUCCESS, "Success.");
}