/*
 * COPYRIGHT:  2007 Christopher Smith
 * LICENCE:    GPL v2: See COPYING in the top level directory
 * PROJECT:    regfs
 * FILE:       fcbtable.c
 * PURPOSE:    Registry filesystem FCB table functions
 */

#include "regfs.h"

RTL_GENERIC_TABLE FcbTable;
RTL_GENERIC_TABLE DcbTable;

//These locks protect FcbTable and DcbTable, as well as the reference counts for all FCBs and DCBs respectively
//Either lock can be acquired individually, but to acquire both, FcbLock must be first

typedef struct _FCB_LOCKS
{
	SIMPLE_MUTEX FcbLock;
	SIMPLE_MUTEX DcbLock;
}FCB_LOCKS, *PFCB_LOCKS;

PFCB_LOCKS Locks;

PSIMPLE_MUTEX FcbLock;
PSIMPLE_MUTEX DcbLock;

RTL_GENERIC_COMPARE_RESULTS CompareTableEntry(PRTL_GENERIC_TABLE Table, PREGFS_DFCB First, PREGFS_DFCB Second)
{
	LONG Result = RtlCompareUnicodeString(&First->FileName, &Second->FileName, FALSE);
	if(Result == 0)
		return GenericEqual;
	if(Result > 0)
		return GenericGreaterThan;
	return GenericLessThan;
}

PVOID AllocateTableEntry(PRTL_GENERIC_TABLE Table, CLONG ByteSize)
{
	return ExAllocatePoolWithTag(NonPagedPool, ByteSize, TAG_DFCB);
}

VOID FreeTableEntry(PRTL_GENERIC_TABLE Table, PVOID Buffer)
{
	ExFreePoolWithTag(Buffer, TAG_DFCB);
}

NTSTATUS NTAPI RegfsInitFcbTable()
{
	Locks = ExAllocatePoolWithTag(SIMPLE_MUTEX_STORAGE, sizeof(FCB_LOCKS), TAG_FCB_TABLE_LOCKS);

	if(!Locks)
		return STATUS_NO_MEMORY;

	FcbLock = &Locks->FcbLock;
	DcbLock = &Locks->DcbLock;

	InitializeSimpleMutex(FcbLock);
	InitializeSimpleMutex(DcbLock);
	RtlInitializeGenericTable(&FcbTable, &CompareTableEntry, &AllocateTableEntry, &FreeTableEntry, NULL);
	RtlInitializeGenericTable(&DcbTable, &CompareTableEntry, &AllocateTableEntry, &FreeTableEntry, NULL);

	return STATUS_SUCCESS;
}

VOID CleanupTable(PRTL_GENERIC_TABLE Table, PSIMPLE_MUTEX Lock)
{
	PVOID PNull = NULL;
	PREGFS_DFCB Dfcb;

	AcquireSimpleMutex(Lock);

	while(Dfcb = RtlEnumerateGenericTableWithoutSplaying(Table, &PNull))
	{
		if(RtlDeleteElementGenericTable(Table, Dfcb))
		{
			ExFreePoolWithTag(Dfcb->FileName.Buffer, TAG_DFCB_NAME);
			DPRINT_V("Successfuly deleted dead element %p", Dfcb);
		}
		else
			DPRINT_V("Failed to delete dead element %p", Dfcb);
	}

	ReleaseSimpleMutex(Lock);
}


VOID RegfsCleanupFcbTable()
{
	CleanupTable(&FcbTable, FcbLock);
	CleanupTable(&DcbTable, DcbLock);

	ExFreePoolWithTag(Locks, TAG_FCB_TABLE_LOCKS);
}

NTSTATUS NTAPI InitializeDfcb(PREGFS_DFCB Dfcb, PUNICODE_STRING Name, HANDLE KeyHandle, PREGFS_DFCB Dcb)
{
	DPRINT_V("Initializing a new Dfcb at %p Dcb:%p Name:%wZ\n", Dfcb, Dcb, Name);

	Dfcb->FileName.Length = Dfcb->FileName.MaximumLength = Name->Length;
	Dfcb->FileName.Buffer = ExAllocatePoolWithTag(PagedPool, Name->Length, TAG_DFCB_NAME);
	if(!Dfcb->FileName.Buffer)
	{
		DPRINT("Error: Out of memory for %x!\n", Name->Length);
		return STATUS_NO_MEMORY;
	}
	RtlCopyMemory(Dfcb->FileName.Buffer, Name->Buffer, Name->Length);
	RegfsSeperateLastName(Dfcb->FileName, &Dfcb->ShortName, NULL); //Dfcb->ShortName = RegfsExtractLastName(Dfcb->FileName);
	Dfcb->RefCount = 1;

	if(Dcb) //it's a FCB
	{
		KEY_VALUE_BASIC_INFORMATION KeyValueInfo;
		PREGFS_FCB_FSRTL FcbFsRtl;

		if(RtlEqualUnicodeString(&Dfcb->ShortName, &REGFS_DEFAULT_VALUE_ALIAS, FALSE)) //special case for unnamed value
			Dfcb->ShortName.Length = 0;

		Dfcb->Fcb.Lock = ExAllocatePoolWithTag(NonPagedPool, sizeof(ERESOURCE), TAG_FCB_LOCK);
		if(!Dfcb->Fcb.Lock)
		{
			ExFreePoolWithTag(Dfcb->FileName.Buffer, TAG_DFCB_NAME);
			DPRINT("Error: Out of memory for ERESOURCE (%xb)!\n", sizeof(ERESOURCE));
			return STATUS_NO_MEMORY;
		}
		ExInitializeResourceLite(Dfcb->Fcb.Lock);

		Dfcb->Fcb.Directory = Dcb;
		Dfcb->Flags = 0;

		//init FcbFsRtl
		FcbFsRtl = ExAllocatePoolWithTag(NonPagedPool, sizeof(REGFS_FCB_FSRTL), TAG_FCB_FSRTL);

		if(!FcbFsRtl)
		{
			RegfsDereferenceDfcb(Dfcb);
			ExFreePoolWithTag(Dfcb->FileName.Buffer, TAG_DFCB_NAME);
			DPRINT("Error: Out of memory for REGFS_FCB_FSRTL (%xb)!\n", sizeof(REGFS_FCB_FSRTL));
			return STATUS_NO_MEMORY;
		}

		RtlZeroMemory(FcbFsRtl, sizeof(REGFS_FCB_FSRTL));

		FcbFsRtl->FsRtl.Resource = Dfcb->Fcb.Lock;
		FcbFsRtl->FsRtl.PagingIoResource = Dfcb->Fcb.Lock;
		FcbFsRtl->FsRtl.AllocationSize.LowPart =
		FcbFsRtl->FsRtl.ValidDataLength.LowPart =
		FcbFsRtl->FsRtl.FileSize.LowPart = 0;
		FcbFsRtl->FsRtl.IsFastIoPossible = FastIoIsNotPossible;

		RtlZeroMemory(&FcbFsRtl->SecPtrs, sizeof(SECTION_OBJECT_POINTERS));

#if KERNEL_VERSION >= 0x501
		ExInitializeFastMutex(&FcbFsRtl->FsRtl2Lock);
		FsRtlSetupAdvancedHeader(&FcbFsRtl->FsRtl2, &FcbFsRtl->FsRtl2Lock);
#endif

		Dfcb->Fcb.FcbFsRtl = FcbFsRtl;
	}
	else //it's a DCB
	{
		Dfcb->Dcb.KeyHandle = KeyHandle;
		Dfcb->Flags = REGFS_DFCB_DIRECTORY;
	}

	return STATUS_SUCCESS;
}

VOID NTAPI RegfsDereferenceDfcb(PREGFS_DFCB Dfcb)
{
	PSIMPLE_MUTEX Lock;
	PRTL_GENERIC_TABLE Table;

	DPRINT_V("Dereferencing Dfcb at %p New refcount:%x Name:%wZ\n", Dfcb, (Dfcb->RefCount) - 1, &Dfcb->FileName);

	if(Dfcb->Flags & REGFS_DFCB_DIRECTORY)
	{
		Lock = DcbLock;
		Table = &DcbTable;
	}
	else
	{
		Lock = FcbLock;
		Table = &FcbTable;
	}

	AcquireSimpleMutex(Lock);

	Dfcb->RefCount--;

	if(Dfcb->RefCount == 0)
	{
		PWSTR NameBuffer = Dfcb->FileName.Buffer;
		DPRINT_V("Deleting.\n");

		if(Dfcb->Flags & REGFS_DFCB_DIRECTORY)
		{
			ZwClose(Dfcb->Dcb.KeyHandle);
			Dfcb->Dcb.KeyHandle = NULL;
		}
		else
		{
			ExDeleteResourceLite(Dfcb->Fcb.Lock);
			ExFreePoolWithTag(Dfcb->Fcb.Lock, TAG_FCB_LOCK);
			ExFreePoolWithTag(Dfcb->Fcb.FcbFsRtl, TAG_FCB_FSRTL);
			RegfsDereferenceDfcb(Dfcb->Fcb.Directory);

#if KERNEL_VERSION >= 0x501
			FsRtlTeardownPerStreamContexts(&Dfcb->Fcb.FcbFsRtl->FsRtl2);
#endif
		}

		if(RtlDeleteElementGenericTable(Table, Dfcb))
		{
			DPRINT_V("Freeing.\n");
			ExFreePoolWithTag(NameBuffer, TAG_DFCB_NAME);	
		}//otherwise, the only thing that stays with a 0-ref Dfcb is the name
	}

	ReleaseSimpleMutex(Lock);
}

//if ValueName is present, Dest is a Fcb; else Dest is a Dcb
//regardless of success, this function takes ownership of KeyHandle regardless if its used for the Dcb or not
NTSTATUS NTAPI RegfsDemandDfcb(OUT PREGFS_DFCB *DestDfcb, IN PUNICODE_STRING KeyName, IN PUNICODE_STRING ValueName OPTIONAL, IN HANDLE KeyHandle)
{
	NTSTATUS Status;
	PREGFS_DFCB Dcb, Fcb;
	REGFS_DFCB EmptyEntry;
	BOOLEAN NewEntry;

	DPRINT_V("Demanding Dfcb for KeyName:%wZ ValueName:%wZ\n", KeyName, ValueName);

	EmptyEntry.FileName = *KeyName;

	AcquireSimpleMutex(DcbLock);

	Dcb = RtlInsertElementGenericTable(&DcbTable, &EmptyEntry, sizeof EmptyEntry, &NewEntry);

	if(!Dcb)
	{
		DPRINT_V("Error: Out of memory!\n");
		ZwClose(KeyHandle);
		ReleaseSimpleMutex(DcbLock);
		return STATUS_NO_MEMORY;
	}

	if(NewEntry)
	{
		Status = InitializeDfcb(Dcb, KeyName, KeyHandle, NULL);

		if(!NT_SUCCESS(Status))
		{
			DPRINT("Error initializing Dcb. Status: %x\n", Status);
			ZwClose(KeyHandle);
			RtlDeleteElementGenericTable(&DcbTable, Dcb);
			ReleaseSimpleMutex(DcbLock);
			return Status;
		}
	}
	else
	{
		if(Dcb->RefCount) //is this Dcb active?
		{
			Dcb->RefCount++;
			ZwClose(KeyHandle);
			DPRINT_V("Using existing Dcb at %p New refcount:%x\n", Dcb, Dcb->RefCount);
		}
		else
		{
			Dcb->RefCount = 1;
			Dcb->Dcb.KeyHandle = KeyHandle;
			DPRINT_V("Revived zero-reference Dcb at %p New refcount:1\n", Dcb);
		}
	}
	//the KeyHandle parameter has either been closed or committed to the Dcb

	ReleaseSimpleMutex(DcbLock);
	
	if(ValueName)
	{
		EmptyEntry.FileName = *ValueName;

		AcquireSimpleMutex(FcbLock);

		Fcb = RtlInsertElementGenericTable(&FcbTable, &EmptyEntry, sizeof EmptyEntry, &NewEntry);

		if(!Fcb)
		{
			DPRINT("Error: Out of memory!\n");
			RegfsDereferenceDfcb(Dcb);
			ReleaseSimpleMutex(FcbLock);
			return STATUS_NO_MEMORY;
		}

		if(NewEntry)
		{
			Status = InitializeDfcb(Fcb, ValueName, NULL, Dcb);

			if(!NT_SUCCESS(Status))
			{
				DPRINT("Error initializing Fcb. Status: %x\n", Status);
				RtlDeleteElementGenericTable(&FcbTable, Fcb);
				RegfsDereferenceDfcb(Dcb);
				ReleaseSimpleMutex(FcbLock);
				return Status;
			}
		}
		else
		{
			Dcb->RefCount--;//referenced through Fcb instead
			Fcb->RefCount++;
			DPRINT_V("Using existing Fcb at %p New refcount:%x\n", Fcb, Fcb->RefCount);
		}

		ReleaseSimpleMutex(FcbLock);
		*DestDfcb = Fcb;
		return STATUS_SUCCESS;
	}
	else
	{
		*DestDfcb = Dcb;
		return STATUS_SUCCESS;
	}
}