/*
 * COPYRIGHT:  2007 Christopher Smith
 * LICENCE:    GPL v2: See COPYING in the top level directory
 * PROJECT:    regfs
 * FILE:       keytree.c
 * PURPOSE:    Registry filesystem manual operation-across-entire-tree support
 */

#include "regfs.h"

#undef DPRINT_V
#define DPRINT_V DPRINT

//these are arranged in a stack
typedef struct _TREE_KEY_NODE
{
	SINGLE_LIST_ENTRY ZoneEntry;
	struct _TREE_KEY_NODE *Up;
	ULONG EnumPos;
	HANDLE SrcKey;
	HANDLE DestKey;
}TREE_KEY_NODE, *PTREE_KEY_NODE;

#define ZONE_BLOCK_LEN		AlignUpTo(sizeof(TREE_KEY_NODE), 4) //64-bit blocks req'd
#define ZONE_SEG_LEN		PAGE_SIZE

NTSTATUS CreateTreeKeyNode(OUT PTREE_KEY_NODE *NewTk, IN OUT PZONE_HEADER Zone)
{
	NTSTATUS Status;

	if(ExIsFullZone(Zone))
	{
		PVOID Seg = ExAllocatePoolWithTag(PagedPool, ZONE_SEG_LEN, TAG_TREE_KEY_NODE);

		if(!Seg)
			return STATUS_NO_MEMORY;

		Status = ExExtendZone(Zone, Seg, ZONE_SEG_LEN);

		if(!NT_SUCCESS(Status))
			return Status;
	}

	*NewTk = ExAllocateFromZone(Zone);

	RtlZeroMemory(*NewTk, sizeof(TREE_KEY_NODE));

	return STATUS_SUCCESS;
}

VOID DestroyZoneSegs(IN PZONE_HEADER Zone)
{
	PSINGLE_LIST_ENTRY Entry = Zone->FreeList.Next;

	while(Entry)
	{
		PSINGLE_LIST_ENTRY NextEntry = Entry->Next;

		ExFreePoolWithTag(Entry, TAG_TREE_KEY_NODE);

		Entry = NextEntry;
	}
}

NTSTATUS CopyKey(IN HANDLE SrcKey, IN HANDLE DestRoot, OUT PHANDLE DestKey, IN PUNICODE_STRING DestName, PKEY_NODE_INFORMATION Kni, IN KPROCESSOR_MODE RequestorMode, IN PSECURITY_SUBJECT_CONTEXT SecurityContext OPTIONAL, PVOID *Buffer, PULONG BufferLength)
{
	NTSTATUS Status;
	ACCESS_MASK GrantedAccess;
	ULONG BufferLengthReq;
	UNICODE_STRING ClassName;

	//query security
	BufferLengthReq = 0x100; //initial guess
	do
	{
		Status = RegfsExpandBufferIfNecessary(Buffer, BufferLength, BufferLengthReq, TAG_TREE_BUFFER);

		if(!NT_SUCCESS(Status))
			return Status;

		Status = ZwQuerySecurityObject(SrcKey, DACL_SECURITY_INFORMATION | SACL_SECURITY_INFORMATION | OWNER_SECURITY_INFORMATION | GROUP_SECURITY_INFORMATION, (PSECURITY_DESCRIPTOR)*Buffer, *BufferLength, &BufferLengthReq);

		if(!NT_SUCCESS(Status) && !IsBufferSizeStatus(Status))
			return RegfsKeyStatusToFileStatus(Status);

	}while(BufferLengthReq > *BufferLength);

	ClassName.Length = ClassName.MaximumLength = (USHORT)Kni->ClassLength;
	if(ClassName.Length)
		ClassName.Buffer = (PWCHAR)((ULONG_PTR)Buffer + Kni->ClassOffset);
	else
		ClassName.Buffer = NULL;

	Status = RegfsCreateKey(DestKey, STANDARD_RIGHTS_WRITE | KEY_CREATE_SUB_KEY | KEY_SET_VALUE, &GrantedAccess, RequestorMode, SecurityContext, DestName, 0, DestRoot, (PSECURITY_DESCRIPTOR)*Buffer, Kni->TitleIndex, &ClassName, 0, PAGED_BIT_BUCKET);

	DPRINT_V("RegfsCreateKey in CopyKey Status:%X DestRoot:%X DestName:%wZ\n", Status, DestRoot, DestName);

	return Status;	
}

/*RegfsKeyTreeOp executes an operation across an entire tree of keys
See REGFS_KEY_OP values for operations
The meaning of Dcb and Name depend on the operation
SecurityContext can be null if RequestorMode == KernelMode

RegfsKeyTreeOp will make the required security check on all keys (if RequestorMode == UserMode, according to SecurityContext)
RegfsKeyTreeOp must be called where it is safe to create and access key handles (same as most IRP dispatch)*/
NTSTATUS NTAPI RegfsKeyTreeOp(IN REGFS_KEY_TREE_OP Op, IN HANDLE BaseKey, IN HANDLE DestRootKey OPTIONAL, IN PUNICODE_STRING DestName OPTIONAL, IN KPROCESSOR_MODE RequestorMode, IN PSECURITY_SUBJECT_CONTEXT SecurityContext OPTIONAL)
{
	ZONE_HEADER Zone;
	union
	{
		PVOID Void;
		PKEY_NODE_INFORMATION Kni;
		PKEY_VALUE_FULL_INFORMATION Kvfi;
	}Buffer;
	ULONG BufferLen, BufferLenReq;

	PVOID Buffer2;
	ULONG Buffer2Len;

	NTSTATUS Status;
	PTREE_KEY_NODE Tk;

	{//init zone and create initial zone segment
		PVOID InitialSeg = ExAllocatePoolWithTag(PagedPool, ZONE_SEG_LEN, TAG_TREE_KEY_NODE);

		if(!InitialSeg)
			return STATUS_NO_MEMORY;

		Status = ExInitializeZone(&Zone, ZONE_BLOCK_LEN, InitialSeg, ZONE_SEG_LEN);

		if(!NT_SUCCESS(Status))
		{
			ExFreePoolWithTag(InitialSeg, TAG_TREE_KEY_NODE);
			return Status;
		}
	}

	//init top node
	Status = CreateTreeKeyNode(&Tk, &Zone);

	if(!NT_SUCCESS(Status))
	{
		DestroyZoneSegs(&Zone);
		return Status;
	}

	Buffer.Void = NULL;
	BufferLen = 0;

	Buffer2 = NULL;
	Buffer2Len = 0;

	Tk->SrcKey = BaseKey;

	if(Op == KtoCopy)
	{
		//query initial key information
		BufferLenReq = 0x80; //initial guess
		do
		{
			Status = RegfsExpandBufferIfNecessary(&Buffer.Void, &BufferLen, BufferLenReq, TAG_TREE_BUFFER);

			if(!NT_SUCCESS(Status))
				goto exit;

			Status = ZwQueryKey(BaseKey, KeyNodeInformation, Buffer.Kni, BufferLen, &BufferLenReq);

			if(!NT_SUCCESS(Status) && !IsBufferSizeStatus(Status))
				goto exit;

		}while(BufferLenReq > BufferLen);
	}

	while(TRUE) //only exited by "goto exit"
	{
		if(Op == KtoCopy)
		{
			ULONG Index = 0;

			//create dest key
			if(!Tk->Up) //if the root key
				Status = CopyKey(Tk->SrcKey, DestRootKey, &Tk->DestKey, DestName, Buffer.Kni, RequestorMode, SecurityContext, &Buffer2, &Buffer2Len);
			else
			{
				UNICODE_STRING Keyname_UCS;

				Keyname_UCS.Length = Keyname_UCS.MaximumLength = (USHORT)Buffer.Kni->NameLength;
				Keyname_UCS.Buffer = Buffer.Kni->Name;

				Status = CopyKey(Tk->SrcKey, Tk->Up->DestKey, &Tk->DestKey, &Keyname_UCS, Buffer.Kni, RequestorMode, SecurityContext, &Buffer2, &Buffer2Len);
			}

			if(!NT_SUCCESS(Status))
				goto exit;

			//copy values
			while(TRUE)
			{
				UNICODE_STRING ValueName;

				BufferLenReq = 0x80; //initial guess
				while(TRUE)
				{
					Status = RegfsExpandBufferIfNecessary(&Buffer.Void, &BufferLen, BufferLenReq, TAG_TREE_BUFFER);

					if(!NT_SUCCESS(Status))
						goto exit;
					
					Status = ZwEnumerateValueKey(Tk->SrcKey, Index, KeyValueFullInformation, Buffer.Kvfi, BufferLen, &BufferLenReq);

					if(IsBufferSizeStatus(Status))
						continue;

					Index++;

					break;
				}

				if(Status == STATUS_NO_MORE_ENTRIES)
					break;

				if(!NT_SUCCESS(Status))
					goto exit;

				ValueName.Length = ValueName.MaximumLength = (USHORT)Buffer.Kvfi->NameLength;
				ValueName.Buffer = Buffer.Kvfi->Name;

				DPRINT_V("Enumerated #%X %wZ\n", Index, &ValueName);

				Status = ZwSetValueKey(Tk->DestKey, &ValueName, Buffer.Kvfi->TitleIndex, Buffer.Kvfi->Type, (PVOID)((ULONG_PTR)Buffer.Kvfi + Buffer.Kvfi->DataOffset), Buffer.Kvfi->DataLength);

				if(!NT_SUCCESS(Status))
					goto exit;
			}
		}

		//enum next subkey or return up the treekeynode stack if none left
		//open next via break
		//go up stack via Tk = Tk->Up, continue
		while(TRUE)
		{
			//enum next subkey
			BufferLenReq = 0x80; //initial guess
			while(TRUE)
			{
				Status = RegfsExpandBufferIfNecessary(&Buffer.Void, &BufferLen, BufferLenReq, TAG_TREE_BUFFER);

				if(!NT_SUCCESS(Status))
					goto exit;

				Status = ZwEnumerateKey(Tk->SrcKey, Tk->EnumPos, KeyNodeInformation, Buffer.Kni, BufferLen, &BufferLenReq);

				if(IsBufferSizeStatus(Status))
					continue;

				Tk->EnumPos++;
				break;
			}

			if(Status == STATUS_NO_MORE_ENTRIES)
			{
				if(Op == KtoDelete)
					ZwDeleteKey(Tk->SrcKey);

				ExFreeToZone(&Zone, Tk);

				Tk = Tk->Up;
				if(!Tk)
				{
					Status = STATUS_SUCCESS;
					goto exit; //if no Tk, we've finished entire stack: the only successful exit path
				}
				continue;
			}

			if(!NT_SUCCESS(Status))
				goto exit;

			break;
		}

		//open next subkey
		{
			PTREE_KEY_NODE NewTk;
			ACCESS_MASK GrantedAccess;

			Status = CreateTreeKeyNode(&NewTk, &Zone);

			if(!NT_SUCCESS(Status))
				goto exit;

			NewTk->Up = Tk;

			Status = RegfsOpenKey2(&NewTk->SrcKey, STANDARD_RIGHTS_READ | KEY_QUERY_VALUE | KEY_ENUMERATE_SUB_KEYS, &GrantedAccess, RequestorMode, SecurityContext, Buffer.Kni->Name, (USHORT)Buffer.Kni->NameLength, 0, Tk->SrcKey);

			Tk = NewTk;

			DPRINT_V("Opening next subkey Status:%X\n", Status);

			if(!NT_SUCCESS(Status))
				goto exit;
		}
	}

//go here for fatal errors or when fininished to cleanup resources and return Status
exit:
	//close tree key node key handles
	while(Tk)
	{
		if(Tk->SrcKey && Tk->SrcKey != (HANDLE)-1 && Tk->Up) //don't close top SrcKey
			ZwClose(Tk->SrcKey);
		if(Tk->DestKey && Tk->DestKey != (HANDLE)-1)
			ZwClose(Tk->DestKey);

		Tk = Tk->Up;
	}

	//free zone segs
	{
		PSINGLE_LIST_ENTRY ZoneSeg = Zone.SegmentList.Next, NextZoneSeg;

		while(ZoneSeg)
		{
			NextZoneSeg = ZoneSeg->Next;
			ExFreePoolWithTag(ZoneSeg, TAG_TREE_KEY_NODE);
			ZoneSeg = NextZoneSeg;
		}
	}

	//free buffers
	if(Buffer.Void)
		ExFreePoolWithTag(Buffer.Void, TAG_TREE_BUFFER);

	if(Buffer2)
		ExFreePoolWithTag(Buffer2, TAG_TREE_BUFFER);

	return RegfsKeyStatusToFileStatus(Status);
};