/*
 * COPYRIGHT:  2007 Christopher Smith
 * LICENCE:    GPL v2: See COPYING in the top level directory
 * PROJECT:    regfs
 * FILE:       queryinf.c
 * PURPOSE:    Registry filesystem file query and volume
 */

#include "regfs.h"

VOID FORCEINLINE NTAPI KeyInfoToFileBasicInfo(PFILE_BASIC_INFORMATION FileInfo, PKEY_BASIC_INFORMATION KeyInfo)
{
	FileInfo->CreationTime = KeyInfo->LastWriteTime;
	FileInfo->LastAccessTime = KeyInfo->LastWriteTime;
	FileInfo->LastWriteTime = KeyInfo->LastWriteTime;
	FileInfo->ChangeTime = KeyInfo->LastWriteTime;
	FileInfo->FileAttributes = FILE_ATTRIBUTE_DIRECTORY;
}

VOID FORCEINLINE NTAPI KeyValueInfoToFileBasicInfo(PFILE_BASIC_INFORMATION FileInfo, PKEY_BASIC_INFORMATION KeyInfo, PKEY_VALUE_FULL_INFORMATION ValueInfo)
{
	FileInfo->CreationTime = KeyInfo->LastWriteTime;
	FileInfo->LastAccessTime = KeyInfo->LastWriteTime;
	FileInfo->LastWriteTime = KeyInfo->LastWriteTime;
	FileInfo->ChangeTime = KeyInfo->LastWriteTime;
	FileInfo->FileAttributes = FILE_ATTRIBUTE_NORMAL;
}

VOID FORCEINLINE NTAPI KeyInfoToFileStdInfo(PFILE_STANDARD_INFORMATION FileInfo, PREGFS_DFCB Dfcb)
{
	FileInfo->AllocationSize.QuadPart = 0;
	FileInfo->EndOfFile.QuadPart = 0;
	FileInfo->NumberOfLinks = 1;
	FileInfo->DeletePending = !!(Dfcb->Flags & REGFS_DFCB_DELETE_ON_CLOSE);
	FileInfo->Directory = TRUE;
}

VOID FORCEINLINE NTAPI KeyValueInfoToFileStdInfo(PFILE_STANDARD_INFORMATION FileInfo, PKEY_VALUE_FULL_INFORMATION ValueInfo, PREGFS_DFCB Dfcb)
{
	DPRINT_V("Returning file size:%X\n", ValueInfo->DataLength);
	FileInfo->AllocationSize.QuadPart = ValueInfo->DataLength;
	FileInfo->EndOfFile.QuadPart = ValueInfo->DataLength;
	FileInfo->NumberOfLinks = 1;
	FileInfo->DeletePending = !!(Dfcb->Flags & REGFS_DFCB_DELETE_ON_CLOSE);
	FileInfo->Directory = FALSE;
}

VOID FORCEINLINE NTAPI KeyInfoToNetworkOpenInfo(PFILE_NETWORK_OPEN_INFORMATION FileInfo, PKEY_BASIC_INFORMATION KeyInfo)
{
	FileInfo->CreationTime = KeyInfo->LastWriteTime;
	FileInfo->LastAccessTime = KeyInfo->LastWriteTime;
	FileInfo->LastWriteTime = KeyInfo->LastWriteTime;
	FileInfo->ChangeTime = KeyInfo->LastWriteTime;
	FileInfo->AllocationSize.QuadPart = 0;
	FileInfo->EndOfFile.QuadPart = 0;
	FileInfo->FileAttributes = FILE_ATTRIBUTE_DIRECTORY;
}

VOID FORCEINLINE NTAPI KeyValueInfoToNetworkOpenInfo(PFILE_NETWORK_OPEN_INFORMATION FileInfo, PKEY_BASIC_INFORMATION KeyInfo, PKEY_VALUE_FULL_INFORMATION ValueInfo)
{
	FileInfo->CreationTime = KeyInfo->LastWriteTime;
	FileInfo->LastAccessTime = KeyInfo->LastWriteTime;
	FileInfo->LastWriteTime = KeyInfo->LastWriteTime;
	FileInfo->ChangeTime = KeyInfo->LastWriteTime;
	FileInfo->AllocationSize.QuadPart = ValueInfo->DataLength;
	FileInfo->EndOfFile.QuadPart = ValueInfo->DataLength;
	FileInfo->FileAttributes = FILE_ATTRIBUTE_DIRECTORY;
}

VOID FORCEINLINE NTAPI FileObjectToFilePosInfo(PFILE_POSITION_INFORMATION FileInfo, PFILE_OBJECT FileObject)
{
	FileInfo->CurrentByteOffset.QuadPart = FileObject->CurrentByteOffset.LowPart;
}

VOID FORCEINLINE NTAPI FileObjectToFileNameInfo(PFILE_NAME_INFORMATION FileInfo, PREGFS_DFCB Dfcb)
{
	FileInfo->FileNameLength = Dfcb->FileName.Length;
}

VOID FORCEINLINE NTAPI FileObjectToFileAttributeTagInfo(PFILE_ATTRIBUTE_TAG_INFORMATION FileInfo, PREGFS_DFCB Dfcb)
{
	if(Dfcb->Flags & REGFS_DFCB_DIRECTORY)
		FileInfo->FileAttributes = FILE_ATTRIBUTE_DIRECTORY;
	else
		FileInfo->FileAttributes = FILE_ATTRIBUTE_NORMAL;
}

NTSTATUS NTAPI RegfsQueryInformation(IN PDEVICE_OBJECT DeviceObject, IN PIRP Irp, IN OUT PVOID IoBuffer, IN ULONG IoBufferLength, IN PSECURITY_SUBJECT_CONTEXT SecurityContext OPTIONAL)
{
	PIO_STACK_LOCATION IrpSp;
	PFILE_OBJECT FileObject;
	NTSTATUS Status;
	HANDLE KeyHandle;
	PREGFS_DFCB Dfcb;
	KEY_BASIC_INFORMATION KeyInfo;
	KEY_VALUE_FULL_INFORMATION KeyValueInfo;
	BOOLEAN NeedKeyInfo;
	BOOLEAN NeedKeyValueInfo;
	ULONG_PTR MaxBufferAddr;
	ULONG NameDestOffset; //0 if not to be copied
		
	IrpSp = (PIO_STACK_LOCATION)IoGetCurrentIrpStackLocation(Irp);
	FileObject = IrpSp->FileObject;

	if(!FileObject->FsContext2)
		return RegfsCompleteStatus(Irp, FALSE, STATUS_INVALID_DEVICE_REQUEST, "Error: QueryInformationFile attempted on raw volume file.");

	Dfcb = ((PREGFS_CCB)FileObject->FsContext2)->Dfcb;
	KeyHandle = RegfsGetDfcbKeyHandle(Dfcb);

	MaxBufferAddr = (ULONG_PTR)IoBuffer + IoBufferLength;

	DPRINT("RegfsQueryInformation DeviceObject:%p, Irp:%p InfoClass:%u FileName:%wZ\n", DeviceObject, Irp, IrpSp->Parameters.QueryFile.FileInformationClass, &Dfcb->FileName);

	DPRINT_V("Setting MaxBufferAddr %p + %X = %p\n", IoBuffer, IoBufferLength, MaxBufferAddr);

	NeedKeyInfo = FALSE;
	NeedKeyValueInfo = FALSE;

	//find out what information is needed
	switch(IrpSp->Parameters.QueryFile.FileInformationClass)
	{
	case FileBasicInformation:	
		DPRINT("-FileBasicInformation\n");
		NeedKeyInfo = TRUE;
		if(!(Dfcb->Flags & REGFS_DFCB_DIRECTORY))
			NeedKeyValueInfo = TRUE;
		break;

	case FileStandardInformation:
		DPRINT("-FileStandardInformation\n");
		if(!(Dfcb->Flags & REGFS_DFCB_DIRECTORY))
			NeedKeyValueInfo = TRUE;
		break;

	case FilePositionInformation:
		DPRINT("-FilePositionInformation\n");
		break;

	case FileNameInformation:
		DPRINT("-FileNameInformation\n");
		break;

	case FileAttributeTagInformation:
		DPRINT("-FileAttributeTagInformation\n");
		break;

	case FileNetworkOpenInformation:
		NeedKeyInfo = TRUE;
		if(!(Dfcb->Flags & REGFS_DFCB_DIRECTORY))
			NeedKeyValueInfo = TRUE;
		DPRINT("-FileNetworkOpenInformation\n");
		break;

	case FileAllInformation:
		DPRINT("-FileAllInformation\n");
		NeedKeyInfo = TRUE;
		if(!(Dfcb->Flags & REGFS_DFCB_DIRECTORY))
			NeedKeyValueInfo = TRUE;
		break;

	case FileEaInformation:
		DPRINT("-FileEaInformation\n");
		break;

	default:
		return RegfsCompleteStatus(Irp, TRUE, STATUS_INVALID_INFO_CLASS, "Error: Unsupported information class.\n");
	}

	//get required information
	if(NeedKeyInfo)
	{
		DPRINT_V("QueryKey KeyHandle:%p", KeyHandle, &Dfcb->ShortName);
		Status = ZwQueryKey(KeyHandle, KeyBasicInformation, &KeyInfo, sizeof KeyInfo, PAGED_BIT_BUCKET);
		if(!NT_SUCCESS(Status) && Status != STATUS_BUFFER_OVERFLOW)
			return RegfsCompleteKeyStatus(Irp, TRUE, Status, "Error: Key query failed.");
	}

	if(NeedKeyValueInfo)
	{
		DPRINT_V("QueryValueKey KeyHandle:%p ShortName:%wZ\n", KeyHandle, &Dfcb->ShortName);
		Status = ZwQueryValueKey(KeyHandle, &Dfcb->ShortName, KeyValueFullInformation, &KeyValueInfo, sizeof KeyValueInfo, PAGED_BIT_BUCKET);
		if(!NT_SUCCESS(Status) && Status != STATUS_BUFFER_OVERFLOW)
			return RegfsCompleteKeyStatus(Irp, TRUE, Status, "Error: Key value query failed.");

		RegfsUpdateFsRtlSizes(FileObject, KeyValueInfo.DataLength);
	}

	NameDestOffset = 0;

	switch(IrpSp->Parameters.QueryFile.FileInformationClass)
	{
	case FileBasicInformation:	
		if(Dfcb->Flags & REGFS_DFCB_DIRECTORY)
			KeyInfoToFileBasicInfo((PFILE_BASIC_INFORMATION)IoBuffer, &KeyInfo);
		else
			KeyValueInfoToFileBasicInfo((PFILE_BASIC_INFORMATION)IoBuffer, &KeyInfo, &KeyValueInfo);
		
		Irp->IoStatus.Information = sizeof(FILE_BASIC_INFORMATION);
		break;

	case FileStandardInformation:
		if(Dfcb->Flags & REGFS_DFCB_DIRECTORY)
			KeyInfoToFileStdInfo((PFILE_STANDARD_INFORMATION)IoBuffer, Dfcb);
		else
			KeyValueInfoToFileStdInfo((PFILE_STANDARD_INFORMATION)IoBuffer, &KeyValueInfo, Dfcb);

		Irp->IoStatus.Information = sizeof(FILE_STANDARD_INFORMATION);
		break;

	case FilePositionInformation:
		FileObjectToFilePosInfo((PFILE_POSITION_INFORMATION)IoBuffer, FileObject);
		
		Irp->IoStatus.Information = sizeof(FILE_POSITION_INFORMATION);
		break;

	case FileNameInformation:
		FileObjectToFileNameInfo((PFILE_NAME_INFORMATION)IoBuffer, Dfcb);
		NameDestOffset = FIELD_OFFSET(FILE_NAME_INFORMATION, FileName);
		break;

	case FileAttributeTagInformation:
		FileObjectToFileAttributeTagInfo((PFILE_ATTRIBUTE_TAG_INFORMATION)IoBuffer, Dfcb);
		
		Irp->IoStatus.Information = sizeof(FILE_ATTRIBUTE_TAG_INFORMATION);
		break;

	case FileNetworkOpenInformation:
		if(Dfcb->Flags & REGFS_DFCB_DIRECTORY)
			KeyInfoToNetworkOpenInfo((PFILE_NETWORK_OPEN_INFORMATION)IoBuffer, &KeyInfo);
		else
			KeyValueInfoToNetworkOpenInfo((PFILE_NETWORK_OPEN_INFORMATION)IoBuffer, &KeyInfo, &KeyValueInfo);

		Irp->IoStatus.Information = sizeof(FILE_NETWORK_OPEN_INFORMATION);
		break;

	case FileEaInformation:
		RegfsQueryEaSetLength(KeyHandle, (Dfcb->Flags & REGFS_DFCB_DIRECTORY) != 0, &((PFILE_EA_INFORMATION)IoBuffer)->EaSize);

		Irp->IoStatus.Information = sizeof(FILE_EA_INFORMATION);
		break;

	case FileAllInformation:
		{
			PFILE_ALL_INFORMATION FileInfo = (PFILE_ALL_INFORMATION)IoBuffer;
			if(Dfcb->Flags & REGFS_DFCB_DIRECTORY)
			{
				KeyInfoToFileBasicInfo(&FileInfo->BasicInformation, &KeyInfo);
				KeyInfoToFileStdInfo(&FileInfo->StandardInformation, Dfcb);
				RtlZeroMemory(&FileInfo->PositionInformation, sizeof(FILE_POSITION_INFORMATION));
			}
			else
			{
				KeyValueInfoToFileBasicInfo(&FileInfo->BasicInformation, &KeyInfo, &KeyValueInfo);
				KeyValueInfoToFileStdInfo(&FileInfo->StandardInformation, &KeyValueInfo, Dfcb);
				FileObjectToFilePosInfo(&FileInfo->PositionInformation, FileObject);
			}
			RtlZeroMemory(&FileInfo->InternalInformation, sizeof(FILE_INTERNAL_INFORMATION));
			FileObjectToFileNameInfo(&FileInfo->NameInformation, Dfcb);
			NameDestOffset = FIELD_OFFSET(FILE_ALL_INFORMATION, NameInformation) + FIELD_OFFSET(FILE_NAME_INFORMATION, FileName);
			RegfsQueryEaSetLength(KeyHandle, (Dfcb->Flags & REGFS_DFCB_DIRECTORY) != 0, &FileInfo->EaInformation.EaSize);
			break;
		}
	}

	//name is always Dfcb->FileName
	if(NameDestOffset)
	{
		ULONG UnusedBuffer = IoBufferLength - NameDestOffset;

		DPRINT_V("Copying name to offset %p\n", NameDestOffset);
		
		if(UnusedBuffer < Dfcb->FileName.Length)
			Status = STATUS_BUFFER_OVERFLOW;
		else
			Status = STATUS_SUCCESS;

		if(UnusedBuffer)
			UnusedBuffer -= RegfsCopyData((PVOID)((ULONG_PTR)IoBuffer + NameDestOffset), Dfcb->FileName.Buffer, Dfcb->FileName.Length, UnusedBuffer, MaxBufferAddr);

		Irp->IoStatus.Information = IoBufferLength - UnusedBuffer;
	}
	else
		Status = STATUS_SUCCESS;

	return RegfsCompleteStatus(Irp, TRUE, Status, NULL);
}

NTSTATUS NTAPI RegfsQueryVolumeInformation(IN PDEVICE_OBJECT DeviceObject, IN PIRP Irp, IN OUT PVOID IoBuffer, IN ULONG IoBufferLength, IN PSECURITY_SUBJECT_CONTEXT SecurityContext OPTIONAL)
{
	PIO_STACK_LOCATION IrpSp;
	PFILE_OBJECT FileObject;
	NTSTATUS Status;
	ULONG ResultLength;
	ULONG_PTR MaxBufferAddr;

	IrpSp = (PIO_STACK_LOCATION)IoGetCurrentIrpStackLocation(Irp);
	FileObject = IrpSp->FileObject;

	MaxBufferAddr = (ULONG_PTR)IoBuffer + IoBufferLength;

	DPRINT("RegfsQueryVolumeInformation DeviceObject:%p, Irp:%p InfoClass:%x\n", DeviceObject, Irp, IrpSp->Parameters.QueryVolume.FsInformationClass);

	DPRINT_V("Setting MaxBufferAddr %p + %X = %p\n", IoBuffer, IoBufferLength, MaxBufferAddr);

	switch(IrpSp->Parameters.QueryVolume.FsInformationClass)
	{
	case FileFsAttributeInformation:
		{
			PFILE_FS_ATTRIBUTE_INFORMATION FsInfo = (PFILE_FS_ATTRIBUTE_INFORMATION)IoBuffer;
			ULONG UnusedBuffer;
			
			FsInfo->FileSystemAttributes = FILE_CASE_PRESERVED_NAMES | FILE_CASE_SENSITIVE_SEARCH | FILE_PERSISTENT_ACLS | FILE_UNICODE_ON_DISK;
			FsInfo->MaximumComponentNameLength = 255;
			
			UnusedBuffer = IoBufferLength - SIZEOF_FILE_FS_ATTRIBUTE_INFORMATION_HEADER;
			FsInfo->FileSystemNameLength = RegfsCopyData(&FsInfo->FileSystemName, REGFS_FS_NAME.Buffer, REGFS_FS_NAME.Length, UnusedBuffer, MaxBufferAddr);
			
			Irp->IoStatus.Information = SIZEOF_FILE_FS_ATTRIBUTE_INFORMATION_HEADER + FsInfo->FileSystemNameLength; //fixed plus used variable

			if(UnusedBuffer < REGFS_FS_NAME.Length)
				return RegfsCompleteStatus(Irp, TRUE, STATUS_BUFFER_OVERFLOW, "IoBuffer overflowed.");
		}
		break;

	case FileFsVolumeInformation:
		{
			PFILE_FS_VOLUME_INFORMATION FsInfo = (PFILE_FS_VOLUME_INFORMATION)IoBuffer;
			ULONG UnusedBuffer;
			LARGE_INTEGER SystemTime;
			LARGE_INTEGER TickCount;

			KeQuerySystemTime(&SystemTime);
			KeQueryTickCount(&TickCount);
			FsInfo->VolumeCreationTime.QuadPart = SystemTime.QuadPart - TickCount.QuadPart; //i.e. system start time
			FsInfo->VolumeSerialNumber = 0;
			FsInfo->SupportsObjects = FALSE;
			FsInfo->VolumeLabelLength = REGFS_VOLUME_NAME.Length;

			UnusedBuffer = IoBufferLength - SIZEOF_FILE_FS_VOLUME_INFORMATION_HEADER;
			FsInfo->VolumeLabelLength = RegfsCopyData(&FsInfo->VolumeLabel, REGFS_VOLUME_NAME.Buffer, REGFS_VOLUME_NAME.Length, UnusedBuffer, MaxBufferAddr);

			Irp->IoStatus.Information = SIZEOF_FILE_FS_VOLUME_INFORMATION_HEADER + FsInfo->VolumeLabelLength;

			if(UnusedBuffer < REGFS_VOLUME_NAME.Length)
				return RegfsCompleteStatus(Irp, TRUE, STATUS_BUFFER_OVERFLOW, "IoBuffer overflowed.");
		}
		break;

	case FileFsControlInformation:
		{
			PFILE_FS_CONTROL_INFORMATION FsInfo = (PFILE_FS_CONTROL_INFORMATION)IoBuffer;
			FsInfo->FreeSpaceStartFiltering.QuadPart = 0x10000000;
			FsInfo->FreeSpaceStopFiltering.QuadPart = 0x8000000;
			FsInfo->DefaultQuotaThreshold.QuadPart = 0;
			FsInfo->DefaultQuotaLimit.QuadPart = 0;
			FsInfo->FileSystemControlFlags = 0;

			Irp->IoStatus.Information = sizeof *FsInfo;
		}
		break;

	case FileFsDeviceInformation:
		{
			PFILE_FS_DEVICE_INFORMATION FsInfo = (PFILE_FS_DEVICE_INFORMATION)IoBuffer;
			FsInfo->DeviceType = DeviceObject->DeviceType;
			FsInfo->Characteristics = DeviceObject->Characteristics;
			Irp->IoStatus.Information = sizeof *FsInfo;
		}
		break;

	case FileFsSizeInformation:
		{
			PFILE_FS_SIZE_INFORMATION FsInfo = (PFILE_FS_SIZE_INFORMATION)IoBuffer;
			SYSTEM_REGISTRY_QUOTA_INFORMATION RegSizeInfo;
			
			Status = ZwQuerySystemInformation(SystemRegistryQuotaInformation, &RegSizeInfo, sizeof RegSizeInfo, &ResultLength);
			if(!NT_SUCCESS(Status))
				return RegfsCompleteStatus(Irp, TRUE, Status, "Error getting registry quota information.");

			FsInfo->TotalAllocationUnits.QuadPart = RegSizeInfo.RegistryQuota;
			FsInfo->AvailableAllocationUnits.QuadPart = RegSizeInfo.RegistryQuota - RegSizeInfo.RegistryQuotaInUse;
			FsInfo->SectorsPerAllocationUnit = 1;
			FsInfo->BytesPerSector = 1;
			Irp->IoStatus.Information = sizeof *FsInfo;
		}
		break;

	case FileFsFullSizeInformation:
		{
			PFILE_FS_FULL_SIZE_INFORMATION FsInfo = (PFILE_FS_FULL_SIZE_INFORMATION)IoBuffer;
			SYSTEM_REGISTRY_QUOTA_INFORMATION RegSizeInfo;
			
			Status = ZwQuerySystemInformation(SystemRegistryQuotaInformation, &RegSizeInfo, sizeof RegSizeInfo, &ResultLength);
			if(!NT_SUCCESS(Status))
				return RegfsCompleteStatus(Irp, TRUE, Status, "Error getting registry quota information.");

			FsInfo->TotalAllocationUnits.QuadPart = RegSizeInfo.RegistryQuota;
			FsInfo->CallerAvailableAllocationUnits.QuadPart = 
			FsInfo->ActualAvailableAllocationUnits.QuadPart = RegSizeInfo.RegistryQuota - RegSizeInfo.RegistryQuotaInUse;
			FsInfo->SectorsPerAllocationUnit = 1;
			FsInfo->BytesPerSector = 1;
			Irp->IoStatus.Information = sizeof *FsInfo;
		}
		break;

	default:
		return RegfsCompleteStatus(Irp, TRUE, STATUS_INVALID_INFO_CLASS, "Error: Unsupported information class.\n");
	}
	return RegfsCompleteStatus(Irp, TRUE, STATUS_SUCCESS, "Success.");
}