/*
 * COPYRIGHT:  2007 Christopher Smith
 * LICENCE:    GPL v2: See COPYING in the top level directory
 * PROJECT:    regfs
 * FILE:       regfs.h
 * PURPOSE:    Registry filesystem common header
 */

#ifndef __DRIVERS_FS_REGFS_H
#define __DRIVERS_FS_REGFS_H

//both the official MS ntifs.h and the GPL'd ntifs.h by Bo Brantn (included) are supported
//move/rename the included ntifs.h to use Microsoft's instead
#include <ntifs.h>

//build environment specific helper definitions
#include "hdrhelp.h"

//#define NODPRINT //to disable debug output in a checked build (already disabled in free)

#ifndef DECLARE_CONST_UNICODE_STRING

#define DECLARE_CONST_UNICODE_STRING(_var, _string) \
const WCHAR _var ## _buffer[] = _string; \
const UNICODE_STRING _var = { sizeof(_string) - sizeof(WCHAR), sizeof(_string), (PWCH) _var ## _buffer }

#endif


#ifndef DECLARE_CONST_SELECTANY_UNICODE_STRING

#define DECLARE_CONST_SELECTANY_UNICODE_STRING(_var, _string) \
const WCHAR SELECTANY _var ## _buffer[] = _string; \
const UNICODE_STRING SELECTANY _var = { sizeof(_string) - sizeof(WCHAR), sizeof(_string), (PWCH) _var ## _buffer }

#endif


#ifndef DECLARE_CONST_STRING

#define DECLARE_CONST_STRING(_var, _string) \
const CHAR _var ## _buffer[] = _string; \
const STRING _var = { sizeof(_string) - sizeof(CHAR), sizeof(_string), (PCH) _var ## _buffer }

#endif

#ifndef FORCEINLINE
#define FORCEINLINE __forceinline
#endif

#if DBG
	#ifdef NODPRINT
	#define DPRINT
	#else
	#define DPRINT DbgPrint
	#endif
#else
#define DPRINT
#endif

#if VERBOSE_MESSAGES
#define DPRINT_V DPRINT
#else
#define DPRINT_V
#endif

//modules are paged by default (subject to override after header include)
#pragma code_seg("PAGEDCODE")
#pragma data_seg("PAGEDDATA")
#pragma bss_seg("PAGEDDATA")
#pragma const_seg("PAGEDDATA")

#define TRY_LEAVE(expr) {expr; leave;}

#define SELECTANY DECLSPEC_SELECTANY

//include os version specific definitions
#include <osver.h>

typedef NTSTATUS (*PREGFS_IRP_DISPATCH)(IN PDEVICE_OBJECT DeviceObject, IN PIRP Irp, IN OUT PVOID IoBuffer, IN ULONG IoBufferLength, IN PSECURITY_SUBJECT_CONTEXT SecurityContext OPTIONAL);

typedef enum _REGFS_ENUMERATION_MODE
{
	EnumDotsMode,	//special files "." and ".."
	EnumKeysMode,	//keys (directories)
	EnumValuesMode,	//values (files)
	EnumEndMode		//no more entries
}REGFS_ENUMERATION_MODE, *PREGFS_ENUMERATION_MODE;

typedef enum _REGFS_DISPATCH_METHOD
{
	DmObjKernelHandle,		//use OBJ_KERNEL_HANDLE to enable dispatch in caller's context
	DmSystemThread,			//use a driver-created system thread to dispatch in system context
	DmExWorkItem,			//use executive work items to dispatch in system context
	DmMaximum = DmExWorkItem//maximum supported mode
}REGFS_DISPATCH_METHOD, *PREGFS_DISPATCH_METHOD;

typedef enum _REGFS_NOTIFY_TYPE
{
	NoTyIrp,		//notify the attached IRP
	NoTySection,	//synch FCB's section
}REGFS_NOTIFY_TYPE, *PREGFS_NOTIFY_TYPE;

typedef enum _REGFS_KEY_TREE_OP
{
	KtoCopy, //Dcb = src, Name = dest
	KtoDelete //Dcb = target, Name = NULL
}REGFS_KEY_TREE_OP, *PREGFS_KEY_TREE_OP;

typedef struct _KEY_QUERY_BUFFER
{
	ULONG Length;
	union
	{
		PUCHAR Addr;
		PKEY_BASIC_INFORMATION KeyInfo;
		PKEY_FULL_INFORMATION KeyFullInfo;
		PKEY_VALUE_FULL_INFORMATION KeyValueInfo;
	};
}KEY_QUERY_BUFFER, *PKEY_QUERY_BUFFER;

//stored in each FILE_OBJECT.FsContext, IIF it has data (i.e. a key value)
//directories only have a Ccb (FsContext2)
//a file cannot have a FcbFsrtl (FsContext) and no Ccb (FsContext2)
//nonpaged
typedef struct _REGFS_FCB_FSRTL
{
	union
	{
		FSRTL_COMMON_FCB_HEADER FsRtl; //common FSRTL Header
#if KERNEL_VERSION >= 0x501
		FSRTL_ADVANCED_FCB_HEADER FsRtl2; //extra fields for XP+
#endif
	};
#if KERNEL_VERSION >= 0x501
	FAST_MUTEX FsRtl2Lock; //lock for FsRtl2
#endif
	SECTION_OBJECT_POINTERS SecPtrs;//mm section object state
}REGFS_FCB_FSRTL, *PREGFS_FCB_FSRTL;

struct _REGFS_DFCB;
typedef struct _REGFS_DFCB REGFS_DFCB;
typedef struct _REGFS_DFCB *PREGFS_DFCB;

typedef struct _REGFS_DCB
{
	HANDLE KeyHandle;				//handle to key available in all contexts where IRPs are handled
}REGFS_DCB, *PREGFS_DCB;

typedef struct _REGFS_FCB
{
	PREGFS_DFCB Directory;			//pointer to parent DCB
	PERESOURCE Lock;				//read/write data lock
	PREGFS_FCB_FSRTL FcbFsRtl;
}REGFS_FCB, *PREGFS_FCB;

typedef struct _REGFS_DFCB //DCB or FCB
{
	UNICODE_STRING FileName;		//full name (paged)
	UNICODE_STRING ShortName;		//short verion of name (no path) (paged, same backing as FileName)
	ULONG RefCount;					//number of CCBs (and FCBs when a DCB) dependent on this DCB | FCB
	ULONG Flags;					//misc flags
	union
	{
		REGFS_DCB Dcb;
		REGFS_FCB Fcb;
	};
}REGFS_DFCB, *PREGFS_DFCB;

//stored in each FILE_OBJECT.FsContext2
//paged
typedef struct _REGFS_CCB
{
	PREGFS_DFCB Dfcb;				//the directory or file the FILE_OBJECT references
	ULONG Flags;
	REGFS_ENUMERATION_MODE EnumMode;//current enumeration mode
	ULONG EnumIndex;				//relative enum position
	ULONG TotalEnumIndex;			//absolute enum position (across enum modes)
	UNICODE_STRING QueryString;		//query string for current enumeration, .Buffer == null if none (paged)
	ULONG EaEnumIndex;				//enum position for EAs
	PSIMPLE_MUTEX Lock;				//lock for CCB's mutable fields
}REGFS_CCB, *PREGFS_CCB;

typedef struct _REGFS_STRUCTURED_EA
{
	ULONG TitleIndex;
	union
	{
		ULONG Type;//key value option
		struct //key options
		{
			BOOLEAN Volatile;
			UNICODE_STRING Class;
		};
	};
}REGFS_STRUCTURED_EA, *PREGFS_STRUCTURED_EA;

typedef enum _REGFS_WORK_ITEM_TYPE
{
	WtTermThread,	//thread should terminate and NOT free item
	WtDispatchIrp	//call attached IRP dispatch function with attached parameters, free item and security context if any
}REGFS_WORK_ITEM_TYPE, *PREGFS_WORK_ITEM_TYPE;

typedef struct _REGFS_WORK_ITEM //must be non-paged
{
	union
	{
		LIST_ENTRY KQueueItem; //for use with a KQUEUE
		WORK_QUEUE_ITEM ExItem; //for use with an executive work item
	};
	REGFS_WORK_ITEM_TYPE Type;
	union
	{
		struct //when Type == WtDispatchIrp
		{
			PREGFS_IRP_DISPATCH Proc;
			PDEVICE_OBJECT DeviceObject;
			PIRP Irp;
			PVOID IoBuffer;
			ULONG IoBufferLength;
			BOOLEAN HasSecurityContext;
			SECURITY_SUBJECT_CONTEXT SecurityContext;
		}DIrp;
	};
}REGFS_WORK_ITEM, *PREGFS_WORK_ITEM;

typedef struct _REGFS_WORK_ITEM_SHORT //must be non-paged
{
	LIST_ENTRY ListEntry;
	REGFS_WORK_ITEM_TYPE Type;
}REGFS_WORK_ITEM_SHORT, *PREGFS_WORK_ITEM_SHORT;

typedef struct _REGFS_DEVICE_EXTENSION
{
	CHAR Pad;
}REGFS_DEVICE_EXTENSION, *PREGFS_DEVICE_EXTENSION;

#define REGFS_DFCB_DIRECTORY			0x00000001 //is this {D/F}CB for a directory (? is a Dcb : is a Fcb)
#define REGFS_DFCB_DELETE_ON_CLOSE		0x00000002 //the file should be deleted when References == 0

#define REGFS_CCB_FILES_ENUMERATED		0x00000001 //were any actual files enumerated?

#define TAG_QUERY_BUFFER				'bqFR'
#define TAG_DFCB						'cfFR'
#define TAG_DFCB_NAME					'nfFR'
#define TAG_CCB							'ccFR'
#define TAG_FCB_FSRTL					'fcFR'
#define TAG_CCB_QUERY_STRING			'qcFR'
#define TAG_EA_KEY_INFO_CLASS			'ceFR'
#define TAG_EA_KEYVAL_INFO				'veFR'
#define TAG_READ_BUFFER					'rBFR'
#define TAG_WRITE_BUFFER				'wBFR'
#define TAG_BLANK_BUFFER				'bBFR'
#define TAG_MOVE_BUFFER					'mBFR'
#define TAG_ACTUAL_NAME					'naFR'
#define TAG_SECURITY_DESC_CHECK			'dsFR'
#define TAG_SECURITY_DESC_QUERY			'qsFR'
#define TAG_WORK_QUEUE					'qwFR'
#define TAG_WORK_ITEM					'iwFR'
#define TAG_IRP_SYSTEM_BUFFER			'bsFR'
#define TAG_FCB_TABLE_LOCKS				'ltFR'
#define TAG_FCB_LOCK					'lfFR'
#define TAG_CCB_LOCK					'lcFR'
#define TAG_NOTIFY_LOCK					'lnFR'
#define TAG_DFCB_HEADER_LOCK			'lhFR'
#define TAG_NOTIFY_ENTRY				'enFR'
#define TAG_EX_LIMIT_SEM				'lxFR'
#define TAG_TREE_KEY_NODE				'ntFR'
#define TAG_TREE_BUFFER					'BtFR'

#define SIZEOF_KEY_BASIC_INFORMATION_HEADER			FIELD_OFFSET(KEY_BASIC_INFORMATION, Name)
#define SIZEOF_KEY_FULL_INFORMATION_HEADER			FIELD_OFFSET(KEY_FULL_INFORMATION, Class)
#define SIZEOF_KEY_NODE_INFORMATION_HEADER			FIELD_OFFSET(KEY_FULL_INFORMATION, Name)

#define SIZEOF_KEY_VALUE_PARTIAL_INFORMATION_HEADER	FIELD_OFFSET(KEY_VALUE_PARTIAL_INFORMATION, Data)
#define SIZEOF_KEY_VALUE_BASIC_INFORMATION_HEADER	FIELD_OFFSET(KEY_VALUE_BASIC_INFORMATION, Name)
#define SIZEOF_KEY_VALUE_FULL_INFORMATION_HEADER	FIELD_OFFSET(KEY_VALUE_FULL_INFORMATION, Name)

#define SIZEOF_FILE_BOTH_DIR_INFORMATION_HEADER		FIELD_OFFSET(FILE_BOTH_DIR_INFORMATION, FileName)
#define SIZEOF_FILE_DIRECTORY_INFORMATION_HEADER	FIELD_OFFSET(FILE_DIRECTORY_INFORMATION, FileName)
#define SIZEOF_FILE_FULL_DIR_INFORMATION_HEADER		FIELD_OFFSET(FILE_FULL_DIR_INFORMATION, FileName)
#define SIZEOF_FILE_NAMES_INFORMATION_HEADER		FIELD_OFFSET(FILE_NAMES_INFORMATION, FileName)

#define SIZEOF_FILE_NAME_INFORMATION_HEADER			FIELD_OFFSET(FILE_NAME_INFORMATION, FileName)

#define SIZEOF_FILE_FS_ATTRIBUTE_INFORMATION_HEADER	FIELD_OFFSET(FILE_FS_ATTRIBUTE_INFORMATION, FileSystemName)
#define SIZEOF_FILE_FS_VOLUME_INFORMATION_HEADER	FIELD_OFFSET(FILE_FS_VOLUME_INFORMATION, VolumeLabel)

#define SIZEOF_FILE_FULL_EA_INFORMATION_HEADER		FIELD_OFFSET(FILE_FULL_EA_INFORMATION, EaName)

//default buffer size for key/value name + header (expanded if necessary)
#define REGFS_DEFAULT_QUERY_BUFFER_SIZE		0x80
#define REGFS_STACK_OVERFLOW_THRESHOLD		0x100

#define REGFS_MAX_WORK_THREADS				32
#define REGFS_MDL_MAXIMUM					(PAGE_SIZE * (MAXUSHORT - sizeof(MDL)) / sizeof(ULONG_PTR))

#define COMMON_ACCESS_MASK (STANDARD_RIGHTS_REQUIRED | GENERIC_READ | GENERIC_WRITE | GENERIC_EXECUTE | GENERIC_ALL | ACCESS_SYSTEM_SECURITY | MAXIMUM_ALLOWED)

DECLARE_CONST_SELECTANY_UNICODE_STRING(REGFS_FS_NAME, L"Regfs");
DECLARE_CONST_SELECTANY_UNICODE_STRING(REGFS_VOLUME_NAME, L"Registry");

DECLARE_CONST_SELECTANY_UNICODE_STRING(REGFS_DEFAULT_VALUE_ALIAS, L"@");

//are EA names case insensitive?
#define REGFS_ARE_EA_NAMES_CINS				(TRUE)

PDRIVER_OBJECT SELECTANY RegfsDriverObject = NULL;
PDEVICE_OBJECT SELECTANY RegfsDeviceObject = NULL;

GENERIC_MAPPING SELECTANY KeyGenericMapping = {KEY_READ, KEY_WRITE, KEY_READ, KEY_ALL_ACCESS};

PRIVILEGE_SET SELECTANY KeyNotifyPrivilegeSet = {1, PRIVILEGE_SET_ALL_NECESSARY, {SE_CHANGE_NOTIFY_PRIVILEGE, 0}};
PRIVILEGE_SET SELECTANY SecurityPrivilegeSet = {1, PRIVILEGE_SET_ALL_NECESSARY, {SE_SECURITY_PRIVILEGE, 0}};

ULONG SELECTANY DefaultObjectAttributes = 0;

HANDLE SELECTANY SystemProcessHandle = NULL;

ULONG SELECTANY DefaultKeyValueType = 0;

ULONG SELECTANY PAGED_BIT_BUCKET[] = {0, 0, 0, 0}; //(at least) 128 bits of unwanted data

NTSTATUS NTAPI DriverEntry(PDRIVER_OBJECT DriverObject, PUNICODE_STRING RegistryPath);
VOID NTAPI RegfsUnload(IN PDRIVER_OBJECT DriverObject);

NTSTATUS NTAPI RegfsCreate(IN PDEVICE_OBJECT DeviceObject, IN PIRP Irp, IN OUT PVOID IoBuffer, IN ULONG IoBufferLength, IN PSECURITY_SUBJECT_CONTEXT SecurityContext OPTIONAL);
NTSTATUS NTAPI RegfsCleanup(IN PDEVICE_OBJECT DeviceObject, IN PIRP Irp, IN OUT PVOID IoBuffer, IN ULONG IoBufferLength, IN PSECURITY_SUBJECT_CONTEXT SecurityContext OPTIONAL);
NTSTATUS NTAPI RegfsClose(IN PDEVICE_OBJECT DeviceObject, IN PIRP Irp, IN OUT PVOID IoBuffer, IN ULONG IoBufferLength, IN PSECURITY_SUBJECT_CONTEXT SecurityContext OPTIONAL);

NTSTATUS NTAPI RegfsReadWrite(IN PDEVICE_OBJECT DeviceObject, IN PIRP Irp, IN OUT PVOID IoBuffer, IN ULONG IoBufferLength, IN PSECURITY_SUBJECT_CONTEXT SecurityContext OPTIONAL);

NTSTATUS NTAPI RegfsDirectoryControl(IN PDEVICE_OBJECT DeviceObject, IN PIRP Irp, IN OUT PVOID IoBuffer, IN ULONG IoBufferLength, IN PSECURITY_SUBJECT_CONTEXT SecurityContext OPTIONAL);
NTSTATUS NTAPI RegfsQueryDirectory(IN PDEVICE_OBJECT DeviceObject, IN PIRP Irp, IN OUT PVOID IoBuffer, IN ULONG IoBufferLength, IN PSECURITY_SUBJECT_CONTEXT SecurityContext OPTIONAL);

NTSTATUS NTAPI RegfsFlushBuffers(IN PDEVICE_OBJECT DeviceObject, IN PIRP Irp, IN OUT PVOID IoBuffer, IN ULONG IoBufferLength, IN PSECURITY_SUBJECT_CONTEXT SecurityContext OPTIONAL);

NTSTATUS NTAPI RegfsFsControl(IN PDEVICE_OBJECT DeviceObject, IN PIRP Irp, IN OUT PVOID IoBuffer, IN ULONG IoBufferLength, IN PSECURITY_SUBJECT_CONTEXT SecurityContext OPTIONAL);

NTSTATUS NTAPI RegfsQueryInformation(IN PDEVICE_OBJECT DeviceObject, IN PIRP Irp, IN OUT PVOID IoBuffer, IN ULONG IoBufferLength, IN PSECURITY_SUBJECT_CONTEXT SecurityContext OPTIONAL);
NTSTATUS NTAPI RegfsSetInformation(IN PDEVICE_OBJECT DeviceObject, IN PIRP Irp, IN OUT PVOID IoBuffer, IN ULONG IoBufferLength, IN PSECURITY_SUBJECT_CONTEXT SecurityContext OPTIONAL);
NTSTATUS NTAPI RegfsQueryVolumeInformation(IN PDEVICE_OBJECT DeviceObject, IN PIRP Irp, IN OUT PVOID IoBuffer, IN ULONG IoBufferLength, IN PSECURITY_SUBJECT_CONTEXT SecurityContext OPTIONAL);

NTSTATUS NTAPI RegfsQueryEa(IN PDEVICE_OBJECT DeviceObject, IN PIRP Irp, IN OUT PVOID IoBuffer, IN ULONG IoBufferLength, IN PSECURITY_SUBJECT_CONTEXT SecurityContext OPTIONAL);
NTSTATUS NTAPI RegfsQueryKeyEaSetLength(IN ULONG ClassLength, OUT PULONG Length);
NTSTATUS NTAPI RegfsQueryEaSetLength(IN HANDLE KeyHandle, IN BOOLEAN IsDirectory, OUT PULONG Length);
NTSTATUS NTAPI RegfsSetEa(IN PDEVICE_OBJECT DeviceObject, IN PIRP Irp, IN OUT PVOID IoBuffer, IN ULONG IoBufferLength, IN PSECURITY_SUBJECT_CONTEXT SecurityContext OPTIONAL);
NTSTATUS NTAPI RegfsStructureEaList(IN BOOLEAN IsDirectory, IN PFILE_FULL_EA_INFORMATION EaBuf, ULONG EaBufLen, IN OUT PREGFS_STRUCTURED_EA EaStruct);

NTSTATUS NTAPI RegfsFlush(IN PDEVICE_OBJECT DeviceObject, IN PIRP Irp, IN OUT PVOID IoBuffer, IN ULONG IoBufferLength, IN PSECURITY_SUBJECT_CONTEXT SecurityContext OPTIONAL);

NTSTATUS NTAPI RegfsQuerySecurity(IN PDEVICE_OBJECT DeviceObject, IN PIRP Irp, IN OUT PVOID IoBuffer, IN ULONG IoBufferLength, IN PSECURITY_SUBJECT_CONTEXT SecurityContext OPTIONAL);
NTSTATUS NTAPI RegfsSetSecurity(IN PDEVICE_OBJECT DeviceObject, IN PIRP Irp, IN OUT PVOID IoBuffer, IN ULONG IoBufferLength, IN PSECURITY_SUBJECT_CONTEXT SecurityContext OPTIONAL);

ACCESS_MASK NTAPI RegfsKeyMaskToFileMask(ACCESS_MASK KeyMask);
ACCESS_MASK NTAPI RegfsKeyValueMaskToFileMask(ACCESS_MASK KeyMask);
ACCESS_MASK NTAPI RegfsFileMaskToKeyMask(ACCESS_MASK FileMask);
ACCESS_MASK NTAPI RegfsFileMaskToKeyValueMask(ACCESS_MASK FileMask);

NTSTATUS NTAPI RegfsKeySDToFileSD(PSECURITY_DESCRIPTOR KeySD);
NTSTATUS NTAPI RegfsFileSDToKeySD(PSECURITY_DESCRIPTOR FileSD);

NTSTATUS NTAPI RegfsKeyStatusToFileStatus(NTSTATUS KeyStatus);

BOOLEAN NTAPI RegfsIsKeyHandleBogus(HANDLE Handle);

VOID NTAPI RegfsSeperateLastName(UNICODE_STRING CONST Src, OUT UNICODE_STRING *Last OPTIONAL, OUT UNICODE_STRING *Rest OPTIONAL);

NTSTATUS NTAPI RegfsCreateKey(PHANDLE KeyHandle, ACCESS_MASK DesiredAccess, PACCESS_MASK GrantedAccess, KPROCESSOR_MODE RequestMode, IN PSECURITY_SUBJECT_CONTEXT SecurityContext OPTIONAL, PUNICODE_STRING Name, ULONG ExtraAttributes, HANDLE Root, PSECURITY_DESCRIPTOR NewSecurity, ULONG TitleIndex, PUNICODE_STRING Class, ULONG CreateOptions, PULONG Disposition);
NTSTATUS NTAPI RegfsCreateKey2(PHANDLE KeyHandle, ACCESS_MASK DesiredAccess, PACCESS_MASK GrantedAccess, KPROCESSOR_MODE RequestMode, IN PSECURITY_SUBJECT_CONTEXT SecurityContext OPTIONAL, PWSTR NameBuf, USHORT NameSize, ULONG ExtraAttributes, HANDLE Root, PSECURITY_DESCRIPTOR Security, ULONG TitleIndex, PUNICODE_STRING Class, ULONG CreateOptions, PULONG Disposition);
NTSTATUS NTAPI RegfsOpenKey(PHANDLE KeyHandle, ACCESS_MASK DesiredAccess, PACCESS_MASK GrantedAccess, KPROCESSOR_MODE RequestMode, IN PSECURITY_SUBJECT_CONTEXT SecurityContext OPTIONAL, PUNICODE_STRING Name, ULONG ExtraAttributes, HANDLE Root);
NTSTATUS NTAPI RegfsOpenKey2(PHANDLE KeyHandle, ACCESS_MASK DesiredAccess, PACCESS_MASK GrantedAccess, KPROCESSOR_MODE RequestMode, IN PSECURITY_SUBJECT_CONTEXT SecurityContext OPTIONAL, PWSTR NameBuf, USHORT NameSize, ULONG ExtraAttributes, HANDLE Root);
NTSTATUS NTAPI RegfsManualKeySecurityCheck(HANDLE KeyHandle, PSECURITY_SUBJECT_CONTEXT Subject, ACCESS_MASK DesiredAccess, PACCESS_MASK GrantedAccess);
NTSTATUS NTAPI RegfsReplaceFileName(IN OUT PFILE_OBJECT File, IN PUNICODE_STRING BaseName, IN PUNICODE_STRING ExtraName OPTIONAL);
VOID RegfsUpdateFsRtlSizes(IN PFILE_OBJECT FileObject, IN ULONG ActualSize);

NTSTATUS NTAPI RegfsStartLocalBuffer(IN PIRP Irp, OUT PVOID *Buffer, OUT PULONG Length, OUT LOCK_OPERATION *Operation);
NTSTATUS NTAPI RegfsStartRemoteBuffer(IN PIRP Irp, OUT PVOID *Buffer, OUT PULONG Length, OUT LOCK_OPERATION *Operation);
VOID NTAPI RegfsEndBuffer(PIRP Irp, PVOID Addr);
ULONG NTAPI RegfsCopyData(PVOID Dest, void const *Src, ULONG SrcLength, ULONG UnusedBuffer, ULONG_PTR MaxBufferAddr);
NTSTATUS NTAPI RegfsExpandBufferIfNecessary(PVOID *Buffer, PULONG CurrentLength, ULONG RequiredLength, ULONG Tag);

NTSTATUS NTAPI RegfsInitFcbTable();
VOID NTAPI RegfsCleanupFcbTable();
NTSTATUS NTAPI RegfsDemandDfcb(OUT PREGFS_DFCB *DestDfcb, IN PUNICODE_STRING KeyName, IN PUNICODE_STRING ValueName OPTIONAL, IN HANDLE KeyHandle);
VOID NTAPI RegfsDereferenceDfcb(PREGFS_DFCB Dfcb);

PDRIVER_DISPATCH NTAPI RegfsInitDispatch(REGFS_DISPATCH_METHOD Method, PNTSTATUS Status);
VOID NTAPI RegfsDispatchMethodWorkThreadSystemContext(PREGFS_WORK_ITEM Wi);

NTSTATUS NTAPI RegfsInitWorkThreads(UCHAR CreateThreadCount, UCHAR ConcurrencyCount);
VOID NTAPI RegfsCleanupWorkThreads();
NTSTATUS NTAPI RegfsInitExWorkItems(ULONG Limit);
VOID NTAPI RegfsCleanupExWorkItems();
NTSTATUS NTAPI RegfsQueueWorkItem(PREGFS_WORK_ITEM CopyWi);
NTSTATUS NTAPI RegfsQueueExWorkItem(PREGFS_WORK_ITEM CopyWi);

NTSTATUS NTAPI RegfsKeyTreeOp(IN REGFS_KEY_TREE_OP Op, IN HANDLE BaseKey, IN HANDLE DestRootKey OPTIONAL, IN PUNICODE_STRING DestName OPTIONAL, IN KPROCESSOR_MODE RequestorMode, IN PSECURITY_SUBJECT_CONTEXT SecurityContext OPTIONAL);

#include "inline.h"

#endif //__DRIVERS_FS_REGFS_H