/*
 * COPYRIGHT:  2007 Christopher Smith
 * LICENCE:    GPL v2: See COPYING in the top level directory
 * PROJECT:    regfs
 * FILE:       rw.c
 * PURPOSE:    Registry filesystem read/write
 */

#include "regfs.h"

NTSTATUS NTAPI RegfsReadWrite(IN PDEVICE_OBJECT DeviceObject, IN PIRP Irp, IN OUT PVOID IoBuffer, IN ULONG IoBufferLength, IN PSECURITY_SUBJECT_CONTEXT SecurityContext OPTIONAL)
{
	PIO_STACK_LOCATION IrpSp;
	PFILE_OBJECT FileObject;
	NTSTATUS Status;
	HANDLE KeyHandle;
	KEY_VALUE_PARTIAL_INFORMATION ValueDataInfo;
	ULONG Offset;
	PKEY_VALUE_PARTIAL_INFORMATION Buffer;
	ULONG BufferLength;
	ULONG ResultSize;
	ULONG ValueDataSize;
	UNICODE_STRING ValueName;
	PREGFS_DFCB Dfcb;

	IrpSp = (PIO_STACK_LOCATION)IoGetCurrentIrpStackLocation(Irp);
	FileObject = IrpSp->FileObject;
	
	DPRINT("RegfsReadWrite DeviceObject:%p, Irp:%p FileName:%wZ\n", DeviceObject, Irp, &FileObject->FileName);

	if(!FileObject->FsContext2)
		return RegfsCompleteStatus(Irp, FALSE, STATUS_INVALID_DEVICE_REQUEST, "Error: No IO on raw device supported");

	Dfcb = ((PREGFS_CCB)FileObject->FsContext2)->Dfcb;

	if(Dfcb->Flags & REGFS_DFCB_DIRECTORY)
		return RegfsCompleteStatus(Irp, FALSE, STATUS_INVALID_DEVICE_REQUEST, "Error: Read/Write on a directory unsupported.");

	KeyHandle = RegfsGetDfcbKeyHandle(Dfcb);

	ValueName = Dfcb->ShortName;

	//acquire the appropriate lock
	if(IrpSp->MajorFunction == IRP_MJ_READ)
		ExAcquireResourceSharedLite(Dfcb->Fcb.Lock, TRUE);
	else //IRP_MJ_WRITE
		ExAcquireResourceExclusiveLite(Dfcb->Fcb.Lock, TRUE);
	
	//key value size information, detect value deletion
	Status = ZwQueryValueKey(KeyHandle, &ValueName, KeyValuePartialInformation, &ValueDataInfo, sizeof ValueDataInfo, &BufferLength);
	switch(Status)
	{
	case STATUS_OBJECT_NAME_NOT_FOUND:
		ExReleaseResourceLite(Dfcb->Fcb.Lock);
		return RegfsCompleteStatus(Irp, TRUE, STATUS_FILE_DELETED, "Error: Value deleted.");
	case STATUS_BUFFER_OVERFLOW:
		break; //expected warning
	default:
		if(!NT_SUCCESS(Status))
		{
			ExReleaseResourceLite(Dfcb->Fcb.Lock);
			return RegfsCompleteKeyStatus(Irp, TRUE, Status, "Error: QueryValueKey for data size failed.");
		}
	}

	ValueDataSize = ValueDataInfo.DataLength;

	//detirmine offset
	if(IrpSp->Parameters.Read.ByteOffset.LowPart == FILE_USE_FILE_POINTER_POSITION && IrpSp->Parameters.Read.ByteOffset.HighPart == -1)
		Offset = FileObject->CurrentByteOffset.LowPart;
	else if(IrpSp->Parameters.Read.ByteOffset.LowPart == FILE_WRITE_TO_END_OF_FILE && IrpSp->Parameters.Read.ByteOffset.HighPart == -1)
		Offset = ValueDataSize;
	else if(IrpSp->Parameters.Read.ByteOffset.HighPart)
	{
		ExReleaseResourceLite(Dfcb->Fcb.Lock);
		return RegfsCompleteStatus(Irp, FALSE, STATUS_END_OF_FILE, "Error: Offset too big.");
	}
	else
		Offset = IrpSp->Parameters.Read.ByteOffset.LowPart;

	DPRINT("Buffer size:%x ValueDataSize:%x SizeofHPD:%x\n", BufferLength, ValueDataSize, SIZEOF_KEY_VALUE_PARTIAL_INFORMATION_HEADER);
	
	if(IrpSp->MajorFunction == IRP_MJ_READ)
	{
		DPRINT_V("Offset:%x ValueDataSize:%x Read.Length:%x BufferLength:%x\n", Offset, ValueDataSize, IrpSp->Parameters.Read.Length, BufferLength);

		//offset past EOF?
		if(Offset >= ValueDataSize)
		{
			Irp->IoStatus.Information = 0;
			ExReleaseResourceLite(Dfcb->Fcb.Lock);
			return RegfsCompleteStatus(Irp, TRUE, STATUS_END_OF_FILE, "Error: Read entirely past end of file.");
		}

		//IoBufferLength = smaller of amount of data past offset and requested length
		IoBufferLength = min(ValueDataSize - Offset, IoBufferLength);

		if(IoBufferLength == 0)
		{
			Irp->IoStatus.Information = 0;
			ExReleaseResourceLite(Dfcb->Fcb.Lock);
			return RegfsCompleteStatus(Irp, TRUE, STATUS_SUCCESS, "Zero read.");
		}

		Buffer = (PKEY_VALUE_PARTIAL_INFORMATION)ExAllocatePoolWithTag(PagedPool, BufferLength, TAG_READ_BUFFER);
		if(!Buffer)
		{
			ExReleaseResourceLite(Dfcb->Fcb.Lock);
			return RegfsCompleteStatus(Irp, TRUE, STATUS_NO_MEMORY, "Error: Out of memory!");
		}

		//Read value into our buffer
		Status = ZwQueryValueKey(KeyHandle, &ValueName, KeyValuePartialInformation, Buffer, BufferLength, &BufferLength);
		if(!NT_SUCCESS(Status))
		{
			ExReleaseResourceLite(Dfcb->Fcb.Lock);
			ExFreePoolWithTag(Buffer, TAG_READ_BUFFER);
			return RegfsCompleteKeyStatus(Irp, TRUE, Status, "Error: QueryValueKey for data failed.");
		}
		
		//copy relevant buffer to user buffer
		DPRINT("Reading from %p to user buffer %p, length %x\n", Buffer, IoBuffer, IoBufferLength);
		Status = STATUS_SUCCESS;

		try
		{
			RtlCopyMemory(IoBuffer, Buffer->Data + Offset, IoBufferLength);
		}
		except(EXCEPTION_EXECUTE_HANDLER)
		{
			Status = GetExceptionCode();
		}

		ExFreePoolWithTag(Buffer, TAG_READ_BUFFER);
		if(!NT_SUCCESS(Status))
		{
			ExReleaseResourceLite(Dfcb->Fcb.Lock);
			DbgBreakPointWithStatus(Status);
			return RegfsCompleteStatus(Irp, TRUE, Status, "Error: Exception during copy.");
		}

		if(Offset + IoBufferLength > ValueDataSize)
			Status = STATUS_END_OF_FILE;
		else
			Status = STATUS_SUCCESS;
	}
	else //must be write
	{
		if(IoBufferLength == 0)
		{
			Irp->IoStatus.Information = 0;
			ExReleaseResourceLite(Dfcb->Fcb.Lock);
			return RegfsCompleteStatus(Irp, TRUE, STATUS_SUCCESS, "Zero write.");
		}
		
		ValueDataSize = max(ValueDataSize, Offset + IoBufferLength); //max of old size + new size
		BufferLength = SIZEOF_KEY_VALUE_PARTIAL_INFORMATION_HEADER + ValueDataSize;

		DPRINT_V("Offset:%x ValueDataSize:%x Read.Length:%x BufferLength:%x IoBufferLength:%x\n", Offset, ValueDataSize, IrpSp->Parameters.Read.Length, BufferLength, IoBufferLength);

		Buffer = ExAllocatePoolWithTag(PagedPool, BufferLength, TAG_WRITE_BUFFER);
		if(!Buffer)
		{
			ExReleaseResourceLite(Dfcb->Fcb.Lock);
			return RegfsCompleteStatus(Irp, TRUE, STATUS_NO_MEMORY, "Error: Out of memory!");
		}

		//get a copy of the old data (to merge with the new)
		Status = ZwQueryValueKey(KeyHandle, &ValueName, KeyValuePartialInformation, Buffer, BufferLength, &ResultSize);
		if(!NT_SUCCESS(Status))
		{
			ExFreePoolWithTag(Buffer, TAG_WRITE_BUFFER);
			ExReleaseResourceLite(Dfcb->Fcb.Lock);
			return RegfsCompleteKeyStatus(Irp, TRUE, Status, "Error: QueryValueKey for data failed.");
		}

		//merge write data into buffer
		Status = STATUS_SUCCESS;

		try
		{
			RtlCopyMemory(Buffer->Data + Offset, IoBuffer, IoBufferLength);
		}
		except(EXCEPTION_EXECUTE_HANDLER)
		{
			Status = GetExceptionCode();
		}

		if(!NT_SUCCESS(Status))
		{
			ExFreePoolWithTag(Buffer, TAG_WRITE_BUFFER);
			ExReleaseResourceLite(Dfcb->Fcb.Lock);
			DbgBreakPointWithStatus(Status);
			return RegfsCompleteStatus(Irp, TRUE, Status, "Error: Exception during copy");
		}

		//set new value
		Status = ZwSetValueKey(KeyHandle, &ValueName, Buffer->TitleIndex, Buffer->Type, Buffer->Data, ValueDataSize);
		ExFreePoolWithTag(Buffer, TAG_WRITE_BUFFER);		
		if(!NT_SUCCESS(Status) && Status != STATUS_BUFFER_OVERFLOW)
		{
			ExReleaseResourceLite(Dfcb->Fcb.Lock);
			return RegfsCompleteKeyStatus(Irp, TRUE, Status, "Error: SetValueKey for data failed.");
		}
	}
	
	RegfsUpdateFsRtlSizes(FileObject, ValueDataSize);

	ExReleaseResourceLite(Dfcb->Fcb.Lock);

	Irp->IoStatus.Information = IoBufferLength;

	if(FileObject->Flags & FO_SYNCHRONOUS_IO)
		InterlockedExchangeAdd(&FileObject->CurrentByteOffset.LowPart, Irp->IoStatus.Information);
	
	return RegfsCompleteStatus(Irp, TRUE, Status, "Success.");
}