/*
 * COPYRIGHT:  2007 Christopher Smith
 * LICENCE:    GPL v2: See COPYING in the top level directory
 * PROJECT:    regfs
 * FILE:       secinf.c
 * PURPOSE:    Registry filesystem query/set security information
 */

#include "regfs.h"

NTSTATUS NTAPI RegfsQuerySecurity(IN PDEVICE_OBJECT DeviceObject, IN PIRP Irp, IN OUT PVOID IoBuffer, IN ULONG IoBufferLength, IN PSECURITY_SUBJECT_CONTEXT SecurityContext OPTIONAL)
{
	PIO_STACK_LOCATION IrpSp;
	PFILE_OBJECT FileObject;
	NTSTATUS Status;
	HANDLE KeyHandle;
	ULONG_PTR MaxBufferAddr;
	PSECURITY_DESCRIPTOR SD;
	ULONG SDLength;

	IrpSp = (PIO_STACK_LOCATION)IoGetCurrentIrpStackLocation(Irp);
	FileObject = IrpSp->FileObject;

	if(!FileObject->FsContext2)
		return RegfsCompleteStatus(Irp, FALSE, STATUS_INVALID_DEVICE_REQUEST, "Error: QuerySecurity attempted on raw volume file.");

	KeyHandle = RegfsGetDfcbKeyHandle(((PREGFS_CCB)FileObject->FsContext2)->Dfcb);

	MaxBufferAddr = (ULONG_PTR)IoBuffer + IoBufferLength;

	DPRINT("RegfsQuerySecurity DeviceObject:%p, Irp:%p FileName:%wZ\n", DeviceObject, Irp, &FileObject->FileName);

	DPRINT_V("Setting MaxBufferAddr %p + %X = %p\n", IoBuffer, IoBufferLength, MaxBufferAddr);

	Status = ZwQuerySecurityObject(KeyHandle, IrpSp->Parameters.SetSecurity.SecurityInformation, NULL, 0, &SDLength);

	if(Status != STATUS_BUFFER_TOO_SMALL && !NT_SUCCESS(Status))
		return RegfsCompleteStatus(Irp, FALSE, Status, "Error querying security size.");

	SD = ExAllocatePoolWithTag(PagedPool, SDLength, TAG_SECURITY_DESC_QUERY);

	if(!SD)
		return RegfsCompleteStatus(Irp, FALSE, STATUS_NO_MEMORY, "Out of memory!");

	Status = ZwQuerySecurityObject(KeyHandle, IrpSp->Parameters.SetSecurity.SecurityInformation, SD, SDLength, PAGED_BIT_BUCKET);

	if(!NT_SUCCESS(Status))
	{
		ExFreePoolWithTag(SD, TAG_SECURITY_DESC_QUERY);
		return RegfsCompleteStatus(Irp, TRUE, Status, "Error querying security."); 
	}

	Status = RegfsKeySDToFileSD(SD);

	if(!NT_SUCCESS(Status))
	{
		ExFreePoolWithTag(SD, TAG_SECURITY_DESC_QUERY);
		return RegfsCompleteStatus(Irp, FALSE, Status, "Error converting security descriptor.");
	}

	Irp->IoStatus.Information = SDLength;

	try
	{
		RegfsCopyData(IoBuffer, SD, SDLength, IoBufferLength, MaxBufferAddr);
	}
	except(EXCEPTION_EXECUTE_HANDLER)
	{
		Status = GetExceptionCode();
	}

	ExFreePoolWithTag(SD, TAG_SECURITY_DESC_QUERY);

	if(!NT_SUCCESS(Status))
		return RegfsCompleteStatus(Irp, TRUE, Status, "Error copying buffer.");

	if(SDLength > IoBufferLength)
		return RegfsCompleteStatus(Irp, TRUE, STATUS_BUFFER_OVERFLOW, "Buffer overflow.");
	else
		return RegfsCompleteStatus(Irp, TRUE, STATUS_SUCCESS, "Success");
}

NTSTATUS NTAPI RegfsSetSecurity(IN PDEVICE_OBJECT DeviceObject, IN PIRP Irp, IN OUT PVOID IoBuffer, IN ULONG IoBufferLength, IN PSECURITY_SUBJECT_CONTEXT SecurityContext OPTIONAL)
{
	PIO_STACK_LOCATION IrpSp;
	PFILE_OBJECT FileObject;
	PREGFS_DFCB Dfcb;
	NTSTATUS Status;
	HANDLE KeyHandle;

	IrpSp = (PIO_STACK_LOCATION)IoGetCurrentIrpStackLocation(Irp);
	FileObject = IrpSp->FileObject;

	if(!FileObject->FsContext2)
		return RegfsCompleteStatus(Irp, FALSE, STATUS_INVALID_DEVICE_REQUEST, "Error: SetSecurity attempted on raw volume file.");

	Dfcb = ((PREGFS_CCB)FileObject->FsContext2)->Dfcb;

	if(!(Dfcb->Flags & REGFS_DFCB_DIRECTORY))
		return RegfsCompleteStatus(Irp, FALSE, STATUS_SUCCESS, "Warning: SetSecurity attempt on a key value ignored.");

	KeyHandle = RegfsGetDfcbKeyHandle(Dfcb);

	DPRINT("RegfsSetSecurity DeviceObject:%p, Irp:%p FileName:%wZ\n", DeviceObject, Irp, &Dfcb->FileName);

	Status = RegfsFileSDToKeySD(IrpSp->Parameters.SetSecurity.SecurityDescriptor);

	if(!NT_SUCCESS(Status))
		return RegfsCompleteStatus(Irp, FALSE, Status, "Error converting security descriptor.");

	Status = ZwSetSecurityObject(KeyHandle, IrpSp->Parameters.SetSecurity.SecurityInformation, IrpSp->Parameters.SetSecurity.SecurityDescriptor);

	if(!NT_SUCCESS(Status))
		return RegfsCompleteStatus(Irp, TRUE, Status, "Error setting key security.");

	return RegfsCompleteStatus(Irp, TRUE, STATUS_SUCCESS, "Success.");
}