/*
 * COPYRIGHT:  2007 Christopher Smith
 * LICENCE:    GPL v2: See COPYING in the top level directory
 * PROJECT:    regfs
 * FILE:       setinf.c
 * PURPOSE:    Registry filesystem set information
 */

#include "regfs.h"

PWCHAR const SYMLINK_VALUE_NAME = L"SymbolicLinkValue";
USHORT const SYMLINK_VALUE_NAME_LENGTH = 17 * sizeof(WCHAR);

typedef union _FILE_INFO
{
	FILE_ALLOCATION_INFORMATION FileAllocationInformation;
	FILE_BASIC_INFORMATION FileBasicInformation;
	FILE_DISPOSITION_INFORMATION FileDispositionInformation;
	FILE_END_OF_FILE_INFORMATION FileEndOfFileInformation;
	FILE_LINK_INFORMATION FileLinkInformation;
	FILE_POSITION_INFORMATION FilePositionInformation;
	FILE_RENAME_INFORMATION FileRenameInformation;
	FILE_VALID_DATA_LENGTH_INFORMATION FileValidDataLengthInformation;
} FILE_INFO, *PFILE_INFO;

//Dfcb must be a FCB
NTSTATUS NTAPI RegfsSetEof(PIRP Irp, PREGFS_DFCB Dfcb, HANDLE KeyHandle, LARGE_INTEGER NewSize)
{
	NTSTATUS Status;
	PKEY_VALUE_PARTIAL_INFORMATION Buffer;
	ULONG BufferLength = SIZEOF_KEY_VALUE_PARTIAL_INFORMATION_HEADER + NewSize.LowPart;

	if(NewSize.HighPart)
		return RegfsCompleteStatus(Irp, FALSE, STATUS_INVALID_DEVICE_REQUEST, "File size too large!");

	if(!MmCanFileBeTruncated(&Dfcb->Fcb.FcbFsRtl->SecPtrs, &NewSize))
		return RegfsCompleteStatus(Irp, FALSE, STATUS_UNSUCCESSFUL, "Mm says we can't truncate the file to that size.");
	
	Buffer = ExAllocatePoolWithTag(PagedPool, BufferLength, TAG_BLANK_BUFFER);
	if(!Buffer)
		return RegfsCompleteStatus(Irp, FALSE, STATUS_NO_MEMORY, "Error: Out of memory!");
	RtlZeroMemory(Buffer, BufferLength);
	
	ExAcquireResourceExclusiveLite(Dfcb->Fcb.Lock, TRUE);
	
	Status = ZwQueryValueKey(KeyHandle, &Dfcb->ShortName, KeyValuePartialInformation, Buffer, BufferLength, PAGED_BIT_BUCKET);
	if(!NT_SUCCESS(Status) && Status != STATUS_BUFFER_OVERFLOW) //if data is being truncated, then that's what the caller wanted
	{
		ExFreePoolWithTag(Buffer, TAG_BLANK_BUFFER);
		ExReleaseResourceLite(Dfcb->Fcb.Lock);
		return RegfsCompleteKeyStatus(Irp, TRUE, Status, "Error during QueryValue.");
	}
	
	Status = ZwSetValueKey(KeyHandle, &Dfcb->ShortName, Buffer->TitleIndex, Buffer->Type, Buffer->Data, NewSize.LowPart);
	ExFreePoolWithTag(Buffer, TAG_BLANK_BUFFER);

	RegfsUpdateFsRtlSizes(IoGetCurrentIrpStackLocation(Irp)->FileObject, NewSize.LowPart);

	ExReleaseResourceLite(Dfcb->Fcb.Lock);
	return RegfsCompleteKeyStatus(Irp, TRUE, Status, NULL);
}

NTSTATUS NTAPI RegfsMoveValue(HANDLE SrcKey, PUNICODE_STRING SrcName, HANDLE DestKey, PUNICODE_STRING DestName, BOOLEAN Delete)
{
	PKEY_VALUE_PARTIAL_INFORMATION Buffer = NULL;
	NTSTATUS Status;
	ULONG BufferLength = 0, ReqLength;

	ReqLength = 0x80; //initial guess
	while(TRUE)
	{
		Status = RegfsExpandBufferIfNecessary(&Buffer, &BufferLength, ReqLength, TAG_MOVE_BUFFER);

		if(!NT_SUCCESS(Status))
			return Status;

		Status = ZwQueryValueKey(SrcKey, SrcName, KeyValuePartialInformation, Buffer, BufferLength, &ReqLength);
		
		if(IsBufferSizeStatus(Status))
			continue;

		DPRINT("RegfsMoveValue: ZwQueryValueKey non-buffer status:%X", Status);

		if(!NT_SUCCESS(Status))
			return Status;

		break;
	}
	
	Status = ZwSetValueKey(DestKey, DestName, Buffer->TitleIndex, Buffer->Type, Buffer->Data, Buffer->DataLength);

	ExFreePoolWithTag(Buffer, TAG_MOVE_BUFFER);
	
	if(!NT_SUCCESS(Status))
		return RegfsKeyStatusToFileStatus(Status);

	if(Delete)
		Status = ZwDeleteValueKey(SrcKey, SrcName);

	return RegfsKeyStatusToFileStatus(Status);
}

NTSTATUS NTAPI DeleteTreeKeyByName(IN HANDLE Root, IN UNICODE_STRING Name, KPROCESSOR_MODE RequestorMode, IN PSECURITY_SUBJECT_CONTEXT SecurityContext OPTIONAL)
{
	HANDLE Key;
	NTSTATUS Status;
	
	Status = RegfsOpenKey(&Key, DELETE, PAGED_BIT_BUCKET, RequestorMode, SecurityContext, &Name, 0, Root);

	if(!NT_SUCCESS(Status))
		return Status;

	Status = RegfsKeyTreeOp(KtoDelete, Key, NULL, NULL, RequestorMode, SecurityContext);

	ZwDeleteKey(Key);

	ZwClose(Key);

	return Status;
}

NTSTATUS NTAPI RegfsProcessMove(PIRP Irp, PIO_STACK_LOCATION IrpSp, PREGFS_DFCB Dfcb, IN PSECURITY_SUBJECT_CONTEXT SecurityContext OPTIONAL)
{
	NTSTATUS Status;
	UNICODE_STRING DestName, DestShortName, DestRootName;
	PFILE_OBJECT DestFile;
	PFILE_RENAME_INFORMATION RenInfo;

	RenInfo = (PFILE_RENAME_INFORMATION)Irp->AssociatedIrp.SystemBuffer;

	DestName.Length = DestName.MaximumLength = (USHORT)RenInfo->FileNameLength;
	DestName.Buffer = RenInfo->FileName;

	RegfsSeperateLastName(DestName, &DestShortName, &DestRootName);

	DestFile = IrpSp->Parameters.SetFile.FileObject;

	if(DestFile) //then full rename
	{
		PREGFS_CCB DestCcb = (PREGFS_CCB)IrpSp->Parameters.SetFile.FileObject->FsContext2;
		PREGFS_DFCB DestDfcb;
		
		if(!DestCcb)
			return RegfsCompleteStatus(Irp, FALSE, STATUS_INVALID_PARAMETER, "Can't rename to volume file!");
		
		DestDfcb = DestCcb->Dfcb;

		if(!(DestDfcb->Flags & REGFS_DFCB_DIRECTORY))
			return RegfsCompleteStatus(Irp, FALSE, STATUS_INVALID_PARAMETER, "SetFile.FileObject must be a directory!");

		if(Dfcb->Flags & REGFS_DFCB_DIRECTORY)
		{
			DPRINT("Moving directory from %wZ to %wZ (%wZ)\n", &Dfcb->FileName, &DestName, &DestShortName);

			if(RenInfo->ReplaceIfExists) //try to delete
				DeleteTreeKeyByName(DestDfcb->Dcb.KeyHandle, DestShortName, Irp->RequestorMode, SecurityContext);

			Status = RegfsKeyTreeOp(KtoCopy, Dfcb->Dcb.KeyHandle, DestDfcb->Dcb.KeyHandle, &DestShortName, Irp->RequestorMode, SecurityContext);

			if(!NT_SUCCESS(Status))
				return RegfsCompleteStatus(Irp, TRUE, Status, "Error copying key.");

			Status = RegfsKeyTreeOp(KtoDelete, Dfcb->Dcb.KeyHandle, NULL, NULL, Irp->RequestorMode, SecurityContext);
		}
		else
			Status = RegfsMoveValue(Dfcb->Fcb.Directory->Dcb.KeyHandle, &Dfcb->FileName, DestDfcb->Dcb.KeyHandle, &DestDfcb->FileName, TRUE);
	}
	else //short name rename
	{
		if(Dfcb->Flags & REGFS_DFCB_DIRECTORY)
		{
			HANDLE DestRootHandle;
			OBJECT_ATTRIBUTES DestRootOa;
			
			DPRINT("Short moving directory from %wZ to %wZ\n", &Dfcb->FileName, &DestName);

			RtlZeroMemory(&DestRootOa, sizeof DestRootOa);
			DestRootOa.Length = sizeof(DestRootOa);
			DestRootOa.ObjectName = &DestRootName;

			DPRINT("Opening %wZ as dest root\n", &DestRootName);

			Status = ZwOpenKey(&DestRootHandle, KEY_ALL_ACCESS, &DestRootOa);

			if(!NT_SUCCESS(Status))
				return RegfsCompleteKeyStatus(Irp, FALSE, Status, "Error opening common root key.");

			if(RenInfo->ReplaceIfExists) //then try to delete
				DeleteTreeKeyByName(DestRootHandle, DestName, Irp->RequestorMode, SecurityContext);

			Status = RegfsKeyTreeOp(KtoCopy, Dfcb->Dcb.KeyHandle, DestRootHandle, &DestName, Irp->RequestorMode, SecurityContext);

			ZwClose(DestRootHandle);

			if(!NT_SUCCESS(Status))
				return RegfsCompleteStatus(Irp, TRUE, Status, "Error copying key.");

			Status = RegfsKeyTreeOp(KtoDelete, Dfcb->Dcb.KeyHandle, NULL, NULL, Irp->RequestorMode, SecurityContext);
		}
		else
			Status = RegfsMoveValue(Dfcb->Fcb.Directory->Dcb.KeyHandle, &Dfcb->FileName, Dfcb->Fcb.Directory->Dcb.KeyHandle, &DestName, TRUE);
	}

	return RegfsCompleteStatus(Irp, TRUE, Status, "");
}

NTSTATUS NTAPI RegfsSetInformation(IN PDEVICE_OBJECT DeviceObject, IN PIRP Irp, IN OUT PVOID IoBuffer, IN ULONG IoBufferLength, IN PSECURITY_SUBJECT_CONTEXT SecurityContext OPTIONAL)
{
	NTSTATUS Status;
	PIO_STACK_LOCATION IrpSp;
	PFILE_OBJECT FileObject;
	PREGFS_DFCB Dfcb;
	HANDLE KeyHandle;
	PFILE_INFO FileInfo;

	IrpSp = (PIO_STACK_LOCATION)IoGetCurrentIrpStackLocation(Irp);
	FileObject = IrpSp->FileObject;

	if(!FileObject->FsContext2)
		return RegfsCompleteStatus(Irp, FALSE, STATUS_INVALID_DEVICE_REQUEST, "Error: SetInformationFile attempted on raw volume file.");

	FileInfo = IoBuffer;

	DPRINT("RegfsSetInformation DeviceObject:%p, Irp:%p FileName:%wZ\n", DeviceObject, Irp, &FileObject->FileName);

	Dfcb = ((PREGFS_CCB)FileObject->FsContext2)->Dfcb;

	if(!Dfcb)
		return RegfsCompleteStatus(Irp, FALSE, STATUS_INVALID_DEVICE_REQUEST, "Error: SetInformation attempted on raw volume file.");

	KeyHandle = RegfsGetDfcbKeyHandle(Dfcb);

	switch(IrpSp->Parameters.SetFile.FileInformationClass)
	{
	case FileAllocationInformation:
		DPRINT("File allocation information set to %x%x\n", FileInfo->FileAllocationInformation.AllocationSize.HighPart, FileInfo->FileAllocationInformation.AllocationSize.LowPart);
		//same as EOF, so fall through:
	case FileEndOfFileInformation:
		DPRINT("File EOF information set to %x%x\n", FileInfo->FileEndOfFileInformation.EndOfFile.HighPart, FileInfo->FileEndOfFileInformation.EndOfFile.HighPart);
		
		if(Dfcb->Flags & REGFS_DFCB_DIRECTORY)
			return RegfsCompleteStatus(Irp, FALSE, STATUS_FILE_IS_A_DIRECTORY, "Error: Can't set file size on a directory.");
		
		return RegfsSetEof(Irp, Dfcb, KeyHandle, FileInfo->FileEndOfFileInformation.EndOfFile);
	
	case FileBasicInformation:
		//if(!(Dfcb->Flags & REGFS_DFCB_DIRECTORY))
			return RegfsCompleteStatus(Irp, FALSE, STATUS_SUCCESS, NULL); //only keys have any times
		//{
		//	KEY_WRITE_TIME_INFORMATION KeyInfo;
		//	KeyInfo.LastWriteTime = FileInfo->FileBasicInformation.LastWriteTime;
		//	Status = ZwSetInformationKey(KeyHandle, KeyWriteTimeInformation, &KeyInfo, sizeof KeyInfo);
		//	return RegfsCompleteKeyStatus(Irp, TRUE, Status, NULL);
		//}

	case FileDispositionInformation:
		if(FileInfo->FileDispositionInformation.DeleteFile)
			if(Dfcb->Flags & REGFS_DFCB_DIRECTORY)
				Status = ZwDeleteKey(KeyHandle);
			else
				Status = ZwDeleteValueKey(KeyHandle, &Dfcb->ShortName);
		else
			Status = STATUS_SUCCESS;
		return RegfsCompleteKeyStatus(Irp, FALSE, Status, NULL);

	case FileLinkInformation:
		if(Dfcb->Flags & REGFS_DFCB_DIRECTORY)
		{
			HANDLE LinkHandle;
			UNICODE_STRING LinkName, LinkValueName;
			OBJECT_ATTRIBUTES LinkOa;

			LinkName.Buffer = FileInfo->FileLinkInformation.FileName;
			LinkName.Length = LinkName.MaximumLength = (USHORT)FileInfo->FileLinkInformation.FileNameLength;
			InitializeObjectAttributes(&LinkOa, &LinkName, OBJ_KERNEL_HANDLE | OBJ_FORCE_ACCESS_CHECK, FileInfo->FileLinkInformation.RootDirectory, NULL);

			Status = ZwCreateKey(&LinkHandle, KEY_ALL_ACCESS, &LinkOa, 0, NULL, REG_OPTION_CREATE_LINK, PAGED_BIT_BUCKET);
			if(!Status)
				return RegfsCompleteStatus(Irp, TRUE, Status, "Error creating link key.");

			LinkValueName.Buffer = SYMLINK_VALUE_NAME;
			LinkValueName.Length = LinkValueName.MaximumLength = SYMLINK_VALUE_NAME_LENGTH;

			Status = ZwSetValueKey(LinkHandle, &LinkValueName, 0, REG_LINK, Dfcb->FileName.Buffer, Dfcb->FileName.Length);
			ZwClose(LinkHandle);
			return RegfsCompleteStatus(Irp, TRUE, Status, NULL);
		}
		else
			return RegfsCompleteStatus(Irp, FALSE, STATUS_INVALID_DEVICE_REQUEST, "Links avaliable on keys only.");

	case FilePositionInformation:
		if(FileInfo->FilePositionInformation.CurrentByteOffset.HighPart)
			return RegfsCompleteStatus(Irp, FALSE, STATUS_INVALID_DEVICE_REQUEST, "File position too large.");
	
		if(IrpSp->FileObject->CurrentByteOffset.QuadPart < FileInfo->FilePositionInformation.CurrentByteOffset.QuadPart) //if file expansion needed
		{
			IrpSp->FileObject->CurrentByteOffset.QuadPart = FileInfo->FilePositionInformation.CurrentByteOffset.QuadPart;
			return RegfsSetEof(Irp, Dfcb, KeyHandle, FileInfo->FilePositionInformation.CurrentByteOffset);
		}
		else
		{
			IrpSp->FileObject->CurrentByteOffset.QuadPart = FileInfo->FilePositionInformation.CurrentByteOffset.QuadPart;
			return STATUS_SUCCESS;
		}
	case FileRenameInformation:
		return RegfsProcessMove(Irp, IrpSp, Dfcb, SecurityContext);

	case FileValidDataLengthInformation:
		return RegfsCompleteStatus(Irp, FALSE, STATUS_SUCCESS, NULL);

	default:
		return RegfsCompleteStatus(Irp, TRUE, STATUS_INVALID_INFO_CLASS, "Error: Unsupported information class.\n");
	}
}

NTSTATUS NTAPI RegfsFlush(IN PDEVICE_OBJECT DeviceObject, IN PIRP Irp, IN OUT PVOID IoBuffer, IN ULONG IoBufferLength, IN PSECURITY_SUBJECT_CONTEXT SecurityContext OPTIONAL)
{	
	PIO_STACK_LOCATION IrpSp;
	PFILE_OBJECT FileObject;
	HANDLE KeyHandle;
	NTSTATUS Status;

	DPRINT("RegfsFlush DeviceObject:%p, Irp:%p", DeviceObject, Irp);

	IrpSp = (PIO_STACK_LOCATION)IoGetCurrentIrpStackLocation(Irp);
	FileObject = IrpSp->FileObject;

	if(!FileObject->FsContext2)
		return RegfsCompleteStatus(Irp, FALSE, STATUS_SUCCESS, "Raw device: nothing to flush.");

	KeyHandle = RegfsGetDfcbKeyHandle(((PREGFS_CCB)FileObject->FsContext2)->Dfcb);

	Status = ZwFlushKey(KeyHandle);

	return RegfsCompleteStatus(Irp, TRUE, Status, NULL);
}