/*
 * COPYRIGHT:  2007 Christopher Smith
 * LICENCE:    GPL v2: See COPYING in the top level directory
 * PROJECT:    regfs
 * FILE:       support.c
 * PURPOSE:    Registry filesystem basic registry/filesystem transformations and misc support
 */

#include "regfs.h"

typedef ACCESS_MASK (NTAPI *PMASK_CONVERTER)(IN ACCESS_MASK Mask);

ACCESS_MASK NTAPI RegfsFileMaskToKeyMask(ACCESS_MASK FileMask)
{
	ACCESS_MASK KeyMask = FileMask & COMMON_ACCESS_MASK; 

	if(FileMask & (FILE_READ_ATTRIBUTES | FILE_READ_EA | FILE_TRAVERSE))
		KeyMask |= KEY_QUERY_VALUE;

	if(FileMask & (FILE_ADD_FILE | FILE_DELETE_CHILD | FILE_WRITE_ATTRIBUTES | FILE_WRITE_EA))
		KeyMask |= KEY_SET_VALUE;

	if(FileMask & (FILE_LIST_DIRECTORY))
		KeyMask |= KEY_QUERY_VALUE | KEY_ENUMERATE_SUB_KEYS;

	if(FileMask & (FILE_ADD_SUBDIRECTORY))
		KeyMask |= KEY_CREATE_SUB_KEY;

	DPRINT_V("RefgsFileMaskToKeyMask FileMask:%X KeyMask:%X\n", FileMask, KeyMask);

	return KeyMask;
}

ACCESS_MASK NTAPI RegfsFileMaskToKeyValueMask(ACCESS_MASK FileMask)
{
	ACCESS_MASK KeyMask = FileMask & COMMON_ACCESS_MASK; 

	if(FileMask & (FILE_READ_DATA | FILE_READ_ATTRIBUTES | FILE_READ_EA | FILE_EXECUTE))
		KeyMask |= KEY_QUERY_VALUE;

	if(FileMask & (FILE_WRITE_DATA | FILE_WRITE_ATTRIBUTES | FILE_WRITE_EA | FILE_APPEND_DATA))
		KeyMask |= KEY_SET_VALUE;

	return KeyMask;
}

ACCESS_MASK NTAPI RegfsKeyMaskToFileMask(ACCESS_MASK KeyMask)
{
	ACCESS_MASK FileMask = KeyMask & COMMON_ACCESS_MASK;

	FileMask |= FILE_TRAVERSE;

	if(KeyMask & (KEY_QUERY_VALUE))
		FileMask |= FILE_LIST_DIRECTORY | FILE_READ_ATTRIBUTES | FILE_READ_EA | FILE_LIST_DIRECTORY;

	if(KeyMask & (KEY_ENUMERATE_SUB_KEYS))
		FileMask |= FILE_LIST_DIRECTORY;

	if(KeyMask & (KEY_SET_VALUE))
		FileMask |= FILE_ADD_FILE | FILE_DELETE_CHILD | FILE_WRITE_ATTRIBUTES | FILE_WRITE_EA;

	if(KeyMask & (KEY_CREATE_SUB_KEY))
		FileMask |= FILE_ADD_SUBDIRECTORY;

	return FileMask;
}

ACCESS_MASK NTAPI RegfsKeyValueMaskToFileMask(ACCESS_MASK KeyMask)
{
	ACCESS_MASK FileMask = KeyMask & COMMON_ACCESS_MASK;

	if(KeyMask & (KEY_QUERY_VALUE))
		FileMask |= FILE_READ_DATA | FILE_READ_EA | FILE_EXECUTE | FILE_READ_ATTRIBUTES;

	if(KeyMask & (KEY_SET_VALUE))
		FileMask |= FILE_WRITE_DATA | FILE_WRITE_EA | FILE_APPEND_DATA | FILE_WRITE_ATTRIBUTES;

	return FileMask;
}

NTSTATUS NTAPI ConvertAcl(IN OUT PACL Acl, IN PMASK_CONVERTER ConvertMask)
{
	USHORT i;
	PACCESS_ALLOWED_ACE Ace; //all the supported ACE types have the same layout

	(ULONG_PTR)Ace = (ULONG_PTR)Acl + sizeof(ACL);

	if(Acl->AclRevision > ACL_REVISION2)
		return STATUS_INVALID_ACL;

	for(i = 0; i < Acl->AceCount; i++)
	{
		Ace->Mask = ConvertMask(Ace->Mask);
		(ULONG_PTR)Ace += Ace->Header.AceSize;
	}

	return STATUS_SUCCESS;
}

NTSTATUS NTAPI ConvertSD(IN OUT struct _SECURITY_DESCRIPTOR *SD, IN PMASK_CONVERTER ConvertMask)
{
	NTSTATUS Status = STATUS_SUCCESS;

	if(SD->Control & SE_SELF_RELATIVE)
	{
		struct _SECURITY_DESCRIPTOR_RELATIVE *SDRel = (struct _SECURITY_DESCRIPTOR_RELATIVE *)SD;

		if(SD->Control & SE_DACL_PRESENT)
			Status = ConvertAcl((PACL)((ULONG_PTR)SDRel + SDRel->Dacl), ConvertMask);

		if(!NT_SUCCESS(Status))
			return Status;

		if(SD->Control & SE_SACL_PRESENT)
			Status = ConvertAcl((PACL)((ULONG_PTR)SDRel + SDRel->Sacl), ConvertMask);
	}
	else
	{
		if(SD->Control & SE_DACL_PRESENT)
			Status = ConvertAcl(SD->Dacl, ConvertMask);

		if(!NT_SUCCESS(Status))
			return Status;

		if(SD->Control & SE_SACL_PRESENT)
			Status = ConvertAcl(SD->Sacl, ConvertMask);
	}

	return Status;
}

NTSTATUS NTAPI RegfsKeySDToFileSD(PSECURITY_DESCRIPTOR KeySD)
{
	return ConvertSD(KeySD, &RegfsKeyMaskToFileMask);
}

NTSTATUS NTAPI RegfsFileSDToKeySD(PSECURITY_DESCRIPTOR FileSD)
{
	return ConvertSD(FileSD, &RegfsFileMaskToKeyMask);
}

BOOLEAN NTAPI RegfsHasSeperator(UNICODE_STRING Name)
{
	USHORT SepPos = Name.Length / sizeof(WCHAR);

	if(SepPos && SepPos == L'\\')//ignore trailing backslash
		SepPos--;

	for(; SepPos; SepPos--)
		if(Name.Buffer[SepPos] == L'\\')
			return TRUE;

	return FALSE;
}

//similar to FsRtlDissectName
VOID NTAPI RegfsSeperateLastName(UNICODE_STRING CONST Src, OUT UNICODE_STRING *Last OPTIONAL, OUT UNICODE_STRING *Rest OPTIONAL)
{
	BOOLEAN HasSep = TRUE;
	USHORT SepPos = Src.Length / sizeof(WCHAR) - 1; //start at last char
	if(SepPos && SepPos == L'\\')//ignore trailing backslash
		SepPos--;
	
	while(Src.Buffer[SepPos] != L'\\')
	{
		if(!SepPos)
		{
			HasSep = FALSE;
			break;
		}
		SepPos--;
	}

	if(Last)
	{
		*Last = Src;

		if(HasSep)
		{
			USHORT LastStart = SepPos + 1;
			LastStart = min(LastStart, Last->Length * sizeof(WCHAR));
			Last->Buffer += LastStart;
			Last->Length -= LastStart * sizeof(WCHAR);
			Last->MaximumLength = Last->Length;
		}
	}

	if(Rest)
	{
		if(HasSep)
		{
			*Rest = Src;
			Rest->Length = Rest->MaximumLength = SepPos * sizeof(WCHAR);
		}
		else
		{
			Rest->Length = Rest->MaximumLength = 0;
			Rest->Buffer = NULL;
		}
	}

	DPRINT_V("RegfsSeperateLastName SepPos:%x HasSep:%x Src:%wZ Last:%wZ Rest:%wZ\n", SepPos, HasSep, &Src, Last, Rest);
}

NTSTATUS NTAPI RegfsKeyStatusToFileStatus(NTSTATUS KeyStatus)
{
	switch(KeyStatus)
	{
	case STATUS_KEY_DELETED:
		return STATUS_FILE_DELETED;
	case STATUS_NO_MORE_ENTRIES:
		return STATUS_NO_MORE_FILES;
	case STATUS_REGISTRY_IO_FAILED:
		return STATUS_DEVICE_DATA_ERROR;
	case STATUS_SYSTEM_HIVE_TOO_LARGE:
		return STATUS_DISK_FULL;
	default:
		return KeyStatus;
	}
}

//it's possible that the CM will return some of these values (at least HKEY_PERFORMANCE_TEXT),
//but they are Win32 pseudohandles, not valid kernel handles. See KB890648
BOOLEAN NTAPI RegfsIsKeyHandleBogus(HANDLE Handle)
{
	switch((ULONG)Handle)
	{
		case 0x80000000: //HKEY_CLASSES_ROOT
		case 0x80000001: //HKEY_CURRENT_USER
		case 0x80000002: //HKEY_LOCAL_MACHINE
		case 0x80000003: //HKEY_USERS
		case 0x80000004: //HKEY_PERFORMANCE_DATA
		case 0x80000005: //HKEY_CURRENT_CONFIG
		case 0x80000006: //HKEY_DYN_DATA
		case 0x80000050: //HKEY_PERFORMANCE_TEXT
		case 0x80000060: //HKEY_PERFORMANCE_NLSTEXT
			return TRUE;
		default:
			return FALSE;
	}
}

NTSTATUS NTAPI RegfsManualKeySecurityCheck(HANDLE KeyHandle, PSECURITY_SUBJECT_CONTEXT Subject, ACCESS_MASK DesiredAccess, PACCESS_MASK GrantedAccess)
{
	NTSTATUS Status;
	ULONG BufferLength;
	PSECURITY_DESCRIPTOR Buffer;

	BufferLength = 0;

	if(DesiredAccess == 0)
	{
		*GrantedAccess = 0;
		return STATUS_SUCCESS;
	}

	Status = ZwQuerySecurityObject(KeyHandle, DACL_SECURITY_INFORMATION | SACL_SECURITY_INFORMATION | OWNER_SECURITY_INFORMATION | GROUP_SECURITY_INFORMATION, NULL, 0, &BufferLength);

	DPRINT_V("ZwQuerySecurityObject KeyHandle:%p Status:%x ReqBuffer:%x\n", KeyHandle, Status, BufferLength);

	if(Status != STATUS_BUFFER_TOO_SMALL && !NT_SUCCESS(Status))
		return Status;

	Buffer = ExAllocatePoolWithTag(PagedPool, BufferLength, TAG_SECURITY_DESC_CHECK);

	if(!Buffer)
		return STATUS_NO_MEMORY;

	Status = ZwQuerySecurityObject(KeyHandle, DACL_SECURITY_INFORMATION | SACL_SECURITY_INFORMATION | OWNER_SECURITY_INFORMATION | GROUP_SECURITY_INFORMATION, Buffer, BufferLength, PAGED_BIT_BUCKET);

	DPRINT_V("ZwQuerySecurityObject KeyHandle:%p Status:%x\n", KeyHandle, Status);

	if(!NT_SUCCESS(Status))
	{
		ExFreePoolWithTag(Buffer, TAG_SECURITY_DESC_CHECK);
		return Status;
	}

	if(SeAccessCheck(Buffer, Subject, FALSE, DesiredAccess, 0, NULL, &KeyGenericMapping, UserMode, GrantedAccess, &Status))
		Status = STATUS_SUCCESS;

	ExFreePoolWithTag(Buffer, TAG_SECURITY_DESC_CHECK);

	return Status;
}

//must be called with either ExtraAttributes & OBJ_KERNEL_HANDLE or in system context
NTSTATUS NTAPI RegfsOpenKey(PHANDLE KeyHandle, ACCESS_MASK DesiredAccess, PACCESS_MASK GrantedAccess, KPROCESSOR_MODE RequestMode, IN PSECURITY_SUBJECT_CONTEXT SecurityContext OPTIONAL, PUNICODE_STRING Name, ULONG ExtraAttributes, HANDLE Root)
{
	NTSTATUS Status;
	OBJECT_ATTRIBUTES Oa;

	InitializeObjectAttributes(&Oa, Name, ExtraAttributes | DefaultObjectAttributes, Root, NULL);

	Status = ZwOpenKey(KeyHandle, KEY_ALL_ACCESS, &Oa);

	if(!NT_SUCCESS(Status))
		return Status;

	if(RegfsIsKeyHandleBogus(*KeyHandle))
		return STATUS_ACCESS_DENIED; //i.e. not accessible from KM

	if(RequestMode == UserMode)
	{
		Status = RegfsManualKeySecurityCheck(*KeyHandle, SecurityContext, DesiredAccess, GrantedAccess);
		if(!NT_SUCCESS(Status))
		{
			ZwClose(*KeyHandle);
			*KeyHandle = NULL;
			return Status;
		}
	}
	else
		*GrantedAccess = KEY_ALL_ACCESS;

	return Status;
}

//must be called with either ExtraAttributes & OBJ_KERNEL_HANDLE or in system context
NTSTATUS NTAPI RegfsOpenKey2(PHANDLE KeyHandle, ACCESS_MASK DesiredAccess, PACCESS_MASK GrantedAccess, KPROCESSOR_MODE RequestMode, IN PSECURITY_SUBJECT_CONTEXT SecurityContext OPTIONAL, PWSTR NameBuf, USHORT NameSize, ULONG ExtraAttributes, HANDLE Root)
{
	UNICODE_STRING Name;
	Name.Buffer = NameBuf;
	Name.Length = Name.MaximumLength = NameSize;
	return RegfsOpenKey(KeyHandle, DesiredAccess, GrantedAccess, RequestMode, SecurityContext, &Name, ExtraAttributes, Root);
}

//must be called with either ExtraAttributes & OBJ_KERNEL_HANDLE or in system context
NTSTATUS NTAPI RegfsCreateKey(OUT PHANDLE KeyHandle, IN ACCESS_MASK DesiredAccess, OUT PACCESS_MASK GrantedAccess, IN KPROCESSOR_MODE RequestMode, IN PSECURITY_SUBJECT_CONTEXT SecurityContext OPTIONAL, IN PUNICODE_STRING Name, IN ULONG ExtraAttributes, IN HANDLE Root, IN PSECURITY_DESCRIPTOR NewSecurity OPTIONAL, IN ULONG TitleIndex, IN PUNICODE_STRING Class OPTIONAL, IN ULONG CreateOptions, IN PULONG Disposition)
{
	NTSTATUS Status;
	OBJECT_ATTRIBUTES Oa;
	*KeyHandle = NULL;

	//InitializeObjectAttributes(&Oa, Name, OBJ_KERNEL_HANDLE | (RequestMode == UserMode ? OBJ_FORCE_ACCESS_CHECK : 0) | ExtraAttributes, Root, Security);
	InitializeObjectAttributes(&Oa, Name, ExtraAttributes | DefaultObjectAttributes, Root, NewSecurity);

	//should the key be opened instead?
	Status = ZwOpenKey(KeyHandle, KEY_ALL_ACCESS, &Oa);

	if(Status == STATUS_OBJECT_NAME_NOT_FOUND)
	{//really create
		if(RequestMode == UserMode) //needs access check on parent for create
		{
			UNICODE_STRING ParentName;
			*KeyHandle = NULL;

			RegfsSeperateLastName(*Name, NULL, &ParentName);

			if(ParentName.Length)
			{
				OBJECT_ATTRIBUTES ParentOa;
				HANDLE ParentHandle;
				ACCESS_MASK ParentGrantedAccess;

				InitializeObjectAttributes(&ParentOa, &ParentName, ExtraAttributes | DefaultObjectAttributes, Root, NewSecurity);

				Status = ZwOpenKey(&ParentHandle, KEY_ALL_ACCESS, &ParentOa);

				if(!NT_SUCCESS(Status)) //the parent must exist
					return Status;

				if(RegfsIsKeyHandleBogus(ParentHandle))
					return STATUS_ACCESS_DENIED; //i.e. not accessible from KM

				Status = RegfsManualKeySecurityCheck(ParentHandle, SecurityContext, KEY_CREATE_SUB_KEY, &ParentGrantedAccess);

				ZwClose(ParentHandle);

				if(!NT_SUCCESS(Status))
					return Status;
			}
			else //direct decendent: we already have a key handle
			{
				ACCESS_MASK ParentGrantedAccess;

				Status = RegfsManualKeySecurityCheck(Root, SecurityContext, KEY_CREATE_SUB_KEY, &ParentGrantedAccess);
				
				if(!NT_SUCCESS(Status))
				{
					return Status;
				}
			}
		}
	}
	else if(NT_SUCCESS(Status))
		ZwClose(*KeyHandle);
	else
	{
		*KeyHandle = NULL;
		return Status;
	}

	Status = ZwCreateKey(KeyHandle, KEY_ALL_ACCESS, &Oa, TitleIndex, Class, CreateOptions, Disposition);

	if(!NT_SUCCESS(Status))
	{
		*KeyHandle = NULL;
		return Status;
	}

	if(RegfsIsKeyHandleBogus(*KeyHandle))
		return STATUS_ACCESS_DENIED; //i.e. not accessible from KM

	if(DesiredAccess & ACCESS_SYSTEM_SECURITY)
	{
		if(!SePrivilegeCheck(&SecurityPrivilegeSet, SecurityContext, RequestMode))
		{
			ZwClose(*KeyHandle);
			*KeyHandle = NULL;
			return STATUS_PRIVILEGE_NOT_HELD;
		}
	}

	*GrantedAccess = DesiredAccess; //creator of an object gets whatever access they want

	return Status;
}

//must be called with either ExtraAttributes & OBJ_KERNEL_HANDLE or in system context
NTSTATUS NTAPI RegfsCreateKey2(PHANDLE KeyHandle, ACCESS_MASK DesiredAccess, PACCESS_MASK GrantedAccess, KPROCESSOR_MODE RequestMode, IN PSECURITY_SUBJECT_CONTEXT SecurityContext OPTIONAL, PWSTR NameBuf, USHORT NameSize, ULONG ExtraAttributes, HANDLE Root, PSECURITY_DESCRIPTOR Security, ULONG TitleIndex, PUNICODE_STRING Class, ULONG CreateOptions, PULONG Disposition)
{
	UNICODE_STRING Name;
	Name.Buffer = NameBuf;
	Name.Length = Name.MaximumLength = NameSize;
	return RegfsCreateKey(KeyHandle, DesiredAccess, GrantedAccess, RequestMode, SecurityContext, &Name, ExtraAttributes, Root, Security, TitleIndex, Class, CreateOptions, Disposition);
}

NTSTATUS NTAPI RegfsReplaceFileName(IN OUT PFILE_OBJECT File, IN PUNICODE_STRING BaseName, IN PUNICODE_STRING ExtraName OPTIONAL)
{
	UNICODE_STRING NewName;

	if(ExtraName)
	{
		USHORT Ct; //character write position

		if((ULONG)BaseName->Length + (ULONG)ExtraName->Length > 0xFFFF)
			return STATUS_NAME_TOO_LONG;

		NewName.Length = BaseName->Length + ExtraName->Length;

		if(BaseName->Buffer[BaseName->Length/sizeof(WCHAR)-1] !=  L'\\')
			NewName.Length += sizeof(WCHAR);

		NewName.Buffer = ExAllocatePool(NonPagedPool, NewName.Length);

		NewName.MaximumLength = NewName.Length;

		if(!NewName.Buffer)
			return STATUS_NO_MEMORY;

		RtlCopyMemory(NewName.Buffer, BaseName->Buffer, BaseName->Length);
		Ct = BaseName->Length / sizeof(WCHAR);
		
		if(BaseName->Buffer[Ct-1] !=  L'\\')
		{
			NewName.Buffer[Ct] = L'\\';
			Ct++;
		}

		RtlCopyMemory(NewName.Buffer + Ct, ExtraName->Buffer, ExtraName->Length);
	}
	else
	{
		NewName.Length = NewName.MaximumLength = BaseName->Length;
		NewName.Buffer = ExAllocatePool(NonPagedPool, NewName.Length);

		if(!NewName.Buffer)
			return STATUS_NO_MEMORY;

		RtlCopyMemory(NewName.Buffer, BaseName->Buffer, NewName.Length);
	}

	ExFreePool(File->FileName.Buffer);
	File->FileName = NewName;

	return STATUS_SUCCESS;
}

VOID RegfsUpdateFsRtlSizes(IN PFILE_OBJECT FileObject, IN ULONG ActualSize)
{
	BOOLEAN ReleaseFcbLock;
	PREGFS_FCB_FSRTL FcbFsRtl = (PREGFS_FCB_FSRTL)FileObject->FsContext;

	if(!FcbFsRtl)
		return;

	if(ExIsResourceAcquired(FcbFsRtl->FsRtl.Resource))
		ReleaseFcbLock = FALSE;
	else
	{
		ExAcquireResourceExclusiveLite(FcbFsRtl->FsRtl.Resource, TRUE);
		ReleaseFcbLock = TRUE;
	}

#if KERNEL_VERSION >= 0x501
	ExAcquireFastMutex(&FcbFsRtl->FsRtl2Lock);
#endif

	FcbFsRtl->FsRtl.AllocationSize.LowPart =
	FcbFsRtl->FsRtl.ValidDataLength.LowPart =
	FcbFsRtl->FsRtl.FileSize.LowPart = ActualSize;

#if KERNEL_VERSION >= 0x501
	ExReleaseFastMutex(&FcbFsRtl->FsRtl2Lock);
#endif

	if(ReleaseFcbLock)
		ExReleaseResourceLite(FcbFsRtl->FsRtl.Resource);
}