/*
 * COPYRIGHT:  2007 Christopher Smith
 * LICENCE:    GPL v2: See COPYING in the top level directory
 * PROJECT:    regfs
 * FILE:       dispatch.c
 * PURPOSE:    Registry filesystem worker thread procs
 */

#include "regfs.h"

PRKQUEUE WorkQueue = NULL;

HANDLE ThreadHandles[REGFS_MAX_WORK_THREADS];
UCHAR ThreadCount = 0;

VOID NTAPI RegfsWorkerThread(PRKQUEUE Queue);

VOID NTAPI KillAllThreads(UCHAR Count)
{
	PKTHREAD Threads[REGFS_MAX_WORK_THREADS];
	REGFS_WORK_ITEM_SHORT TerminateItems[REGFS_MAX_WORK_THREADS];
	KWAIT_BLOCK WaitBlocks[REGFS_MAX_WORK_THREADS];
	UCHAR i;

	for(i = 0; i < REGFS_MAX_WORK_THREADS; i++)
	{
		TerminateItems[i].Type = WtTermThread;
		KeInsertQueue(WorkQueue, (PLIST_ENTRY)&TerminateItems[i]);
	}

	for(i = 0; i < Count; i++)
	{
		ObReferenceObjectByHandle(ThreadHandles[i], SYNCHRONIZE, NULL, KernelMode, &Threads[i], NULL);
		ZwClose(ThreadHandles[i]);
	}

	KeWaitForMultipleObjects(Count, Threads, WaitAll, Executive, KernelMode, FALSE, NULL, WaitBlocks);

	for(i = 0; i < Count; i++)
		ObDereferenceObject(Threads[i]);		
}

#pragma code_seg(push, INIT)

NTSTATUS NTAPI RegfsInitWorkThreads(UCHAR CreateThreadCount, UCHAR ConcurrencyCount)
{
	UCHAR i;

	if(CreateThreadCount > REGFS_MAX_WORK_THREADS)
		return STATUS_TOO_MANY_THREADS;

	WorkQueue = ExAllocatePoolWithTag(NonPagedPool, sizeof(KQUEUE), TAG_WORK_QUEUE);
	
	if(!WorkQueue)
		return STATUS_NO_MEMORY;

	KeInitializeQueue(WorkQueue, ConcurrencyCount);

	ThreadCount = CreateThreadCount;

	for(i = 0; i < CreateThreadCount; i++)
	{
		NTSTATUS Status;
		
		Status = PsCreateSystemThread(&ThreadHandles[i], SYNCHRONIZE, NULL, NULL, NULL, (PKSTART_ROUTINE)&RegfsWorkerThread, WorkQueue);

		if(!NT_SUCCESS(Status))
		{
			if(i)
				KillAllThreads(i - 1);

			ExFreePoolWithTag(WorkQueue, TAG_WORK_QUEUE);
			return Status;
		}
	}

	return STATUS_SUCCESS;
}

#pragma code_seg(pop, INIT)

VOID NTAPI RegfsCleanupWorkThreads()
{
	if(ThreadCount)
		KillAllThreads(ThreadCount);

	if(WorkQueue)
		ExFreePoolWithTag(WorkQueue, TAG_WORK_QUEUE);
}

NTSTATUS NTAPI RegfsQueueWorkItem(PREGFS_WORK_ITEM CopyWi)
{
	NTSTATUS Status;
	PREGFS_WORK_ITEM Wi;
	
	Wi = ExAllocatePoolWithTag(NonPagedPool, sizeof(REGFS_WORK_ITEM), TAG_WORK_ITEM);

	if(!Wi)
		return STATUS_NO_MEMORY;

	RtlCopyMemory(Wi, CopyWi, sizeof(REGFS_WORK_ITEM));

	KeInsertQueue(WorkQueue, &Wi->KQueueItem);

	return STATUS_SUCCESS;
}

VOID NTAPI RegfsWorkerThread(PRKQUEUE Queue)
{
	PREGFS_WORK_ITEM Wi;
	while(TRUE)
	{
		DPRINT_V("Waiting for work item...\n");

		Wi = (PREGFS_WORK_ITEM)KeRemoveQueue(Queue, UserMode, NULL);

		DPRINT_V("Recieved item\n");

		switch(Wi->Type)
		{
		case WtTermThread:
			DPRINT_V("Terminating.\n");
			PsTerminateSystemThread(STATUS_SUCCESS);
			return;

		case WtDispatchIrp:
			DPRINT_V("Dispatching Irp.\n");
			RegfsDispatchMethodWorkThreadSystemContext(Wi);
			break;
		}

		ExFreePoolWithTag(Wi, TAG_WORK_ITEM);
	}
}

//executive work item support:

PRKSEMAPHORE ExWorkItemLimit = NULL; //to avoid tying up all executive work items (could cause a deadlock)

VOID NTAPI RegfsExecuteExWorkItem(PREGFS_WORK_ITEM Wi)
{
	DPRINT_V("Recieved ExWorkItem.\n");

	switch(Wi->Type)
	{
	case WtTermThread:
		ASSERTMSG("Can't terminate executive worker!", FALSE);
		return;

	case WtDispatchIrp:
		DPRINT_V("Dispatching Irp.\n");
		RegfsDispatchMethodWorkThreadSystemContext(Wi);
		break;
	}

	ExFreePoolWithTag(Wi, TAG_WORK_ITEM);

	KeReleaseSemaphore(ExWorkItemLimit, 1, 1, FALSE);
}

NTSTATUS NTAPI RegfsInitExWorkItems(ULONG Limit)
{
	ExWorkItemLimit = ExAllocatePoolWithTag(PagedPool, sizeof(KSEMAPHORE), TAG_EX_LIMIT_SEM);

	if(!ExWorkItemLimit)
		return STATUS_NO_MEMORY;

	KeInitializeSemaphore(ExWorkItemLimit, Limit, Limit);

	return STATUS_SUCCESS;
}

VOID NTAPI RegfsCleanupExWorkItems()
{
	if(ExWorkItemLimit)
		ExFreePoolWithTag(ExWorkItemLimit, TAG_EX_LIMIT_SEM);
}

NTSTATUS NTAPI RegfsQueueExWorkItem(PREGFS_WORK_ITEM CopyWi)
{
	NTSTATUS Status;
	PREGFS_WORK_ITEM Wi;
	
	Wi = ExAllocatePoolWithTag(NonPagedPool, sizeof(REGFS_WORK_ITEM), TAG_WORK_ITEM);

	if(!Wi)
		return STATUS_NO_MEMORY;

	RtlCopyMemory(Wi, CopyWi, sizeof(REGFS_WORK_ITEM));

	Wi->ExItem.Parameter = &Wi;
	Wi->ExItem.WorkerRoutine = &RegfsExecuteExWorkItem;

	KeWaitForSingleObject(ExWorkItemLimit, WrUserRequest, KernelMode, FALSE, NULL);

	ExQueueWorkItem(&Wi->ExItem, DelayedWorkQueue);

	return STATUS_SUCCESS;
}
